/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  offload.c                                                                *
 *                                                                           *
 * Description:                                                              *
 *  TCP offload support.                                                     *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/08/08 18:25:22 $ $RCSfile: offload.c,v $ $Revision: 1.16 $ */

#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/notifier.h>
#include <linux/toedev.h>
#include <net/sock.h>
#include <net/tcp.h>
#include <net/offload.h>
#include <asm/semaphore.h>

#ifdef sk_state					/* 2.6 */
# define inet_addr_info(sk) inet_sk(sk)
#else						/* 2.4 */
# define sk_route_caps route_caps
# define sk_family family
# define sk_backlog_rcv backlog_rcv
# define inet_addr_info(sk) sk
#endif

/* Replace this with a R/W semaphore someday.  See kernel/profile.c */
static DECLARE_MUTEX(notify_mutex);
static struct notifier_block *listen_offload_notify_list;

int register_listen_offload_notifier(struct notifier_block *nb)
{
	int err;

	down(&notify_mutex);
	err = notifier_chain_register(&listen_offload_notify_list, nb);
	up(&notify_mutex);
	return err;
}

int unregister_listen_offload_notifier(struct notifier_block *nb)
{
	int err;

	down(&notify_mutex);
	err = notifier_chain_unregister(&listen_offload_notify_list, nb);
	up(&notify_mutex);
	return err;
}

/*
 * Called when an active open has been requested through connect(2).  Decides
 * if the connection may be offloaded based on the system's offload policies
 * and the capabilities of the egress interface.
 *
 * Returns 1 if the connection is offloaded and 0 otherwise.
 */
int tcp_connect_offload(struct sock *sk)
{
	if (sk->sk_route_caps & NETIF_F_TCPIP_OFFLOAD) {
		struct toedev *dev = TOEDEV(__sk_dst_get(sk)->dev);

		if (!dev || !dev->can_offload(dev, sk))
			return 0;
		/* XXX check offload policies */
		if (dev->connect(dev, sk) == 0)
			return 1;
	}
	return 0;
}

/*
 * TOE capable backlog handler.  This is used for offloaded listening sockets
 * so they can deal with non-IP (TOE) packets queued in their backlogs.  We
 * distinguish TOE from IP packets easily as the former lack network headers.
 * Such TOE packets are fed to a TOE-specific backlog handler.
 */
static int listen_backlog_rcv(struct sock *sk, struct sk_buff *skb)
{
	if (likely(skb->h.raw != skb->nh.raw))
		return tcp_v4_do_rcv(sk, skb);
	BLOG_SKB_CB(skb)->backlog_rcv(sk, skb);
	return 0;
}

/*
 * Called when the SW stack has transitioned a socket to listen state.
 * We check if the socket should be offloaded according to the current 
 * offloading policies, and if so, publish an OFFLOAD_LISTEN_START event.
 */
int tcp_listen_offload(struct sock *sk)
{
	/* IPv4 only for now */
	if (sk->sk_family != PF_INET)
		return 0;

	/* filter out loopback listens */
	if (LOOPBACK(inet_addr_info(sk)->rcv_saddr))
		return 0;

	/* if (nf_sock_hook(PF_INET, NF_IP_OFFLOAD, sk) != NF_ACCEPT)
		return 0; */

	/* Install a TOE capable backlog handler */
	sk->sk_backlog_rcv = listen_backlog_rcv;

	down(&notify_mutex);
	notifier_call_chain(&listen_offload_notify_list,
				OFFLOAD_LISTEN_START, sk);
	up(&notify_mutex);
	return 1;
}

/*
 * Called through a netfilter hook when a socket starts listening.
 * Publishes an OFFLOAD_LISTEN_START event.
 */
static int tcp_listen_offload_start(struct sk_buff *skb)
{
	down(&notify_mutex);
	notifier_call_chain(&listen_offload_notify_list, OFFLOAD_LISTEN_START,
				skb->sk);
	up(&notify_mutex);
	return 0;
}

/*
 * Called when the SW stack is preparing to close an existing listening socket.
 * We publish an OFFLOAD_LISTEN_STOP event.
 */
int tcp_listen_offload_stop(struct sock *sk)
{
	down(&notify_mutex);
	notifier_call_chain(&listen_offload_notify_list,
				OFFLOAD_LISTEN_STOP, sk);
	up(&notify_mutex);
	return 0;
}

EXPORT_SYMBOL(register_listen_offload_notifier);
EXPORT_SYMBOL(unregister_listen_offload_notifier);
