/*
 * This file implements the interface between the socket layer and
 * the HW TCP/CPL, including the protocol operations for Chelsio's HW TCP.
 *
 * Large portions of this file are taken from net/ipv4/tcp.c.
 * See that file for copyrights of the original code.
 * Any additional code is
 *
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/toedev.h>
#include <linux/tcp_diag.h>

#include <net/offload.h>
#include <net/tcp.h>
#include <net/ip.h>

#include <asm/uaccess.h>
#include <asm/ioctls.h>

#include "../toe_compat.h"
#include "defs.h"
#include "../cpl5_cmd.h"
#include "tom.h"
#include "l2t.h"

#ifdef CONFIG_CHELSIO_T1_CM
#include "t1_cml/cml_osdep.h"
#endif


static const int TX_HEADER_LEN = sizeof(struct cpl_tx_data);

static inline int chelsio_wspace(struct sock *sk)
{
    struct toedev *dev = TOE_DEV(tcp_sk(sk));

    return TOM_TUNABLE(dev, max_host_sndbuf) - sk->sk_wmem_queued;
}

/*
 * TCP socket write_space callback.
 */
void t1_write_space(struct sock *sk)
{
    struct socket *sock = sk->sk_socket;

    if (chelsio_wspace(sk) >= sk_stream_min_wspace(sk) && sock) {
	clear_bit(SOCK_NOSPACE, &sock->flags);

	if (sk->sk_sleep && waitqueue_active(sk->sk_sleep))
	    wake_up_interruptible(sk->sk_sleep);

	if (sock->fasync_list && !(sk->sk_shutdown & SEND_SHUTDOWN))
	    sock_wake_async(sock, 2, POLL_OUT);
    }
}

static inline struct sk_buff *chelsio_alloc_skb(struct sock *sk, int size,
						int extra_mem, int gfp)
{
    struct sk_buff *skb = alloc_skb(size + TX_HEADER_LEN, gfp);

    if (skb) {
	skb->truesize += extra_mem;
	skb_reserve(skb, TX_HEADER_LEN);
    }
    return skb;
}

#ifndef sk_stream_wait_connect
# define TCP_PAGE(sk)   (sk->sk_sndmsg_page)
# define TCP_OFF(sk)    (sk->sk_sndmsg_off)
#else
extern int wait_for_tcp_connect(struct sock *sk, int flags, long *timeo_p);

#ifdef ip_route_output_flow
# define TCP_PAGE(sk)	(tcp_sk(sk)->sndmsg_page)
# define TCP_OFF(sk)	(tcp_sk(sk)->sndmsg_off)
#else
# define TCP_PAGE(sk)	(inet_sk(sk)->sndmsg_page)
# define TCP_OFF(sk)	(inet_sk(sk)->sndmsg_off)
#endif

static inline int skb_can_coalesce(struct sk_buff *skb, int i,
				   struct page *page, int off)
{
    if (i) {
	skb_frag_t *frag = &skb_shinfo(skb)->frags[i - 1];
	return page == frag->page && off == frag->page_offset + frag->size;
    }
    return 0;
}

#ifdef ip_route_output_flow
static inline void skb_fill_page_desc(struct sk_buff *skb, int i,
				      struct page *page, int off, int size)
{
    skb_frag_t *frag = &skb_shinfo(skb)->frags[i];
    frag->page = page;
    frag->page_offset = off;
    frag->size = size;
    skb_shinfo(skb)->nr_frags = i + 1;
}
#endif

static int sk_stream_error(struct sock *sk, int flags, int err)
{
    if (err == -EPIPE)
	err = sock_error(sk) ? : -EPIPE;
    if (err == -EPIPE && !(flags & MSG_NOSIGNAL))
	send_sig(SIGPIPE, current, 0);
    return err;
}

/*
 * Release an skb if it is no longer needed. This routine
 * must be called with interrupts disabled or with the
 * socket locked so that the sk_buff queue operation is ok.
 */
static inline void sk_eat_skb(struct sock *sk, struct sk_buff *skb)
{
    __skb_unlink(skb, &sk->sk_receive_queue);
    __kfree_skb(skb);
}

#define FINAL_ACK_WAITING_STATE (TCPF_FIN_WAIT1 | TCPF_CLOSING | TCPF_LAST_ACK)

/*
 * Returns true if we are waiting for an ACK to our FIN.  Note that our FIN
 * may not even have gone out yet (still in the write queue).
 */
static inline int closing(struct sock *sk)
{
    return (1 << sk->sk_state) & FINAL_ACK_WAITING_STATE;
}

/*
 * Wait until our FIN is ACKed.
 */
static void sk_stream_wait_close(struct sock *sk, long timeout)
{
    struct task_struct *tsk = current;

#ifndef DEFINE_WAIT
    DECLARE_WAITQUEUE(wait, tsk);

    add_wait_queue(sk->sleep, &wait);
    do {
	set_current_state(TASK_INTERRUPTIBLE);
	if (!closing(sk))
	    break;
	release_sock(sk);
	timeout = schedule_timeout(timeout);
	lock_sock(sk);
    } while (!signal_pending(tsk) && timeout);

    tsk->state = TASK_RUNNING;
    remove_wait_queue(sk->sleep, &wait);
#else
    DEFINE_WAIT(wait);

    do {
	prepare_to_wait(sk->sk_sleep, &wait, TASK_INTERRUPTIBLE);
	if (!closing(sk))
	    break;
	release_sock(sk);
	timeout = schedule_timeout(timeout);
	lock_sock(sk);
    } while (!signal_pending(tsk) && timeout);

    finish_wait(sk->sk_sleep, &wait);
#endif
}
#endif

static inline int tcp_memory_free(struct sock *sk)
{
    struct toedev *dev = TOE_DEV(tcp_sk(sk));

    return sk->sk_wmem_queued < TOM_TUNABLE(dev, max_host_sndbuf);
}

/*
 * Wait for memory to become available, either space in a socket's send buffer
 * or system memory.
 */
static int wait_for_mem(struct sock *sk, long *timeout)
{
	int err = 0;
	long vm_wait = 0;
	long current_timeo = *timeout;

#ifdef DEFINE_WAIT
	DEFINE_WAIT(wait);
#else
	DECLARE_WAITQUEUE(wait, current);
#endif

	if (tcp_memory_free(sk))
		current_timeo = vm_wait = (net_random() % (HZ / 5)) + 2;

#ifndef DEFINE_WAIT
	add_wait_queue(sk->sk_sleep, &wait);
#endif
	for (;;) {
		set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);

#ifdef DEFINE_WAIT
		prepare_to_wait(sk->sk_sleep, &wait, TASK_INTERRUPTIBLE);
#else
		set_current_state(TASK_INTERRUPTIBLE);
#endif
		if (sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN))
			goto do_error;
		if (!*timeout)
			goto do_nonblock;
		if (signal_pending(current))
			goto do_interrupted;
		clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);
		if (tcp_memory_free(sk) && !vm_wait)
			break;

		set_bit(SOCK_NOSPACE, &sk->sk_socket->flags);
		WRITE_PENDING(sk, tcp_sk(sk))++;
		release_sock(sk);
		if (!tcp_memory_free(sk) || vm_wait)
			current_timeo = schedule_timeout(current_timeo);
		lock_sock(sk);
		WRITE_PENDING(sk, tcp_sk(sk))--;

		if (vm_wait) {
			vm_wait -= current_timeo;
			current_timeo = *timeout;
			if (current_timeo != MAX_SCHEDULE_TIMEOUT &&
			    (current_timeo -= vm_wait) < 0)
				current_timeo = 0;
			vm_wait = 0;
		}
		*timeout = current_timeo;
	}
out:
#ifdef DEFINE_WAIT
	finish_wait(sk->sk_sleep, &wait);
#else
	current->state = TASK_RUNNING;
	remove_wait_queue(sk->sk_sleep, &wait);
#endif
	return err;

do_error:
	err = -EPIPE;
	goto out;
do_nonblock:
	err = -EAGAIN;
	goto out;
do_interrupted:
	err = sock_intr_errno(*timeout);
	goto out;
}

static inline void skb_set_ulp_mode(struct sk_buff *skb, int mode, int submode)
{
    skb_ulp_mode(skb) = (mode << 4) | submode;
}

void skb_entail(struct sock *sk, struct sk_buff *skb)
{
    struct tcp_opt *tp = tcp_sk(sk);

    TCP_SKB_CB(skb)->seq = tp->write_seq;
    TCP_SKB_CB(skb)->flags = 0;
    __skb_queue_tail(&sk->sk_write_queue, skb);
    sk->sk_wmem_queued += skb->truesize;
    // tcp_charge_skb(sk, skb);

    // Do not share pages across sk_buffs
    if (TCP_PAGE(sk) && TCP_OFF(sk)) {
	put_page(TCP_PAGE(sk));
	TCP_PAGE(sk) = NULL;
    }
}

/*
 * Returns true if a connection is in a PDU based mode.
 */
static inline int uses_pdus(struct tcp_opt *tp)
{
    return ULP_MODE(tp) >= ULP_MODE_ISCSI;
}

/*
 * Returns true if a new send may be merged into existing sk_buffs in the
 * send queue.  This is not possible for ULPs that need to preserve PDU
 * boundaries.
 */
static inline int may_merge(struct tcp_opt *tp, int flags)
{
    return !uses_pdus(tp);
}

/*
 * Returns true if a connection should send more data to the TOE ASAP.
 */
static inline int should_push(struct tcp_opt *tp)
{
    struct toedev *dev;

    if ((tp == NULL) || ((dev = TOE_DEV(tp)) == NULL))
	return 0;

    return tp->snd_nxt - tp->snd_una <= TOM_TUNABLE(dev, tx_hold_thres) ||
	(tp->nonagle & TCP_NAGLE_OFF);
}

/*
 * Returns true if a TCP socket is corked.
 */
static inline int corked(struct tcp_opt *tp, int flags)
{
    return (flags & MSG_MORE) || (tp->nonagle & TCP_NAGLE_CORK);
}

/*
 * Returns true if a send should try to push new data.
 */
static inline int send_should_push(struct tcp_opt *tp, int flags)
{
    return should_push(tp) && !corked(tp, flags);
}

static inline void mark_tx_ready(struct sk_buff *skb)
{
    TCP_SKB_CB(skb)->flags |= TCPCB_FLAG_PSH;
}

static inline void tcp_mark_urg(struct tcp_opt *tp, int flags,
				struct sk_buff *skb)
{
    if (flags & MSG_OOB) {
	tp->urg_mode = 1;
	tp->snd_up = tp->write_seq;
	TCP_SKB_CB(skb)->flags |= TCPCB_FLAG_URG;
    }
}

static inline void tcp_push(struct sock *sk, struct tcp_opt *tp, int flags,
			    int mss)
{
    int qlen = skb_queue_len(&sk->sk_write_queue);

    if (qlen) {
	struct sk_buff *skb = sk->sk_write_queue.prev;

	tcp_mark_urg(tp, flags, skb);
	if (skb->len == mss || !corked(tp, flags)) {
	    mark_tx_ready(skb);
	    if (qlen == 1 && (skb->len == mss || should_push(tp)))
		t1_push_frames(sk);
	}
    }
}

static void tcp_uncork(struct sock *sk)
{
    struct tcp_opt *tp = tcp_sk(sk);

    if (tp->nonagle & TCP_NAGLE_CORK) {
	tp->nonagle &= ~TCP_NAGLE_CORK;
	tcp_push(sk, tp, 0, -1);
    }
}

static inline void push_frames_if_head(struct sock *sk)
{
    if (skb_queue_len(&sk->sk_write_queue) == 1)
	t1_push_frames(sk);
}

static int wait_for_pm_memory(struct sock *sk, long *timeo)
{
    int err = 0;
    DECLARE_WAITQUEUE(wait, current);

    add_wait_queue(sk->sk_sleep, &wait);
    while (1) {
	long current_timeo, t;

	set_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);
	if (unlikely(sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN))) {
	    err = -EPIPE;
	    break;
	}
	if (!*timeo) {
	    err = -EAGAIN;
	    break;
	}
	if (unlikely(signal_pending(current))) {
	    err = sock_intr_errno(*timeo);
	    break;
	}
	clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);

	if (t1_alloc_tx_pages(sk, 0))
		break;

	set_current_state(TASK_INTERRUPTIBLE);
	set_bit(SOCK_NOSPACE, &sk->sk_socket->flags);
	t = (net_random() % HZ) + 5;
	if (t > *timeo) t = *timeo;

	release_sock(sk);

	current_timeo = schedule_timeout(t);

	lock_sock(sk);
	t -= current_timeo;
	if (*timeo != MAX_SCHEDULE_TIMEOUT)
	    *timeo -= t;
    }

    current->state = TASK_RUNNING;
    remove_wait_queue(sk->sk_sleep, &wait);
    return err;
}

#ifdef CONFIG_CHELSIO_T1_CM
static int wait_for_cm_finish(struct sock *sk, long *timeo)
{
    int err = 0;
    DECLARE_WAITQUEUE(wait, current);
                                                                                
    add_wait_queue(sk->sk_sleep, &wait);
    do {
        long current_timeo, t;
                                                                                
        if (unlikely(sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN))) {
            err = -EPIPE;
            break;
        }
        if (!*timeo) {
            err = -EAGAIN;
            break;
        }
        if (unlikely(signal_pending(current))) {
            err = sock_intr_errno(*timeo);
            break;
        }
                                                                                
        set_current_state(TASK_INTERRUPTIBLE);
        t = (net_random() % HZ) + 5;
        if (t > *timeo) t = *timeo;
                                                                                
        release_sock(sk);
                                                                                
        current_timeo = schedule_timeout(t);
                                                                                
        lock_sock(sk);
        t -= current_timeo;
        if (*timeo != MAX_SCHEDULE_TIMEOUT)
            *timeo -= t;
    } while (sock_flag(sk, SOCK_MIGRATE));
                                                                                
    current->state = TASK_RUNNING;
    remove_wait_queue(sk->sk_sleep, &wait);
    return err;
}
#endif

static int chelsio_sendpage(struct sock *sk, struct page *page, int offset,
			    size_t size, int flags)
{
    struct tcp_opt *tp;
    int mss, err, copied = 0;
    long timeo;
    struct toedev *tdev;
    struct l2t_entry *e;
    struct tom_data *d;

    lock_sock(sk);
    tp = tcp_sk(sk);
    tdev = TOE_DEV(tp);
    e = L2T_ENTRY(tp);

    timeo = sock_sndtimeo(sk, flags & MSG_DONTWAIT);

    /* Wait for connection establishment to finish. */
    if ((1 << sk->sk_state) & ~(TCPF_ESTABLISHED | TCPF_CLOSE_WAIT))
	if ((err = sk_stream_wait_connect(sk, &timeo)) != 0)
	    goto out_err;

    clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);

    err = -EPIPE;
    if (sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN))
	goto out_err;

    if (!TX_PAGES_ALLOCED(tp) && !t1_alloc_tx_pages(sk, 0) &&
	(err = wait_for_pm_memory(sk, &timeo)) != 0)
	goto out_err;

#ifdef CONFIG_CHELSIO_T1_CM
    if ( sock_flag(sk, SOCK_MIGRATE) &&
         (err = wait_for_cm_finish(sk, &timeo)) != 0 ) {
         goto out_err;
    }
#endif

    d = TOM_DATA(tdev);
    mss = min((int)tp->mss_cache, (int)d->lmss[e->real_dev->if_port]);

    while (size > 0) {
	int copy, i;
	struct sk_buff *skb = skb_peek_tail(&sk->sk_write_queue);

	if (!skb || (copy = mss - skb->len) <= 0) {
new_segment:
	    if (!tcp_memory_free(sk))
		goto wait_for_sndbuf;

	    skb = alloc_skb(TX_HEADER_LEN, sk->sk_allocation);
	    if (!skb)
		goto wait_for_memory;

	    skb_reserve(skb, TX_HEADER_LEN);
	    skb_entail(sk, skb);
	    if (ULP_SUBMODE(tp))
	        skb_set_ulp_mode(skb, ULP_MODE(tp), ULP_SUBMODE(tp));
	    copy = mss;
	}

	if (copy > size)
	    copy = size;

	i = skb_shinfo(skb)->nr_frags;
	if (skb_can_coalesce(skb, i, page, offset)) {
	    skb_shinfo(skb)->frags[i - 1].size += copy;
	} else if (i < MAX_SKB_FRAGS) {
	    get_page(page);
	    skb_fill_page_desc(skb, i, page, offset, copy);
	} else {
	    mark_tx_ready(skb);
	    push_frames_if_head(sk);
	    goto new_segment;
	}

	skb->len += copy;
	skb->data_len += copy;
	skb->truesize += copy;
	sk->sk_wmem_queued += copy;
	tp->write_seq += copy;
	copied += copy;
	offset += copy;
	size -= copy;
	if (!size)
	    goto out;

	if (skb->len == mss) {
	    mark_tx_ready(skb);
	    push_frames_if_head(sk);
	}
	continue;

wait_for_sndbuf:
	set_bit(SOCK_NOSPACE, &sk->sk_socket->flags);
wait_for_memory:
	if ((err = wait_for_mem(sk, &timeo)) != 0)
	    goto do_error;
    }

out:
    if (copied) {
    	struct sk_buff *last_skb;
    	last_skb = skb_peek_tail(&sk->sk_write_queue);
	if (last_skb) 
    		TCP_SKB_CB(last_skb)->flags |= TCPCB_FLAG_ACK;
	tcp_push(sk, tp, flags, mss);
    }
done:
    release_sock(sk);
    return copied;

do_error:
    if (copied)
	goto out;
out_err:
    copied = sk_stream_error(sk, flags, err);
    goto done;
}

/*
 * Add a list of skbs to a socket send queue.  This interface is intended for
 * use by in-kernel ULPs.  The skbs must comply with the max size limit of the
 * device and have a headroom of at least TX_HEADER_LEN bytes.
 */
static int chelsio_sendskb(struct sock *sk, struct sk_buff *skb, int flags)
{
    struct sk_buff *last_skb, *next;
    struct tcp_opt *tp;
    int err, copied = 0;
    long timeo;

    lock_sock(sk);
    tp = tcp_sk(sk);

    /* Won't be adding anything to the last skb, so push it on */
    last_skb = skb_peek_tail(&sk->sk_write_queue);
    if (last_skb)
	mark_tx_ready(last_skb);

    timeo = sock_sndtimeo(sk, flags & MSG_DONTWAIT);

    /* Wait for connection establishment to finish. */
    if ((1 << sk->sk_state) & ~(TCPF_ESTABLISHED | TCPF_CLOSE_WAIT))
	if ((err = sk_stream_wait_connect(sk, &timeo)) != 0)
	    goto out_err;

    clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);

    err = -EPIPE;
    if (sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN))
	goto out_err;

    if (!TX_PAGES_ALLOCED(tp) && !t1_alloc_tx_pages(sk, 0) &&
	(err = wait_for_pm_memory(sk, &timeo)) != 0)
	goto out_err;

    /*
     * We check for send buffer space once for the whole skb list.  It isn't
     * critical if we end up overrunning the send buffer limit as we do not
     * allocate any new memory.  The benefit is we don't need to perform
     * intermediate packet pushes.
     */
    while (!tcp_memory_free(sk)) {
	set_bit(SOCK_NOSPACE, &sk->sk_socket->flags);
	if ((err = wait_for_mem(sk, &timeo)) != 0)
	    goto out_err;
    }

    while (skb) {
	if (unlikely(skb_headroom(skb) < TX_HEADER_LEN)) {
	    err = -EINVAL;
	    goto out_err;
	}

	if (unlikely(skb->len > tp->mss_cache)) {
	    err = -EMSGSIZE;
	    goto out_err;
	}

	next = skb->next;
	skb->next = NULL;
	skb_entail(sk, skb);
	mark_tx_ready(skb);
	copied += skb->len;
	tp->write_seq += skb->len;
	skb = next;
    }
done:
    if (likely(skb_queue_len(&sk->sk_write_queue))) {
    	last_skb = skb_peek_tail(&sk->sk_write_queue);
	if (last_skb)
    		TCP_SKB_CB(last_skb)->flags |= TCPCB_FLAG_ACK;
	t1_push_frames(sk);
    }
    release_sock(sk);
    return copied;

out_err:
    if (copied == 0)
	copied = sk_stream_error(sk, flags, err);
    goto done;
}

/*
 * Add data to an sk_buff page fragment.
 */
static inline int tcp_copy_to_page(struct sock *sk, char *from,
				   struct sk_buff *skb, struct page *page,
				   int off, int copy)
{
    if (copy_from_user(page_address(page) + off, from, copy))
	return -EFAULT;

    skb->len += copy;
    skb->data_len += copy;
    skb->truesize += copy;
    sk->sk_wmem_queued += copy;
    return 0;
}

/*
 * Add data to the main data portion of an sk_buff.
 */
static inline int ch_skb_add_data(struct sk_buff *skb, char *from, int copy)
{
    int orig_len = skb->len;

    if (!copy_from_user(skb_put(skb, copy), from, copy))
	return 0;

    __skb_trim(skb, orig_len);
    return -EFAULT;
}

/*
 * Calculate the size for a new send sk_buff.  It's maximum size so we can
 * pack lots of data into it, unless we plan to send it immediately, in which
 * case we size it more tightly.
 *
 * Note: we don't bother compensating for MSS < PAGE_SIZE because it doesn't
 * arise in normal cases and when it does we are just wasting memory.
 */
static inline int select_size(struct tcp_opt *tp, int io_len, int mss,
			      int flags)
{
    const int pgbreak = SKB_MAX_HEAD(SMP_CACHE_BYTES);

#if 0
    /*
     * If the current send can fill an entire buffer there are a few different
     * ways to distribute the data into the main sk_buff part and additional
     * fragments.  If HW has immediate data support the best is to place the
     * header alone in the main part and all the data in fragments.  This is
     * probably not as good otherwise.
     */
    if (io_len >= mss) return 0;
#endif
    if (io_len >= pgbreak || !send_should_push(tp, flags)) return pgbreak;
    return io_len;
}

/*
 * Remaining buffer capacity below which we should allocate a new sk_buff
 * rather than adding a new fragment to an existing one.  Goal is to avoid
 * small inefficient DMAs.
 */
#define NEW_FRAG_THRES 512

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static int chelsio_sendmsg(struct sock *sk, struct msghdr *msg, int size)
#else
static int chelsio_sendmsg(struct kiocb *iocb, struct sock *sk,
			   struct msghdr *msg, size_t size)
#endif
{
    struct iovec *iov;
    struct tcp_opt *tp;
    struct sk_buff *skb;
    int iovlen, flags, mss, err, extend_last_skb, copied = 0;
    long timeo;
    struct toedev *tdev;
    struct l2t_entry *e;
    struct tom_data *d;

    lock_sock(sk);
    tp = tcp_sk(sk);
    tdev = TOE_DEV(tp);
    e = L2T_ENTRY(tp);

    flags = msg->msg_flags;
    timeo = sock_sndtimeo(sk, flags & MSG_DONTWAIT);

    /* Wait for connection establishment to finish. */
    if ((1 << sk->sk_state) & ~(TCPF_ESTABLISHED | TCPF_CLOSE_WAIT))
	if ((err = sk_stream_wait_connect(sk, &timeo)) != 0)
	    goto out_err;

    /* This should be in poll */
    clear_bit(SOCK_ASYNC_NOSPACE, &sk->sk_socket->flags);

    err = -EPIPE;
    if (sk->sk_err || (sk->sk_shutdown & SEND_SHUTDOWN))
	goto out_err;

#ifdef CONFIG_CHELSIO_T1_CM
    if ( sock_flag(sk, SOCK_MIGRATE) &&
         (err = wait_for_cm_finish(sk, &timeo)) != 0 ) {
         goto out_err;
    }
#endif

    d = TOM_DATA(tdev);
    mss = min((int)tp->mss_cache, (int)d->lmss[e->real_dev->if_port]);
    extend_last_skb = may_merge(tp, flags);
    if (uses_pdus(tp) && size > mss) {      // PDUs must fit in one sk_buff
	err = -EMSGSIZE;
	goto out_err;
    }

    if (!TX_PAGES_ALLOCED(tp) && !t1_alloc_tx_pages(sk, 0) &&
	(err = wait_for_pm_memory(sk, &timeo)) != 0)
	goto out_err;

    /* OK commence sending. */
    for (iovlen = msg->msg_iovlen, iov = msg->msg_iov; iovlen--; iov++) {
	int seglen = iov->iov_len;
	unsigned char *from = iov->iov_base;

	while (seglen > 0) {
	    int copy, tailroom;

	    skb = skb_peek_tail(&sk->sk_write_queue);
	    if (!skb ||	(copy = mss - skb->len) <= 0 || !extend_last_skb) {
new_segment:
		if (!tcp_memory_free(sk))
		    goto wait_for_sndbuf;
		skb = alloc_skb(select_size(tp, size, mss, flags) +
				TX_HEADER_LEN, sk->sk_allocation);
		if (!skb)
		    goto wait_for_memory;

		skb_reserve(skb, TX_HEADER_LEN);
		skb_entail(sk, skb);
	        if (ULP_SUBMODE(tp))
	    	    skb_set_ulp_mode(skb, ULP_MODE(tp), ULP_SUBMODE(tp));
		copy = mss;
		extend_last_skb = 1;
	    }

	    if (copy > seglen)
		copy = seglen;

	    /*
	     * Try to append data to the end of skb.  We first try the main
	     * part of the skb, then its page fragments, and finally it's time
	     * to grab a new skb.
	     */
	    tailroom = skb_tailroom(skb);
	    if (tailroom > 0) {
		/* We have some space in skb head. Superb! */
		if (copy > tailroom)
		    copy = tailroom;
		if ((err = ch_skb_add_data(skb, from, copy)) != 0)
		    goto do_fault;
	    } else {
		int i = skb_shinfo(skb)->nr_frags;
		struct page *page = TCP_PAGE(sk);
		int merge = 0, off = TCP_OFF(sk);

		if (off < PAGE_SIZE && skb_can_coalesce(skb, i, page, off)) {
		    merge = 1;	/* We can extend the last page fragment. */
		} else if (i == MAX_SKB_FRAGS ||
			   mss - skb->len < NEW_FRAG_THRES) {
		    /*
		     * Need to add new fragment and cannot because all
		     * the page slots are used up.
		     */
		    mark_tx_ready(skb);
		    push_frames_if_head(sk);
		    goto new_segment;
		} else if (page) {
		    /* If page is cached, align offset to L1 cache boundary */
		    off = L1_CACHE_ALIGN(off);
		    if (off == PAGE_SIZE) {
			put_page(page);
			TCP_PAGE(sk) = page = NULL;
		    }
		}

		if (!page) { 		    /* Allocate new cache page. */
		    page = alloc_pages(sk->sk_allocation, 0);
		    if (!page) goto wait_for_memory;
		    off = 0;
		}

		if (copy > PAGE_SIZE - off)
		    copy = PAGE_SIZE - off;

		/* Time to copy data. We are close to the end! */
		err = tcp_copy_to_page(sk, from, skb, page, off, copy);
		if (err) {
		    /* If this page was new, give it to the
		     * socket so it does not get leaked.
		     */
		    if (!TCP_PAGE(sk)) {
			TCP_PAGE(sk) = page;
			TCP_OFF(sk) = 0;
		    }
		    goto do_error;
		}

		/* Update the skb. */
		if (merge) {
		    skb_shinfo(skb)->frags[i - 1].size += copy;
		} else {
		    skb_fill_page_desc(skb, i, page, off, copy);
		    if (off + copy < PAGE_SIZE) { // space left, keep page
			get_page(page);
			TCP_PAGE(sk) = page;
		    } else if (TCP_PAGE(sk))
			TCP_PAGE(sk) = NULL;
		}

		TCP_OFF(sk) = off + copy;
	    }

	    tp->write_seq += copy;
	    from += copy;
	    copied += copy;
	    seglen -= copy;
	    size -= copy;
	    if (size == 0)
		goto out;

	    if (skb->len == mss) {
		mark_tx_ready(skb);
    		if (iovlen == 1 && seglen == 0) 
    			TCP_SKB_CB(skb)->flags |= TCPCB_FLAG_ACK;
		push_frames_if_head(sk);
	    }
	    continue;

wait_for_sndbuf:
	    set_bit(SOCK_NOSPACE, &sk->sk_socket->flags);
wait_for_memory:
	    if ((err = wait_for_mem(sk, &timeo)) != 0)
		goto do_error;
	}
    }

out:
    if (copied) {
    	struct sk_buff *last_skb;
    	last_skb = skb_peek_tail(&sk->sk_write_queue);
	if (last_skb) {
    		TCP_SKB_CB(last_skb)->flags |= TCPCB_FLAG_ACK;
	}
	if (uses_pdus(tp)) mss = copied;
	tcp_push(sk, tp, flags, mss);
    }

done:
    release_sock(sk);
    return copied;

do_fault:
    if (!skb->len) {
	__skb_unlink(skb, skb->list);
	// tcp_free_skb(sk, skb);
	sk->sk_wmem_queued -= skb->truesize;
	__kfree_skb(skb);
    }

do_error:
    if (copied)
	goto out;
out_err:
    copied = sk_stream_error(sk, flags, err);
    goto done;
}

extern int tcp_recv_urg(struct sock *sk, long timeo, struct msghdr *msg,
			int len, int flags, int *addr_len);

/*
 * Set of states for which we should return RX credits.
 */
#define CREDIT_RETURN_STATE (TCPF_ESTABLISHED | TCPF_FIN_WAIT1 | TCPF_FIN_WAIT2)

/*
 * Called after some received data has been read.  It returns RX credits
 * to the HW for the amount of data processed.
 */
static void cleanup_rbuf(struct sock *sk, int copied)
{
    if (((1 << sk->sk_state) & CREDIT_RETURN_STATE)) {
	struct tcp_opt *tp = tcp_sk(sk);
	struct toedev *dev = TOE_DEV(tp);
	u32 thres = TOM_TUNABLE(dev, rx_credit_thres);

	if (likely(thres)) {
	    u32 credits = tp->copied_seq - tp->rcv_wup;

	    /*
	     * For coalescing to work effectively ensure the receive window
	     * has at least 16KB left.
	     */
	    int must_send = credits + 16384 >= tp->rcv_wnd;

	    if (must_send || credits >= thres)
		tp->rcv_wup += t1_send_rx_credits(dev, TID(tp), credits,
						  must_send);
	}
    }
}

static inline struct sk_buff *tcp_recv_skb(struct sock *sk, u32 seq, u32 *off)
{
    struct sk_buff *skb = NULL;
    u32 offset;

    skb_queue_walk(&sk->sk_receive_queue, skb) {
	offset = seq - TCP_SKB_CB(skb)->seq;
	if (offset < skb->len) {
	    *off = offset;
	    break;
	}
    }
    return skb;
}

/*
 * Receive data from a socket into an application buffer.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
static int chelsio_recvmsg(struct sock *sk, struct msghdr *msg,
			   int len, int nonblock, int flags, int *addr_len)
#else
static int chelsio_recvmsg(struct kiocb *iocb, struct sock *sk,
			   struct msghdr *msg, size_t len, int nonblock,
			   int flags, int *addr_len)
#endif
{
    struct tcp_opt *tp;
    int copied = 0;
    u32 peek_seq, *seq;
    unsigned long avail;  /* amount of available data in current skb */
    int target;		  /* Read at least this many bytes */
    long timeo;

    lock_sock(sk);
    tp = tcp_sk(sk);

    timeo = sock_rcvtimeo(sk, nonblock);

    /* Urgent data needs to be handled specially. */
    if (unlikely(flags & MSG_OOB)) {
	copied = tcp_recv_urg(sk, timeo, msg, len, flags, addr_len);
	goto out;
    }

    seq = &tp->copied_seq;
    if (flags & MSG_PEEK) {
	peek_seq = tp->copied_seq;
	seq = &peek_seq;
    }

    target = sock_rcvlowat(sk, flags & MSG_WAITALL, len);

    do {
	struct sk_buff *skb;
	u32 offset;

	/* Are we at urgent data? Stop if we have read anything. */
	if (tp->urg_data && copied && tp->urg_seq == *seq)
	    break;

	/* We need to check signals first, to get correct SIGURG
	 * handling. FIXME: Need to check this doesn't impact 1003.1g
	 * and move it down to the bottom of the loop
	 */
	if (signal_pending(current)) {
	    if (!copied)
		copied = timeo ? sock_intr_errno(timeo) : -EAGAIN;
	    break;
	}

	skb = skb_peek(&sk->sk_receive_queue);
	if (!skb) goto empty_rcvq;

	/* The loop iterates only for MSG_PEEK */
	do {
	    offset = *seq - TCP_SKB_CB(skb)->seq;
	    if (offset < skb->len)
		goto found_ok_skb;
	    skb = skb->next;
	} while (skb != (struct sk_buff *)&sk->sk_receive_queue);

empty_rcvq:
	/* Well, if we have backlog, try to process it now yet. */
	if (copied >= target && !sk->sk_backlog.tail)
	    break;

	if (copied) {
	    if (sk->sk_err || sk->sk_state == TCP_CLOSE ||
		(sk->sk_shutdown & RCV_SHUTDOWN) || !timeo ||
		(flags & MSG_PEEK))
		break;
	} else {
	    if (sock_flag(sk, SOCK_DONE))
		break;
	    if (sk->sk_err) {
		copied = sock_error(sk);
		break;
	    }
	    if (sk->sk_shutdown & RCV_SHUTDOWN)
		break;
	    if (sk->sk_state == TCP_CLOSE) {
		if (!sock_flag(sk, SOCK_DONE)) {
		    /*
		     * This occurs when user tries to read from never
		     * connected socket.
		     */
		    copied = -ENOTCONN;
		}
		break;
	    }
	    if (!timeo) {
		copied = -EAGAIN;
		break;
	    }
	}

	cleanup_rbuf(sk, copied);

	if (copied >= target) {
	    /* Do not sleep, just process backlog. */
	    release_sock(sk);
	    lock_sock(sk);
	} else
	    sk_wait_data(sk, &timeo);

	if ((flags & MSG_PEEK) && peek_seq != tp->copied_seq) {
	    if (net_ratelimit())
		printk(KERN_DEBUG "TCP(%s:%d): Application bug, race in MSG_PEEK.\n",
		       current->comm, current->pid);
	    peek_seq = tp->copied_seq;
	}
	continue;

found_ok_skb:
	avail = skb->len - offset;
	if (len < avail)
	    avail = len;

	/* Do we have urgent data here? We need to skip over the urgent byte */
	if (unlikely(tp->urg_data)) {
	    u32 urg_offset = tp->urg_seq - *seq;

	    if (urg_offset < avail) {
		/*
		 * The amount of data we are preparing to copy contains
		 * urgent data.
		 */
		if (!urg_offset) {
		    if (!sock_flag(sk, SOCK_URGINLINE)) {
			++*seq;
			offset++;
			avail--;
			if (!avail)
			    goto skip_copy;
		    }
		} else
		    avail = urg_offset;  /* stop short of the urgent data */
	    }
	}

	/*
	 * If MSG_TRUNC is specified the data is discarded.
	 */
	if (!(flags & MSG_TRUNC)) {
	    if (skb_copy_datagram_iovec(skb, offset, msg->msg_iov, avail)) {
		if (!copied)
		    copied = -EFAULT;
		break;
	    }
	}

	*seq += avail;
	copied += avail;
	len -= avail;

skip_copy:
	if (tp->urg_data && after(tp->copied_seq, tp->urg_seq)) {
	    tp->urg_data = 0;
	}
	/* Free the buffer if it's fully consumed */
	if (avail + offset >= skb->len && !(flags & MSG_PEEK))
	    sk_eat_skb(sk, skb);
    } while (len > 0);

    if (copied)
	cleanup_rbuf(sk, copied);
out:
    release_sock(sk);
    return copied;
}

/*
 * Close a connection by sending a CPL_CLOSE_CON_REQ message.  Cannot fail
 * under any circumstances.  We take the easy way out and always queue the
 * message to the write_queue.  We can optimize the case where the queue is
 * already empty though the optimization is probably not worth it.
 */
static void close_conn(struct sock *sk)
{
    struct sk_buff *skb = alloc_skb_nofail(sizeof(struct cpl_close_con_req));

    tcp_uncork(sk);
    skb_entail(sk, skb);
    TCP_SKB_CB(skb)->flags = TCPCB_FLAG_FIN | TCPCB_FLAG_PSH;
    if (sk->sk_state != TCP_SYN_SENT)
	push_frames_if_head(sk);
}

/*
 * State transitions and actions for close.  Note that if we are in SYN_SENT
 * we remain in that state as we cannot control a connection while it's in
 * SYN_SENT; such connections are allowed to establish and are then aborted.
 */
static unsigned char new_state[16] = {
  /* current state:     new state:      action:	*/
  /* (Invalid)	     */ TCP_CLOSE,
  /* TCP_ESTABLISHED */ TCP_FIN_WAIT1 | TCP_ACTION_FIN,
  /* TCP_SYN_SENT    */ TCP_SYN_SENT,
  /* TCP_SYN_RECV    */ TCP_FIN_WAIT1 | TCP_ACTION_FIN,
  /* TCP_FIN_WAIT1   */ TCP_FIN_WAIT1,
  /* TCP_FIN_WAIT2   */ TCP_FIN_WAIT2,
  /* TCP_TIME_WAIT   */ TCP_CLOSE,
  /* TCP_CLOSE	     */ TCP_CLOSE,
  /* TCP_CLOSE_WAIT  */ TCP_LAST_ACK  | TCP_ACTION_FIN,
  /* TCP_LAST_ACK    */ TCP_LAST_ACK,
  /* TCP_LISTEN	     */ TCP_CLOSE,
  /* TCP_CLOSING     */ TCP_CLOSING,
};

/*
 * Perform a state transition during close and return the actions indicated
 * for the transition.  Do not make this function inline, the main reason
 * it exists at all is to avoid multiple inlining of tcp_set_state.
 */
static int make_close_transition(struct sock *sk)
{
    int next = (int)new_state[sk->sk_state];

    tcp_set_state(sk, next & TCP_STATE_MASK);
    return next & TCP_ACTION_FIN;
}

#define SHUTDOWN_ELIGIBLE_STATE (TCPF_ESTABLISHED | TCPF_SYN_RECV | TCPF_CLOSE_WAIT)

/*
 * Shutdown the sending side of a connection. Much like close except
 * that we don't receive shut down or set_sock_flag(sk, SOCK_DEAD).
 *
 * Note: this does not do anything for SYN_SENT state as tcp_shutdown
 * does, however this function is not really called for SYN_SENT because
 * inet_shutdown handles that state specially.  So no harm.
 */
static void chelsio_shutdown(struct sock *sk, int how)
{
    if ((how & SEND_SHUTDOWN) &&
	((1 << sk->sk_state) & SHUTDOWN_ELIGIBLE_STATE) &&
	make_close_transition(sk)) {
#ifdef CONFIG_CHELSIO_T1_CM
	if (sock_flag(sk, SOCK_MIGRATE))
	    t1_cml_sk_abort(sk, 1);
#endif
	close_conn(sk);
    }
}

static void chelsio_close(struct sock *sk, long timeout)
{
    int data_lost;

    lock_sock(sk);
    sk->sk_shutdown |= SHUTDOWN_MASK;

#ifdef CONFIG_CHELSIO_T1_CM
    if (sock_flag(sk, SOCK_MIGRATE)) {
        t1_cml_sk_abort(sk, 0);
    }
#endif

    /*
     * We need to flush the receive buffs.  We do this only on the
     * descriptor close, not protocol-sourced closes, because the
     * reader process may not have drained the data yet!  Make a note
     * of whether any received data will be lost so we can decide whether
     * to FIN or RST.
     */
    data_lost = skb_queue_len(&sk->sk_receive_queue);
    __skb_queue_purge(&sk->sk_receive_queue);

    if (sk->sk_state == TCP_CLOSE) // Do nothing if we are already closed
	;
    else if (data_lost || sk->sk_state == TCP_SYN_SENT) {
	// Unread data was tossed, zap the connection.
	NET_INC_STATS_USER(LINUX_MIB_TCPABORTONCLOSE);
	t1_send_reset(sk, CPL_ABORT_SEND_RST, NULL);
	if (is_bound(sk))
	    tcp_put_port(sk);
	goto unlock;
    } else if (sock_flag(sk, SOCK_LINGER) && !sk->sk_lingertime) {
	/* Check zero linger _after_ checking for unread data. */
	sk->sk_prot->disconnect(sk, 0);
	NET_INC_STATS_USER(LINUX_MIB_TCPABORTONDATA);
    } else if (make_close_transition(sk)) {  // Regular FIN-based close
	close_conn(sk);
    }

    if (timeout)
	sk_stream_wait_close(sk, timeout);

 unlock:
    release_sock(sk);  /* Final release_sock in the connection's lifetime. */

    /*
     * There are no more user references at this point.  Grab the socket
     * spinlock and finish the close.
     */
    local_bh_disable();
    bh_lock_sock(sk);

    sock_hold(sk);    // must survive past the potential tcp_destroy_sock
    sock_orphan(sk);

#if 0
    if (sk->sk_state == TCP_FIN_WAIT2) {
	struct tcp_opt *tp = tcp_sk(sk);
	if (tp->linger2 < 0) {
	    tcp_set_state(sk, TCP_CLOSE);
	    tcp_send_active_reset(sk, GFP_ATOMIC);
	    NET_INC_STATS_BH(TCPAbortOnLinger);
	} else {
	    int tmo = tcp_fin_time(tp);

	    if (tmo > TCP_TIMEWAIT_LEN) {
		tcp_reset_keepalive_timer(sk, tcp_fin_time(tp));
	    } else {
		atomic_inc(&tcp_orphan_count);
		tcp_time_wait(sk, TCP_FIN_WAIT2, tmo);
		goto out;
	    }
	}
    }
    if (sk->sk_state != TCP_CLOSE) {
	tcp_mem_reclaim(sk);
	if (atomic_read(&tcp_orphan_count) > sysctl_tcp_max_orphans ||
	    (sk->sk_wmem_queued > SOCK_MIN_SNDBUF &&
	     atomic_read(&tcp_memory_allocated) > sysctl_tcp_mem[2])) {
	    if (net_ratelimit())
		printk(KERN_INFO "TCP: too many orphaned sockets\n");
	    tcp_set_state(sk, TCP_CLOSE);
	    tcp_send_active_reset(sk, GFP_ATOMIC);
	    NET_INC_STATS_BH(TCPAbortOnMemory);
	}
    }
#endif

    atomic_inc(&tcp_orphan_count);
    if (sk->sk_state == TCP_CLOSE)
	tcp_destroy_sock(sk);
#if 0
out:
#endif
    bh_unlock_sock(sk);
    local_bh_enable();
    sock_put(sk);
}

/*
 * Our analog of tcp_free_skb().
 */
static inline void chelsio_tcp_free_skb(struct sock *sk, struct sk_buff *skb)
{
    sk->sk_wmem_queued -= skb->truesize;
    __kfree_skb(skb);
}

void t1_purge_write_queue(struct sock *sk)
{
    struct sk_buff *skb;

    while ((skb = __skb_dequeue(&sk->sk_write_queue)))
	chelsio_tcp_free_skb(sk, skb);
    // tcp_mem_reclaim(sk);
}

/*
 * Switch a socket to the SW TCP's protocol operations.
 */
static inline void install_standard_ops(struct sock *sk)
{
    sk->sk_prot = &tcp_prot;
    sk->sk_backlog_rcv = sk->sk_prot->backlog_rcv;
    sk->sk_write_space = sk_stream_write_space;
    tcp_sk(sk)->toe_specific = NULL;
}

/*
 * Wait until we get an ABORT_RPL.  Note that on return the backlog must be
 * empty because there can be no CPL messages after an ABORT_RPL.
 */
static void wait_for_abort_rpl(struct sock *sk)
{
    wait_queue_head_t waitq;
    struct task_struct *tsk = current;
    DECLARE_WAITQUEUE(wait, tsk);
    struct tcp_opt *tp = tcp_sk(sk);

    /*
     * We want this to work even when there's no associated struct socket.
     * In that case we provide a temporary wait_queue_head_t.
     */
    if (!sk->sk_sleep) {
	init_waitqueue_head(&waitq);
	sk->sk_sleep = &waitq;
    }

    add_wait_queue(sk->sk_sleep, &wait);
    while (CLOSE_FLAGS(tp) & ABORT_RPL_PENDING) {
	set_task_state(tsk, TASK_UNINTERRUPTIBLE);
	release_sock(sk);
	if (CLOSE_FLAGS(tp) & ABORT_RPL_PENDING)
	    schedule();
	lock_sock(sk);
    }
    __set_task_state(tsk, TASK_RUNNING);
    remove_wait_queue(sk->sk_sleep, &wait);

    if (sk->sk_sleep == &waitq)
	sk->sk_sleep = NULL;

    BUG_TRAP(sk->sk_state == TCP_CLOSE);
}

static int chelsio_disconnect(struct sock *sk, int flags)
{
    struct tcp_opt *tp = tcp_sk(sk);

#ifdef CONFIG_CHELSIO_T1_CM
    if (sock_flag(sk, SOCK_MIGRATE)) {
        t1_cml_sk_abort(sk, 1);
    }
#endif

    __skb_queue_purge(&sk->sk_receive_queue);
    __skb_queue_purge(&tp->out_of_order_queue);
    t1_purge_write_queue(sk);

    if (sk->sk_state != TCP_CLOSE) {
	sk->sk_err = ECONNRESET;
	t1_send_reset(sk, CPL_ABORT_SEND_RST, NULL);
	wait_for_abort_rpl(sk);
    }

    /*
     * We don't know the correct value for max_window but we know an
     * upper limit.
     */
    tp->max_window = 0xFFFF << tp->send_wscale;

    /*
     * Now switch to Linux's TCP operations and let it finish the job.
     */
    install_standard_ops(sk);
    return tcp_disconnect(sk, flags);
}

/*
 * Our version of tcp_v4_destroy_sock().  We need to do this because
 * tcp_writequeue_purge() that is used in the original doesn't quite match
 * our needs.  If we ever hook into the memory management of the SW stack we
 * may be able to use tcp_v4_destroy_sock() directly.
 */
static int chelsio_destroy_sock(struct sock *sk)
{
    t1_purge_write_queue(sk);
    return tcp_prot.destroy(sk);
}

// IP socket options we do not support on offloaded connections
#define UNSUP_IP_SOCK_OPT ((1 << IP_OPTIONS))

/*
 * Socket option code for IP.  We do not allow certain options while a
 * connection is offloaded.  Some of the other options we handle specially,
 * and the rest are directed to the SW IP for their usual processing.
 */
static int chelsio_ip_setsockopt(struct sock *sk, int level, int optname,
				 char __user *optval, int optlen)
{
    if (level != SOL_IP) return -ENOPROTOOPT;

    // unsupported options
    if ((1 << optname) & UNSUP_IP_SOCK_OPT) {
	printk(KERN_WARNING
	       "IP option %d ignored on offloaded TCP connection\n", optname);
	return -ENOPROTOOPT;
    }

    // specially handled options
    if (optname == IP_TOS) {
	struct inet_opt *inet = inet_sk(sk);
	int val = 0, err = 0;

	if (optlen >= sizeof(int)) {
	    if (get_user(val, (int __user *) optval))
		return -EFAULT;
	} else if (optlen >= sizeof(char)) {
	    unsigned char ucval;

	    if (get_user(ucval, (unsigned char __user *) optval))
		return -EFAULT;
	    val = (int) ucval;
	}

	lock_sock(sk);

	val &= ~3;
	val |= inet->tos & 3;
	if (IPTOS_PREC(val) >= IPTOS_PREC_CRITIC_ECP && 
	    !capable(CAP_NET_ADMIN))
	    err = -EPERM;
	else if (inet->tos != val) {
	    inet->tos = val;
	    sk->sk_priority = rt_tos2priority(val);
	    t1_set_tos(sk);
	}

	release_sock(sk);
        return err;
    }

    return tcp_sk(sk)->af_specific->setsockopt(sk, level, optname, optval,
					       optlen);
}

/*
 * Socket option code for TCP.  We override any option processing that needs to
 * be handled specially for a TOE and leave the other options to SW TCP.
 */
static int chelsio_setsockopt(struct sock *sk, int level, int optname,
			      char __user *optval, int optlen)
{
    struct tcp_opt *tp;
    int val, err = 0;

    if (level != SOL_TCP)
	return chelsio_ip_setsockopt(sk, level, optname, optval, optlen);

    if (optlen < sizeof(int))
	return -EINVAL;

    if (get_user(val, (int __user *) optval))
	return -EFAULT;

    lock_sock(sk);
    tp = tcp_sk(sk);

    switch (optname) {
    case TCP_NODELAY: {
	int oldval = tp->nonagle;

	if (val) {
	    tp->nonagle |= TCP_NAGLE_OFF;
	} else {
	    tp->nonagle &= ~TCP_NAGLE_OFF;
	}
	if (oldval != tp->nonagle)
	    t1_set_nagle(sk);
	break;
    }

    case TCP_CORK:
	/*
	 * When set indicates to always queue non-full frames.
	 * Later the user clears this option and we transmit
	 * any pending partial frames in the queue.  This is
	 * meant to be used alongside sendfile() to get properly
	 * filled frames when the user (for example) must write
	 * out headers with a write() call first and then use
	 * sendfile to send out the data parts.
	 *
	 * TCP_CORK can be set together with TCP_NODELAY and it is
	 * stronger than TCP_NODELAY.
	 */
	if (val)
	    tp->nonagle |= TCP_NAGLE_CORK;
	else
	    tcp_uncork(sk);
	break;

    case TCP_KEEPIDLE:
	if (val < 1 || val > MAX_TCP_KEEPIDLE)
	    err = -EINVAL;
	else {
	    tp->keepalive_time = val * HZ;
	}
	break;

    case TCP_QUICKACK:
	if (!val) {
	    tp->ack.pingpong = 1;
	} else {
	    tp->ack.pingpong = 0;
	}
	break;

    case TCP_ULP_MODE: {
	int mode = val & 0xff, submode = (val >> 8) & 0xff;
	int pg_size = (val >> 16) & 0xff;

	if (mode > ULP_MODE_IWARP || submode > 0xf ||
	    (pg_size > 2 && pg_size != 12))
	    err = -EINVAL;
	else {
	    /*
	     * Changing ULP modes automatically uncorks a socket lest we are
	     * left with non-full packets on the send queue that can never be
	     * pushed out.
	     */
	    if (ULP_MODE(tp) != mode)
		tcp_uncork(sk);

	    ULP_MODE(tp) = mode;
	    ULP_SUBMODE(tp) = submode;
	    if (val & HARD_ULP_MODE_CHG)
		t1_set_ulp_mode(sk, pg_size);
	}
	break;
    }

    default:
	release_sock(sk);
	err = tcp_setsockopt(sk, level, optname, optval, optlen);
	goto out;
    }
    release_sock(sk);
out:
    return err;
}

static void set_keepalive(struct sock *sk, int on_off)
{
    int old = sock_flag(sk, SOCK_KEEPOPEN) != 0;

    if (on_off ^ old)
	t1_set_keepalive(sk, on_off);
}

/*
 * Fill a tcpdiag_offload structure TOE specific information.
 */
static void tcpdiag_offload_fill_info(const struct sock *sk,
				      struct tcpdiag_offload *oinfo)
{
    const struct tcp_opt *tp = tcp_sk(sk);

    oinfo->offload_cookie = TID(tp);
    oinfo->offload_dev_idx = TOE_DEV(tp)->name[3] - '0';  // XXX up to 10 TOEs
    oinfo->mem = TX_PAGES_ALLOCED(tp);
}

struct toe_funcs chelsio_t1_toe_funcs = {
    .rcv_consumed = cleanup_rbuf,
    .set_keepalive = set_keepalive,
    .tcpdiag_offload_info = tcpdiag_offload_fill_info,
    .sendskb = chelsio_sendskb,
};

struct proto chelsio_t1_prot;

/*
 * Set up the offload protocol operations vector.  We start with TCP's and
 * override some of the operations.  Note that we do not override the backlog
 * handler here.
 */
void __init t1_init_offload_ops(void)
{
    chelsio_t1_prot = tcp_prot;

    chelsio_t1_prot.close       = chelsio_close;
    chelsio_t1_prot.disconnect  = chelsio_disconnect;
    chelsio_t1_prot.destroy     = chelsio_destroy_sock;
    chelsio_t1_prot.shutdown    = chelsio_shutdown;
    chelsio_t1_prot.setsockopt  = chelsio_setsockopt;
    chelsio_t1_prot.sendmsg     = chelsio_sendmsg;
    chelsio_t1_prot.recvmsg     = chelsio_recvmsg;
    chelsio_t1_prot.sendpage    = chelsio_sendpage;
}
