/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _CHELSIO_DEFS_H
#define _CHELSIO_DEFS_H

#include <linux/skbuff.h>
#include "tom_types.h"
#ifdef CONFIG_CHELSIO_T1_CM
#include "bond.h"
#endif

/* CPL message correctness validation switches */
#define VALIDATE_TID 1
#define VALIDATE_LEN 1
#define VALIDATE_SEQ 1

struct sock;
struct tcp_opt;
struct proc_dir_entry;
struct toedev;
struct tom_data;

/* CPL message priority levels */
enum {
    CPL_PRIORITY_DATA = 0,     /* data messages */
    CPL_PRIORITY_SETUP = 0,    /* connection setup/teardown messages */
    CPL_PRIORITY_LISTEN = 1,   /* listen start/stop messages */
    CPL_PRIORITY_ACK = 1,      /* RX ACK messages */
    CPL_PRIORITY_CONTROL = 1   /* TOE control messages */
};

/* Flags for return value of CPL message handlers */
enum {
    CPL_RET_BUF_DONE = 1,   /* buffer processing done, buffer may be freed */
    CPL_RET_BAD_MSG = 2,    /* bad CPL message (e.g., unknown opcode) */
    CPL_RET_UNKNOWN_TID = 4 /* unexpected unknown TID */
};

/*
 * These flags track some close related events.
 *
 * - ABORT_SHUTDOWN: cannot send an abort request
 * - ABORT_RPL_PENDING: an abort reply is expected
 * - CLOSE_CON_REQUESTED: we have sent a close_connection_request
 */
#define ABORT_SHUTDOWN      0x1
#define ABORT_RPL_PENDING   0x2
#define CLOSE_CON_REQUESTED 0x4
/* wait for tx to drain, for failover */
#define TX_HALT             0x8

/*
 * These extend the TCP socket options in tcp.h.
 *
 * TCP_ULP_MODE - set the ULP mode and submode
 */
#define TCP_ULP_MODE 32

/* The ULP mode/submode of an skbuff */
#define skb_ulp_mode(skb)  (TCP_SKB_CB(skb)->sacked)

/*
 * Requests a hard ULP mode change (i.e., updates HW) if supplied in the
 * TCP_ULP_MODE socket option.
 */
#define HARD_ULP_MODE_CHG 0x1000000

/*
 * We use otherwise unused socket fields to store offloaded state.  These
 * are the accessor functions that hide this "reinterpretation" of the fields.
 * Please put all such fields here so they are in one place.
 */

/* We store the TID in pred_flags */
#define TID(tcp_opt) ((tcp_opt)->pred_flags)

/* Maximum send sequence number to stay within our TX page allotment. */
#define SND_MAX(tcp_opt) ((tcp_opt)->snd_sml)

/* Store the TOE device in accept_queue */
#define TOE_DEV(tcp_opt) (*(struct toedev **)&(tcp_opt)->accept_queue)

/* Store a pointer to the L2T entry in listen_opt. */
#define L2T_ENTRY(tcp_opt) (*(struct l2t_entry **)&(tcp_opt)->listen_opt)

/* Store the number of TX pages allocated to the connection in packets_out */
#define TX_PAGES_ALLOCED(tcp_opt) ((tcp_opt)->packets_out)

/* Store close flags in ca_state */
#define CLOSE_FLAGS(tcp_opt) ((tcp_opt)->ca_state)

/* Store the ULP mode and submode in dsack and eff_sacks, respectively */
#define ULP_MODE(tcp_opt)    ((tcp_opt)->dsack)
#define ULP_SUBMODE(tcp_opt) ((tcp_opt)->eff_sacks)
#define ULP_DATACNT(tcp_opt) ((tcp_opt)->frto_counter)
#define ULP_DATASKB(tcp_opt) (*(struct sk_buff **)&(tcp_opt)->accept_queue_tail)

#ifdef CONFIG_CHELSIO_T1_CM
/* reuse for cml data structure, assume it is not a ULP conn. */
#define CML(tcp_opt) ((Cml_Info_T *)(tcp_opt)->accept_queue_tail)
#endif


typedef void (*defer_handler_t)(struct toedev *dev, struct sk_buff *skb);

/*
 * Stores information used to send deferred CPL replies from process context.
 */
struct deferred_skb_cb {
    defer_handler_t handler;
    struct toedev *dev;
};

#define DEFERRED_SKB_CB(skb) ((struct deferred_skb_cb *)(skb)->cb)

/*
 * Returns a pointer to the first byte of the CPL header in an sk_buff that
 * contains a CPL message.
 */
static inline void *cplhdr(struct sk_buff *skb) 
{
    return skb->data;
}

/*
 * Allocate an sk_buff when allocation failure is not an option.
 */
static inline struct sk_buff *alloc_skb_nofail(unsigned int len)
{
#ifdef __GFP_NOFAIL
    return alloc_skb(len, GFP_KERNEL | __GFP_NOFAIL);
#else
    for (;;) {
	struct sk_buff *skb = alloc_skb(len, GFP_KERNEL);
	if (skb) return skb;
	yield();
    }
#endif
}

void t1_register_cpl_handler(unsigned int opcode, cpl_handler_t h);
void t1_listen_start(struct toedev *dev, struct sock *sk);
void t1_listen_stop(struct toedev *dev, struct sock *sk);
int t1_push_frames(struct sock *sk);
int t1_alloc_tx_pages(struct sock *sk, unsigned int n);
void t1_purge_write_queue(struct sock *sk);
void t1_set_nagle(struct sock *sk);
void t1_set_tos(struct sock *sk);
void t1_set_keepalive(struct sock *sk, int on_off);
void t1_set_ulp_mode(struct sock *sk, int page_size);
void t1_write_space(struct sock *sk);
int t1_send_reset(struct sock *sk, int mode, struct sk_buff *skb);
int t1_connect(struct toedev *dev, struct sock *sk);
void t1_disconnect_acceptq(struct sock *sk);
u32 t1_send_rx_credits(struct toedev *dev, tid_t tid, u32 credits, int nofail);
int t1_listen_proc_setup(struct proc_dir_entry *dir, struct tom_data *d);
void t1_listen_proc_free(struct proc_dir_entry *dir);
void t1_defer_reply(struct sk_buff *skb, struct toedev *dev,
		    defer_handler_t handler);

// initialization
void t1_init_offload_ops(void);
void t1_init_listen_cpl_handlers(void);
void t1_init_route_cpl_handlers(void);
int t1_init_cpl_io(void);
#endif

// Local Variables:
// c-basic-offset: 4
// End:
