/*
 * Chelsio TOE layer 2 table management.
 *
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/toedev.h>
#include <net/neighbour.h>
#include "defs.h"
#include "l2t.h"
#include "../cpl5_cmd.h"
#include "cpl5_defs.h"
#include "tcam_rte.h"

/*
 * Note: We do not take refereces to net_devices in this module because both 
 * the TOE and the sockets already hold references to the interfaces and the
 * lifetime of an L2T entry is fully contained in the lifetime of the TOE.
 */

/*
 * Code taken from arp.c.
 */
static inline unsigned int arp_hash(u32 key, const struct net_device *dev,
				    int htab_size)
{
    unsigned int hash_val;

    hash_val = key;
    hash_val ^= hash_val >> 16;
    hash_val ^= hash_val >> 8;
    hash_val ^= hash_val >> 3;
    hash_val = (hash_val ^ dev->ifindex) & (htab_size - 1);
    return hash_val;
}

static inline void neigh_replace(struct l2t_entry *e, struct neighbour *n)
{
    neigh_hold(n);
    if (e->neigh) neigh_release(e->neigh);
    e->neigh = n;
}

/*
 * L2T entry is up to date, just send the packet.
 */
static int fast_send(struct toedev *dev, struct sk_buff *skb,
		     struct l2t_entry *e)
{
    skb->real_dev = e->real_dev;
    return toe_send(dev, skb);
}

/*
 * Set up an L2T entry and send any packets waiting in the arp queue.  The
 * supplied skb is used for the CPL_L2T_WRITE_REQ.  Must be called with the
 * entry locked.
 */
static int setup_l2e_send_pending(struct toedev *dev, struct sk_buff *skb,
				  struct l2t_entry *e)
{
    struct cpl_l2t_write_req *req;
    int idx = l2e_idx(L2DATA(dev), e);

    if (!skb) {
	skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
	if (!skb) return -1;
    }

    req = (struct cpl_l2t_write_req *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_L2T_WRITE_REQ, idx));
    req->params = htonl(V_L2TOP_IDX(idx) | V_L2TOP_VLAN(e->vlan) |
			V_L2TOP_IFF(e->real_dev->if_port));
    memcpy(req->dst_mac, e->neigh->ha, sizeof(req->dst_mac));
    memcpy(e->dmac, req->dst_mac, sizeof(e->dmac));
    skb->priority = CPL_PRIORITY_CONTROL;

    skb->real_dev = e->real_dev;
    toe_send(dev, skb);

    while (e->arpq_head) {
	skb = e->arpq_head;
	e->arpq_head = skb->next;
	skb->next = NULL;
    	skb->real_dev = e->real_dev;
	toe_send(dev, skb);
    }
    e->arpq_tail = NULL;
    e->send = fast_send;

    {
        /* Update the TCAM route */
        struct tcam_rte_entry tcam_rt;
	memset(&tcam_rt, 0, sizeof(tcam_rt));
	tcam_rt.rte_faddr = e->addr;
	tcam_rt.rte_netmask = TCAM_NM32(32);
	tcam_rt.rte_nm_len = 32;
	tcam_rt.rte_l2t_ix = idx;
	tcam_rte_insert(dev, &tcam_rt);
    }

    return 0;
}

/*
 * L2T entry is stale, kick off revalidation.
 */
static int validate_send(struct toedev *dev, struct sk_buff *skb,
			 struct l2t_entry *e)
{
    neigh_event_send(e->neigh, NULL);
    spin_lock_bh(&e->lock);
    if (e->send == validate_send) 
	e->send = fast_send;
    spin_unlock_bh(&e->lock);

    skb->real_dev = e->real_dev;
    return toe_send(dev, skb);
}

/*
 * L2T entry not yet programmed, queue packet and start the process of setting
 * up the L2T entry.
 */
static int resolve_send(struct toedev *dev, struct sk_buff *skb,
			struct l2t_entry *e)
{
    spin_lock_bh(&e->lock);
    if (e->send != &resolve_send) {       /* ARP already completed */
	spin_unlock_bh(&e->lock);
	return e->send(dev, skb, e);
    }
    skb->next = NULL;
    if (e->arpq_head)
	e->arpq_tail->next = skb;
    else
	e->arpq_head = skb;
    e->arpq_tail = skb;
    spin_unlock_bh(&e->lock);

    if (!neigh_event_send(e->neigh, NULL)) {
	skb = alloc_skb(sizeof(struct cpl_l2t_write_req), GFP_ATOMIC);
	if (!skb) goto out;                         /* XXX */

	spin_lock_bh(&e->lock);
	if (e->arpq_head)
	    setup_l2e_send_pending(dev, skb, e);
	else                                      /* we lost the race */
	    kfree_skb(skb);
	spin_unlock_bh(&e->lock);
    }
 out:
    return 0;
}
	
/*
 * Allocate a free l2 entry.  Must be called with l2t_data.lock held.
 */
static struct l2t_entry *alloc_l2e(struct l2t_data *d)
{
    struct l2t_entry *e = NULL;

    if (atomic_read(&d->nfree)) {  /* there's definitely a free entry */
	struct l2t_entry *end = &d->l2tab[d->nentries];

	for (e = d->rover; e != end; ++e)
	    if (atomic_read(&e->refcnt) == 0) goto found;
	for (e = &d->l2tab[1]; atomic_read(&e->refcnt); ++e)
	    ;
    found:
	d->rover = e + 1;
	atomic_dec(&d->nfree);

	/*
	 * The entry we found may be an inactive entry that is presently in
	 * the hash table.  We need to remove it.
	 */
	if (e->dev) {
	    int hash = arp_hash(e->addr, e->dev, d->nentries);
	    struct l2t_entry **p = &d->l2tab[hash].first;

	    for (p = &d->l2tab[hash].first; *p; p = &(*p)->next)
		if (*p == e) {
		    *p = e->next;
		    break;
		}
	    e->dev = e->real_dev = NULL;
	}
    }
    return e;
}

/*
 * Called when an L2T entry has no more users.  The entry is left in the hash
 * table since it is likely to be reused but we also bump nfree to indicate
 * that the entry can be reallocated for a different neighbor.  We also drop
 * the existing neighbor reference in case the neighbor is going away and is
 * waiting on our reference.
 */
void l2e_free(struct l2t_data *d, struct l2t_entry *e)
{
    if (e->neigh) {
	neigh_release(e->neigh);
	e->neigh = NULL;
    }
    if (e->send == fast_send) 
	e->send = validate_send;
    atomic_inc(&d->nfree);
    /* XXX what do we do with the arpq? */
}

/*
 * Lookup an entry in the hash table.  Must be called with the hash
 * table lock held.
 */
static inline struct l2t_entry *__lookup(const struct l2t_data *d,
					 const struct neighbour *n)
{
    struct l2t_entry *e;
    u32 addr = *(u32 *)n->primary_key;
    int hash = arp_hash(addr, n->dev, d->nentries);

    for (e = d->l2tab[hash].first; e; e = e->next)
	if (e->addr == addr && e->dev == n->dev) {
	    l2t_hold(e);
	    break;
	}
    return e;
}

#if 0
static struct l2t_entry *lookup(struct l2t_data *d, const struct neighbour *n)
{
    struct l2t_entry *e;

    read_lock_bh(&d->lock);
    e = __lookup(d, n);
    read_unlock_bh(&d->lock);
    return e;
}
#endif

struct l2t_entry *l2t_get(struct toedev *dev, struct neighbour *neigh)
{
    u32 addr;
    int hash;
    struct l2t_data *d;
    struct l2t_entry *e;

    if (!neigh) return NULL;

    d = L2DATA(dev);
    addr = *(u32 *)neigh->primary_key;
    hash = arp_hash(addr, neigh->dev, d->nentries);

    write_lock_bh(&d->lock);
    for (e = d->l2tab[hash].first; e; e = e->next)
	if (e->addr == addr && e->dev == neigh->dev) {
	    l2t_hold(e);
	    goto done;
	}

    /* Need to allocate a new entry */
    e = alloc_l2e(d);
    if (e) {
        /* Clean up routing tcam entry if necessary */
	if (e->addr) {
            struct tcam_rte_entry tcam_rt;
	    int idx;

	    memset(&tcam_rt, 0, sizeof(tcam_rt));
	    tcam_rt.rte_faddr = e->addr;
	    tcam_rt.rte_netmask = TCAM_NM32(32);
	    tcam_rt.rte_nm_len = 32;
	    idx = l2e_idx(L2DATA(dev), e);
	    tcam_rt.rte_l2t_ix = idx;
	    tcam_rte_delete(dev, &tcam_rt);
	}

	e->next = d->l2tab[hash].first;
	d->l2tab[hash].first = e;
	e->addr = addr;
	e->dev = e->real_dev = neigh->dev;
	atomic_set(&e->refcnt, 1);
	e->send = resolve_send;
	e->vlan = dev->ttid == TOE_ID_CHELSIO_T1 ? 0 : 0xFFF;
    }
 done:
    write_unlock_bh(&d->lock);

    if (e) {
	spin_lock_bh(&e->lock);
	/* Cache most recent neighbor with given (IP, dev) */
	if (neigh != e->neigh)
	    neigh_replace(e, neigh);
	spin_unlock_bh(&e->lock);
    }
    return e;
}

static void handle_failed_resolution(struct toedev *dev, struct l2t_entry *e)
{
    while (e->arpq_head) {
	struct sk_buff *skb = e->arpq_head;
	struct l2t_skb_cb *cb = L2T_SKB_CB(skb);

	e->arpq_head = skb->next;
	skb->next = NULL;
	if (cb->arp_failure_handler)
	    cb->arp_failure_handler(dev, skb);
	else
	    toe_send(dev, skb);
    }
    e->arpq_tail = NULL;
}

void l2t_update(struct toedev *dev, struct neighbour *neigh, int flags, struct net_device *real_dev)
{
    struct l2t_data *d = L2DATA(dev);
    struct l2t_entry *e;

    read_lock_bh(&d->lock);
    e = __lookup(d, neigh);
    read_unlock_bh(&d->lock);

    if (!e) return;

    spin_lock_bh(&e->lock);
    if (neigh != e->neigh)
	neigh_replace(e, neigh);

    if (e->real_dev != real_dev) {
        e->send = resolve_send;
        e->real_dev = real_dev;
    }

    if (e->send == resolve_send) {
	if (neigh->nud_state & NUD_FAILED)
	    handle_failed_resolution(dev, e);
	else if (neigh_is_connected(neigh) && e->arpq_head)
	    setup_l2e_send_pending(dev, NULL, e);
    } else {
	e->send = neigh_is_connected(neigh) ? fast_send : validate_send;
	if (flags & NEIGH_ADDR_CHANGED)
	    setup_l2e_send_pending(dev, NULL, e);
    }
    spin_unlock_bh(&e->lock);
    l2t_release(d, e);
}
	
struct l2t_data *init_l2t(int l2t_capacity)
{
    struct l2t_data *d;
    int i, size = sizeof(*d) + l2t_capacity * sizeof(struct l2t_entry);

    d = kmalloc(size, GFP_KERNEL);
    if (!d) goto out;

    memset(d, 0, size);
    d->nentries = l2t_capacity;
    d->rover = &d->l2tab[1];                  /* entry 0 is not used */
    atomic_set(&d->nfree, l2t_capacity - 1);
    d->lock = RW_LOCK_UNLOCKED;

    for (i = 0; i < l2t_capacity; ++i) {
	d->l2tab[i].lock = SPIN_LOCK_UNLOCKED;
	atomic_set(&d->l2tab[i].refcnt, 0);
    }

 out:
    return d;
}

void free_l2t(struct l2t_data *d)
{
    kfree(d);
}

#ifdef CONFIG_PROC_FS
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include "../toe_compat.h"

static inline void *l2t_get_idx(struct seq_file *seq, loff_t pos)
{
    struct l2t_data *d = seq->private;

    return pos >= d->nentries ? NULL : &d->l2tab[pos];
}

static void *l2t_seq_start(struct seq_file *seq, loff_t *pos)
{
    return *pos ? l2t_get_idx(seq, *pos) : (void *)1;
}

static void *l2t_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
    if (v == (void *)1) {
	v = l2t_get_idx(seq, 1);
    } else {
	struct l2t_entry *e = v;
	struct l2t_data *d = seq->private;

        ++e;
	v = e < &d->l2tab[d->nentries] ? e : NULL;
    }
    ++*pos;
    return v;
}

static void l2t_seq_stop(struct seq_file *seq, void *v)
{
}

static char l2e_state(const struct l2t_entry *e)
{
    if (!e->send) return 'U';                 // unallocated entry
    if (e->send == fast_send) return 'F';     // fast path entry
    if (e->send == validate_send) return 'V'; // validating entry
    return e->arpq_head ? 'A' : 'R';          // resolving/will resolve entry
}

static int l2t_seq_show(struct seq_file *seq, void *v)
{
    if (v == (void *)1)
	seq_puts(seq, "Index IP address      Ethernet address   VLAN  State   Users Iff\n");
    else {
	char ip[20];
	struct l2t_entry *e = v;
	struct l2t_data *d = seq->private;

	sprintf(ip, "%u.%u.%u.%u", NIPQUAD(e->addr));
	seq_printf(seq,
		   "%-5d %-15s %02x:%02x:%02x:%02x:%02x:%02x  %4d    %c   %7u %s,%s\n",
		   l2e_idx(d, e), ip, e->dmac[0], e->dmac[1], e->dmac[2],
		   e->dmac[3], e->dmac[4], e->dmac[5], e->vlan, l2e_state(e),
		   atomic_read(&e->refcnt), e->dev ? e->dev->name : "?",
		   e->real_dev ? e->real_dev->name : "?");
    }
    return 0;
}

static struct seq_operations l2t_seq_ops = {
    .start  = l2t_seq_start,
    .next   = l2t_seq_next,
    .stop   = l2t_seq_stop,
    .show   = l2t_seq_show
};

static int l2t_seq_open(struct inode *inode, struct file *file)
{
    int rc = seq_open(file, &l2t_seq_ops);

    if (!rc) {
	struct proc_dir_entry *dp = PDE(inode);
	struct seq_file *seq = file->private_data;

	seq->private = dp->data;
    }
    return rc;
}

static struct file_operations l2t_seq_fops = {
    .owner   = THIS_MODULE,
    .open    = l2t_seq_open,
    .read    = seq_read,
    .llseek  = seq_lseek,
    .release = seq_release,
};

/*
 * Create the proc entries for the L2 table under dir.
 */
int l2t_proc_setup(struct proc_dir_entry *dir, struct l2t_data *d)
{
    struct proc_dir_entry *p;

    if (!dir) return -EINVAL;

    p = create_proc_entry("l2t", S_IRUGO, dir);
    if (!p) return -ENOMEM;

    p->proc_fops = &l2t_seq_fops;
    p->data = d;
    return 0;
}

void l2t_proc_free(struct proc_dir_entry *dir)
{
    if (dir) {
	remove_proc_entry("l2t", dir);
    }
}
#endif

// Local Variables:
// c-basic-offset: 4
// End:
