/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _CHELSIO_L2T_H
#define _CHELSIO_L2T_H

#include <linux/config.h>
#include <linux/spinlock.h>
#include <asm/atomic.h>

struct net_devive;
struct neighbour;
struct toedev;
struct sk_buff;

struct l2t_entry {
    u32 addr;                         /* dest IP */
    struct net_device *dev;           /* egress interface */
    struct net_device *real_dev;      /* egress interface physical */
    struct neighbour *neigh;          /* associated neighbour */
    struct l2t_entry *next;           /* next l2t_entry on chain */
    struct l2t_entry *first;          /* start of hash chain */
    struct sk_buff *arpq_head;        /* holds sk_buffs awaiting resolution */
    struct sk_buff *arpq_tail;
    int (*send)(struct toedev *dev, struct sk_buff *skb, struct l2t_entry *e);
    spinlock_t lock;
    atomic_t refcnt;                  /* entry reference count */
    u16 vlan;                         /* vlan id */
    u8 dmac[6];                       /* neighbour's MAC address */
};

struct l2t_data {
    int nentries;              /* number of entries */
    struct l2t_entry *rover;   /* starting point for next allocation */
    atomic_t nfree;            /* number of free entries */
    rwlock_t lock;
    struct l2t_entry l2tab[0];
};

/*
 * Callback stored in an skb to handle address resolution failure.
 */
struct l2t_skb_cb {
    void (*arp_failure_handler)(struct toedev *dev, struct sk_buff *skb);
};

#define L2T_SKB_CB(skb) ((struct l2t_skb_cb *)(skb)->cb)
#define SET_ARP_FAILURE_HANDLER(skb, hnd) L2T_SKB_CB(skb)->arp_failure_handler = (hnd)

/*
 * Getting to the L2 data from a toe device.
 */
#define L2DATA(dev) ((dev)->l2opt)

extern void l2e_free(struct l2t_data *d, struct l2t_entry *e);
extern void l2t_update(struct toedev *dev, struct neighbour *neigh, int flags, struct net_device *ldev);
extern struct l2t_entry *l2t_get(struct toedev *dev, struct neighbour *neigh);
extern struct l2t_data *init_l2t(int l2t_capacity);
extern void free_l2t(struct l2t_data *d);

#ifdef CONFIG_PROC_FS
extern int l2t_proc_setup(struct proc_dir_entry *dir, struct l2t_data *d);
extern void l2t_proc_free(struct proc_dir_entry *dir);
#else
#define l2t_proc_setup(dir, d) 0
#define l2t_proc_free(dir)
#endif

/*
 * Returns the L2 table index of an L2T entry.
 */
static inline int l2e_idx(struct l2t_data *d, struct l2t_entry *e)
{
    return e - d->l2tab;
}

static inline int l2t_send(struct toedev *dev, struct sk_buff *skb,
			   struct l2t_entry *e)
{
    return e->send(dev, skb, e);
}

static inline void l2t_hold(struct l2t_entry *e)
{
    atomic_inc(&e->refcnt);
}

static inline void l2t_release(struct l2t_data *d, struct l2t_entry *e)
{
    if (atomic_dec_and_test(&e->refcnt))
	atomic_inc(&d->nfree);
}
#endif

// Local Variables:
// c-basic-offset: 4
// End:
