/*
 * This file handles offloading of listening sockets.
 *
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <net/tcp.h>
#include <linux/toedev.h>
#include <net/offload.h>
#include "../toe_compat.h"
#include "defs.h"
#include "tom.h"
#include "../cpl5_cmd.h"
#include "cpl5_defs.h"

/*
 * Returns a new server TID or -1 if all TIDs are taken.  If sk is non-NULL
 * the newly allocated stid_tab entry is set to point to this socket and the
 * reference count of the socket is incremented.  Note that any lookups for an
 * STID will fail until the table entry becomes non-NULL.
 */
static int alloc_stid(struct tid_info *t, struct sock *sk)
{
    int stid = -1;

    spin_lock_bh(&t->stid_lock);
    if (t->sfree) {
	union listen_entry *p = t->sfree;

	stid = (p - t->stid_tab) + t->stid_base;
	t->sfree = p->next;
	if (sk) sock_hold(sk);
	p->sk = sk;
	t->stids_in_use++;
    }
    spin_unlock_bh(&t->stid_lock);
    return stid;
}

/*
 * Free a server TID and return it to the free pool.  If the stid_tab entry is
 * non-NULL we drop the socket reference.
 */
static void free_stid(struct tid_info *t, tid_t stid)
{
    union listen_entry *p = &STID2ENTRY(t, stid);

    if (p->sk) sock_put(p->sk);
    spin_lock_bh(&t->stid_lock);
    p->next = t->sfree;
    t->sfree = p;
    t->stids_in_use--;
    spin_unlock_bh(&t->stid_lock);
}

static inline int listen_hashfn(const struct sock *sk)
{
    return ((unsigned long)sk >> 10) & (LISTEN_INFO_HASH_SIZE - 1);
}

/*
 * Create and add a listen_info entry to the listen hash table.  This and the
 * listen hash table functions below cannot be called from softirqs.
 */
static struct listen_info *listen_hash_add(struct tom_data *d, struct sock *sk,
					   tid_t stid)
{
    struct listen_info *p = kmalloc(sizeof(struct listen_info), GFP_KERNEL);

    if (p) {
	int bucket = listen_hashfn(sk);

	p->sk = sk;      /* this is just a key, no need to take a reference */
	p->stid = stid;
	write_lock(&d->listen_lock);
	p->next = d->listen_hash_tab[bucket];
	d->listen_hash_tab[bucket] = p;
	write_unlock(&d->listen_lock);
    }
    return p;
}

#if 0
/*
 * Given a pointer to a listening socket return its server TID by consulting
 * the socket->stid map.  Returns -1 if the socket is not in the map.
 */
static int listen_hash_find(struct tom_data *d, struct sock *sk)
{
    int stid = -1, bucket = listen_hashfn(sk);
    struct listen_info *p;

    read_lock(&d->listen_lock);
    for (p = d->listen_hash_tab[bucket]; p; p = p->next)
	if (p->sk == sk) {
	    stid = p->stid;
	    break;
	}
    read_unlock(&d->listen_lock);
    return stid;
}
#endif

/*
 * Delete the listen_info structure for a listening socket.  Returns the server
 * TID for the socket if it is present in the socket->stid map, or -1.
 */
static int listen_hash_del(struct tom_data *d, struct sock *sk)
{
    int stid = -1, bucket = listen_hashfn(sk);
    struct listen_info *p, **prev = &d->listen_hash_tab[bucket];

    write_lock(&d->listen_lock);
    for (p = *prev; p; prev = &p->next, p = p->next)
	if (p->sk == sk) {
	    stid = p->stid;
	    *prev = p->next;
	    kfree(p);
	    break;
	}
    write_unlock(&d->listen_lock);
    return stid;
}

/*
 * Start a listening server by sending a passive open request to HW.
 */
void t1_listen_start(struct toedev *dev, struct sock *sk)
{
    int stid;
    struct sk_buff *skb;
    struct cpl_pass_open_req *req;
    struct tom_data *d = TOM_DATA(dev);

    stid = alloc_stid(&d->tid_maps, NULL);
    if (stid < 0) return;

    skb = alloc_skb(sizeof(*req), GFP_KERNEL);
    if (!skb) goto free_stid;

    if (!listen_hash_add(d, sk, stid)) goto free_buf;

    sock_hold(sk);    /* XXX can be merged with alloc_stid */
    STID2ENTRY(&d->tid_maps, stid).sk = sk;

    req = (struct cpl_pass_open_req *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_PASS_OPEN_REQ, stid));
    req->local_port = inet_addr_info(sk)->sport;
    req->peer_port = 0;
    req->local_ip = inet_addr_info(sk)->rcv_saddr;
    req->peer_ip = 0;
    req->peer_netmask = 0;

    if (TOM_TUNABLE(dev, toe_automode_on)) {
        req->opt0h = htonl(V_DELACK);
	req->opt0l = htonl(V_RCV_BUFSIZ(63));
        req->opt1 = htonl(V_CONN_POLICY(CPL_CONN_POLICY_AUTO));
    } else {
        req->opt0h = htonl(V_DELACK | V_TCAM_BYPASS);
	req->opt0l = htonl(V_RCV_BUFSIZ(16));  // overriden in PASS_ACCEPT_RPL
        req->opt1 = htonl(V_CONN_POLICY(CPL_CONN_POLICY_ASK));
    }

    toe_send(dev, skb);
    return;

 free_buf:
    __kfree_skb(skb);
 free_stid:
    free_stid(&d->tid_maps, stid);
}

/*
 * Stop a listening server by sending a close_listsvr request to HW.
 * The server TID is freed when we get the reply.
 */
void t1_listen_stop(struct toedev *dev, struct sock *sk)
{
    struct sk_buff *skb;
    struct cpl_close_listserv_req *req;

    int stid = listen_hash_del(TOM_DATA(dev), sk);
    if (stid < 0) return;

    skb = alloc_skb_nofail(sizeof(*req));
    req = (struct cpl_close_listserv_req *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_CLOSE_LISTSRV_REQ, stid));
    toe_send(dev, skb);

    t1_disconnect_acceptq(sk);
}

/*
 * Process a CPL_CLOSE_LISTSRV_RPL message.  If the status is good we release
 * the STID.
 */
static int do_close_server_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct tom_data *d = TOM_DATA(dev);
    struct cpl_close_listserv_rpl *rpl = cplhdr(skb);
    tid_t stid = GET_TID(rpl);

#if VALIDATE_TID
    if (!lookup_stid(&d->tid_maps, stid))
	return CPL_RET_UNKNOWN_TID | CPL_RET_BUF_DONE;
#endif
    /* handle bad status */
    free_stid(&d->tid_maps, stid);
    return CPL_RET_BUF_DONE;
}

/*
 * Process a CPL_PASS_OPEN_RPL message.  Remove the socket from the listen hash
 * table and free the STID if there was any error, otherwise nothing to do.
 */
static int do_pass_open_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_pass_open_rpl *rpl = (struct cpl_pass_open_rpl *)skb->data;

    if (rpl->status != CPL_ERR_NONE) {
	int stid = GET_TID(rpl);
	struct tom_data *d = TOM_DATA(dev);
	struct sock *sk = lookup_stid(&d->tid_maps, stid);

#if VALIDATE_TID
	if (!sk)
	    return CPL_RET_UNKNOWN_TID | CPL_RET_BUF_DONE;
#endif
	/*
	 * Note: It is safe to unconditionally call listen_hash_del() at this
	 * point without risking unhashing a reincarnation of an already closed
	 * socket (i.e., there is no listen, close, listen, free the sock for
	 * the second listen while processing a message for the first race)
	 * because we are still holding a reference on the socket.  It is
	 * possible that the unhash will fail because the socket is already
	 * closed, but we cannot unhash the wrong socket because it is
	 * impossible for the socket to which this message refers to have
	 * reincarnated.
	 */
	listen_hash_del(d, sk);
	free_stid(&d->tid_maps, stid);  /* this drops the socket reference */
    }
    return CPL_RET_BUF_DONE;
}

void __init t1_init_listen_cpl_handlers(void)
{
    t1_register_cpl_handler(CPL_PASS_OPEN_RPL, do_pass_open_rpl);
    t1_register_cpl_handler(CPL_CLOSE_LISTSRV_RPL, do_close_server_rpl);
}

#ifdef CONFIG_PROC_FS
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#define PROFILE_LISTEN_HASH 1

#if PROFILE_LISTEN_HASH
# define BUCKET_FIELD_NAME "  Bucket"
# define BUCKET_FMT " %d"
# define BUCKET(sk) , listen_hashfn(sk)
#else
# define BUCKET_FIELD_NAME
# define BUCKET_FMT
# define BUCKET(sk)
#endif

/*
 * Return the first entry in the listen hash table that's in
 * a bucket >= start_bucket.
 */
static void *listen_get_first(struct seq_file *seq, int start_bucket)
{
    struct tom_data *d = seq->private;

    for ( ; start_bucket < LISTEN_INFO_HASH_SIZE; ++start_bucket)
	if (d->listen_hash_tab[start_bucket])
	    return d->listen_hash_tab[start_bucket];
    return NULL;
}

static void *listen_get_next(struct seq_file *seq, void *cur)
{
    struct listen_info *p = cur;

    return p->next ? p->next : listen_get_first(seq, listen_hashfn(p->sk) + 1);
}

/*
 * Must be called with the listen_lock held.
 */
static void *listen_get_idx(struct seq_file *seq, loff_t pos)
{
    void *p = listen_get_first(seq, 0);

    if (p)
	while (pos && (p = listen_get_next(seq, p)))
	    pos--;

    return pos ? NULL : p;
}

static void *listen_get_idx_lock(struct seq_file *seq, loff_t pos)
{
    struct tom_data *d = seq->private;

    read_lock(&d->listen_lock);
    return listen_get_idx(seq, pos);
}

static void *listen_seq_start(struct seq_file *seq, loff_t *pos)
{
    return *pos ? listen_get_idx_lock(seq, *pos - 1) : SEQ_START_TOKEN;
}

static void *listen_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
    if (v == SEQ_START_TOKEN)
	v = listen_get_idx_lock(seq, 0);
    else
	v = listen_get_next(seq, v);
    ++*pos;
    return v;
}

static void listen_seq_stop(struct seq_file *seq, void *v)
{
    if (v != SEQ_START_TOKEN)
	read_unlock(&((struct tom_data *)seq->private)->listen_lock);
}

static int listen_seq_show(struct seq_file *seq, void *v)
{
    if (v == SEQ_START_TOKEN)
	seq_puts(seq, "TID     IP address      Port" BUCKET_FIELD_NAME "\n");
    else {
	char ip[20];
	struct listen_info *p = v;
	struct sock *sk = p->sk;
	u32 saddr = inet_addr_info(sk)->rcv_saddr;

	sprintf(ip, "%u.%u.%u.%u", NIPQUAD(saddr));
	seq_printf(seq, "%-7u %-15s %-5u" BUCKET_FMT "\n", p->stid, ip,
		   inet_addr_info(sk)->num BUCKET(sk));
    }
    return 0;
}

static struct seq_operations listen_seq_ops = {
    .start  = listen_seq_start,
    .next   = listen_seq_next,
    .stop   = listen_seq_stop,
    .show   = listen_seq_show
};

static int listen_seq_open(struct inode *inode, struct file *file)
{
    int rc = seq_open(file, &listen_seq_ops);

    if (!rc) {
	struct proc_dir_entry *dp = PDE(inode);
	struct seq_file *seq = file->private_data;

	seq->private = dp->data;
    }
    return rc;
}

static struct file_operations listen_seq_fops = {
    .owner   = THIS_MODULE,
    .open    = listen_seq_open,
    .read    = seq_read,
    .llseek  = seq_lseek,
    .release = seq_release
};

#define LISTEN_PROC_NAME "listeners"

/*
 * Create the proc entry for the listening servers under dir.
 */
int t1_listen_proc_setup(struct proc_dir_entry *dir, struct tom_data *d)
{
    struct proc_dir_entry *p;

    if (!dir) return -EINVAL;

    p = create_proc_entry(LISTEN_PROC_NAME, S_IRUGO, dir);
    if (!p) return -ENOMEM;

    p->proc_fops = &listen_seq_fops;
    p->data = d;
    return 0;
}

void t1_listen_proc_free(struct proc_dir_entry *dir)
{
    if (dir) {
	remove_proc_entry(LISTEN_PROC_NAME, dir);
    }
}
#endif

// Local Variables:
// c-basic-offset: 4
// End:

