/*
 * Chelsio TOE routing table management.
 *
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Divy Le Ray
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/types.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <net/ip.h>
#include <net/route.h>
#include <net/ip_fib.h>
#include <linux/netlink.h>
#include "tcam_rte.h"
#include "defs.h"
#include "../toe_ctl_defs.h"

#ifdef CONFIG_PROC_FS
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/toedev.h>
#include "../cpl5_cmd.h"
#include "cpl5_defs.h"
#endif /* CONFIG_PROC_FS */

#include "../toe_compat.h"

/*
 * PLO_OPT algorithm related routines.
 */

/* Initializes PLO_OPT regions */
static void tcam_opt_reg_init(struct tcam_rte_unit *t)
{
	struct tcam_opt_region *rg = t->opt_reg;
	int i;

	for (i = REG_FIRST; i <= REG_GAP0; i++) {
		rg[i].rg_lo = t->rte_n_entries - 1;
		rg[i].rg_hi = t->rte_n_entries - 1;
	}

	for (i = REG_GAP0 + 1; i <= REG_LAST; i++) {
		rg[i].rg_lo = 0;
		rg[i].rg_hi = 0;
	}
}

/* Returns TRUE if TCAM route area is full. */
static int tcam_rte_full(struct tcam_rte_unit *t)
{
	return ((atomic_read(&t->tcam_rte_cnt) >= t->rte_n_entries));
}


/* Returns the index of an existing route. */
static int tcam_rte_index(struct toedev *dev, struct tcam_rte_entry *r)
{
	int i, nml, min, max;
	struct tcam_opt_region *rg;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);

	nml = r->rte_nm_len;
	rg = t->opt_reg;

	if (nml > REG_GAP0) {
		min = rg[nml].rg_lo;
		max = rg[nml].rg_hi;
	} else {
		min = rg[nml].rg_lo + 1;
		max = rg[nml].rg_hi + 1;
	}

	for (i = min; i < max; i++) {
		if (r->rte_faddr == t->cache_rte[i].rte_faddr)
			break;
	}

	return ((i < max) ? i : -ESRCH);
}

/* Inserts the given route at the given index. */
static int tcam_rte_write_index(struct toedev *dev,
			      struct tcam_rte_entry *r,
			      int index,
			      int control)
{
	struct cpl_rte_write_req *req;
	struct sk_buff *skb;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
	struct tcam_rte_entry *tab_rt, *gw_rt = NULL;
	u8 l2t_ix;
	int gw_dep_update = 0;
	int w_control;
	u32 tcam_index;

	if (!t) return -1;

	tab_rt = &t->cache_rte[index];

	r->rte_index = index;

	/* Retrieve the gateway for an indirect route */
	if (r->rte_flags & RTE_IND) {
		struct tcam_rte_entry gw_rte;
		int gw_ix;

		/* Retrieve the gateway */
		gw_rte.rte_faddr = r->gw_opt_addr;
		gw_rte.rte_nm_len = r->gw_opt_nm_len;

		gw_ix = tcam_rte_index(dev, &gw_rte);
		if (gw_ix < 0) return -1;

		gw_rt = &t->cache_rte[gw_ix];
	}

	if (control == TCAM_RTE_NEW) {
		memcpy(tab_rt, r, sizeof(*r));

		/* Add a new gateway */
		if (r->rte_flags & RTE_GW) {
			INIT_LIST_HEAD(&tab_rt->gw_dep);
		}

		/* Add a new indirect route */
		if (r->rte_flags & RTE_IND) {
			tab_rt->rte_l2t_ix = gw_rt->rte_l2t_ix;
			INIT_LIST_HEAD(&tab_rt->gw_dep);
			list_add(&tab_rt->gw_dep, &gw_rt->gw_dep);
		}
	}

	if (control == TCAM_RTE_SHIFT) {
		memcpy(tab_rt, r, sizeof(*r));

		/* Shift an existing gateway */
		if (r->rte_flags & RTE_GW) {
			if (!list_empty(&tab_rt->gw_dep)) {
				struct tcam_rte_entry *r_next;
				/* Use the element as the list head */
				r_next = list_entry(tab_rt->gw_dep.next,
						    struct tcam_rte_entry,
						    gw_dep);
				list_del_init(&tab_rt->gw_dep);
				list_add(&tab_rt->gw_dep, &r_next->gw_dep);
			}
		}

		/* Shift an existing indirect route */
		if (r->rte_flags & RTE_IND) {
			list_del_init(&tab_rt->gw_dep);
			list_add(&tab_rt->gw_dep, &gw_rt->gw_dep);
		}
	}

	if (control == TCAM_RTE_UPDATE) {
		/* A new indirect route will share the gateway */
		if ((r->rte_flags & RTE_GW) && (tab_rt->rte_flags & RTE_GW))
			goto out;

		/* The existing entry is made a gateway */
		if ((r->rte_flags & RTE_GW) && !(tab_rt->rte_flags & RTE_GW)) {
			tab_rt->rte_flags |= RTE_GW;
			INIT_LIST_HEAD(&tab_rt->gw_dep);
			goto out;
		}

		/* The gateway's ARP info is modified */
		if (!(r->rte_flags & RTE_GW) && (tab_rt->rte_flags & RTE_GW)) {
			tab_rt->rte_l2t_ix = r->rte_l2t_ix;

			/* We have to update the dependant routes */
			gw_dep_update = 1;
		}

		/* The indirect route is going through a new gateway */
		if ((r->rte_flags & RTE_IND) && (tab_rt->rte_flags & RTE_IND)) {
			tab_rt->rte_l2t_ix = gw_rt->rte_l2t_ix;
			list_del_init(&tab_rt->gw_dep);
			list_add(&tab_rt->gw_dep, &gw_rt->gw_dep);
		}

		/* The existing entry is made an indirect route */
		if ((r->rte_flags & RTE_IND) && !(tab_rt->rte_flags & RTE_IND)) {
			tab_rt->rte_flags |= RTE_IND;
			tab_rt->rte_l2t_ix = gw_rt->rte_l2t_ix;
			INIT_LIST_HEAD(&tab_rt->gw_dep);
			list_add(&tab_rt->gw_dep, &gw_rt->gw_dep);
		}

		/* The indirect route is made direct */
		if (!(r->rte_flags & RTE_IND) && (tab_rt->rte_flags & RTE_IND)) {
			list_del_init(&tab_rt->gw_dep);
			tab_rt->rte_flags &= ~RTE_IND;
			list_del_init(&tab_rt->gw_dep);
			tab_rt->rte_l2t_ix = r->rte_l2t_ix;
		}
	}

	/* Update the TCAM */
	skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
	if (!skb) return -ENOBUFS;

	req = (struct cpl_rte_write_req *)__skb_put(skb, sizeof(*req));
	memset(req, 0, sizeof(*req));
	tcam_index = t->tcam_rte_base + index;
	l2t_ix = tab_rt->rte_l2t_ix;
	w_control = (control == TCAM_RTE_UPDATE ? TCAM_RTE_CTRL_UPDATE
						  : TCAM_RTE_CTRL_INSERT);

	OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_RTE_WRITE_REQ,
				RTE_BASE_ATID + index));
	req->params = htonl(V_RTE_WRITE_REQ_L2T_IX(l2t_ix)
			    | V_RTE_WRITE_REQ_LUT_IX(tcam_index)
			    | V_RTE_WRITE_REQ_CTL(w_control));
	req->faddr = tab_rt->rte_faddr;
	req->netmask = htonl(tab_rt->rte_netmask); /* XXX why htonl() ? */
	toe_send(dev, skb);

	/* Update the dependant routes of a gateway */
	if (gw_dep_update) {
		struct list_head *entry;
		struct tcam_rte_entry *ind_rt;

		list_for_each(entry, &tab_rt->gw_dep) {
			ind_rt = list_entry(entry,
					    struct tcam_rte_entry,
					    gw_dep);

			if (!(ind_rt->rte_flags & RTE_IN_TCAM))
				continue;

			ind_rt->rte_l2t_ix = l2t_ix;

			skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
			if (!skb) return -ENOBUFS;

			req = (struct cpl_rte_write_req *)__skb_put(skb, sizeof(*req));
			memset(req, 0, sizeof(*req));

			index = ind_rt->rte_index;
			tcam_index = t->tcam_rte_base + index;
			w_control = TCAM_RTE_CTRL_UPDATE;

			OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_RTE_WRITE_REQ,
						RTE_BASE_ATID + index));
			req->params = htonl(V_RTE_WRITE_REQ_L2T_IX(l2t_ix)
					| V_RTE_WRITE_REQ_LUT_IX(tcam_index)
					| V_RTE_WRITE_REQ_CTL(w_control));

			req->faddr = ind_rt->rte_faddr;
			req->netmask = htonl(ind_rt->rte_netmask);
			toe_send(dev, skb);
		}
	}
out:
	return 0;
}

/*
 * Process an RTE_WRITE CPL message.
 */
static int do_rte_write_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_rte_write_rpl *rpl = cplhdr(skb);
    struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
    tid_t tid = GET_TID(rpl);
    struct tcam_rte_entry *rt;
    u8 status;
    int index;

    index = tid - RTE_BASE_ATID;
    rt = &t->cache_rte[index];

    status = rpl->status;

    if (!status) {
        rt->rte_flags |= RTE_IN_TCAM;
    } else {
        /* TCAM write failed. Update the cache */
        rt->rte_flags &= ~RTE_IN_TCAM;
        tcam_rte_delete(dev, rt);
        printk(KERN_ERR
               "Unexpected status %d for CPL_RTE_WRITE_RPL for index %u\n",
               rpl->status, index);
    }
    return CPL_RET_BUF_DONE;
}

/*
 * Delete the route at the given index.
 * Cleans up the TCAM if the route was inserted.
 */
static int tcam_rte_del_index(struct toedev *dev, int index)
{
	struct tcam_rte_entry *r;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
	struct cpl_rte_delete_req *req = NULL;
	struct sk_buff *skb;
	u32 tcam_index;

	if (!t) return -1;

	r = &t->cache_rte[index];
	if ((r->rte_flags & RTE_IN_TCAM)) {
		/* Update the TCAM */
		skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
		if (!skb) return -ENOBUFS;

		req = (struct cpl_rte_delete_req *)__skb_put(skb, sizeof(*req));
		memset(req, 0, sizeof(*req));
		tcam_index = t->tcam_rte_base + index;
		OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_RTE_DELETE_REQ,
					RTE_BASE_ATID + index));
		req->params = htonl(V_RTE_DELETE_REQ_LUT_IX(tcam_index));
		toe_send(dev, skb);
	}

	/* Update the soft cache */
	memset(r, 0, sizeof(*r));

	return 0;
}

/*
 * Process an RTE_DELETE CPL message.
 */
static int do_rte_delete_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_rte_delete_rpl *rpl = cplhdr(skb);
    struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
    tid_t tid = GET_TID(rpl);
    u8 status;
    int index;

    index = tid - RTE_BASE_ATID;
    status = rpl->status;

    atomic_dec(&t->tcam_rte_cnt);
    if (status) {
        printk(KERN_ERR
               "Unexpected status %d for CPL_RTE_DELETE_RPL for index %u\n",
               status, index);
    }
    return CPL_RET_BUF_DONE;
}

/*
 * Shift the given entry to the given index.
 * Cleans the previous spot if requested.
 */
static void tcam_rte_shift(struct toedev *dev,
			   struct tcam_rte_entry *r,
			   int nix, int del)
{
	int old_ix = r->rte_index;

	tcam_rte_write_index(dev, r, nix, TCAM_RTE_SHIFT);
	if (del)
		tcam_rte_del_index(dev, old_ix);

}

/* Exported TCAM route insertion/update routine */
int tcam_rte_insert(struct toedev *dev, struct tcam_rte_entry *r)
{
	int nml, i, index;
	struct tcam_opt_region *rg;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
	struct tcam_rte_entry *rt = NULL;

	if (!t) return -1;

	nml = r->rte_nm_len;
	rg = t->opt_reg;

	/* If the route exists, update it */
	index = tcam_rte_index(dev, r);
	if (index >= 0) {
		tcam_rte_write_index(dev, r, index,
				     TCAM_RTE_UPDATE);
		return 0;
	}

	/* Lock the structure before checking to the route counter */
	down(&t->rte_mutex);
	if (tcam_rte_full(t)) {
		up(&t->rte_mutex);
		return -1;
	}
	atomic_inc(&t->tcam_rte_cnt);
	up(&t->rte_mutex);

	if (nml > REG_GAP0) {
		for (i = REG_GAP0 + 1; i <= nml; i++) {
			if (rg[i].rg_lo != rg[i].rg_hi) {
				/*
				 * Current region is not empty.
				 * Shift its lower entry above.
				 */
				rt = &t->cache_rte[rg[i].rg_lo];
				index = rg[i].rg_hi;
				tcam_rte_shift(dev, rt, index, 0);
			}
			rg[i].rg_hi++;
			if (i != nml)
				rg[i].rg_lo++;
		}
		tcam_rte_write_index(dev, r, rg[nml].rg_lo,
				     TCAM_RTE_NEW);
	} else {
		for (i = REG_GAP0; i >= nml; i--) {
			if (rg[i].rg_lo != rg[i].rg_hi) {
				/*
				 * Current region is not empty.
				 * Shift its higher entry below.
				 */
				rt = &t->cache_rte[rg[i].rg_hi];
				index = rg[i].rg_lo;
				tcam_rte_shift(dev, rt, index, 0);
			}
			rg[i].rg_lo--;
			if (i != nml)
				rg[i].rg_hi--;
		}
		tcam_rte_write_index(dev, r, rg[nml].rg_hi,
				     TCAM_RTE_NEW);
	}
	return 0;
}

/* Exported TCAM route deletion routine */
int tcam_rte_delete(struct toedev *dev, struct tcam_rte_entry *r)
{
	int nml, i;
	struct tcam_opt_region *rg;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
	struct tcam_rte_entry *rt = NULL;
	int index;

	if (!t) return -1;
	index = tcam_rte_index(dev, r);
	if (index < 0) return -1;

	rt = &t->cache_rte[index];

	if ((rt->rte_flags & RTE_GW) && !list_empty(&rt->gw_dep)) {
		/*
		 * Gateway with depending routes.
		 * Build a write request to reset the L2T index.
		 */
		struct tcam_rte_entry tcam_rte;
		memset(&tcam_rte, 0, sizeof(tcam_rte));
		tcam_rte.rte_faddr = rt->rte_faddr;
		tcam_rte.rte_nm_len = rt->rte_nm_len;
		tcam_rte.rte_netmask = TCAM_NM32(rt->rte_nm_len);
		tcam_rte.rte_l2t_ix = 0;
		tcam_rte_write_index(dev, &tcam_rte, index, TCAM_RTE_UPDATE);
		return 0;
	}

	if (rt->rte_flags & RTE_IND)
		list_del_init(&rt->gw_dep);

	nml = r->rte_nm_len;
	rg = t->opt_reg;

	if (nml > REG_GAP0) {
		/* Shift the region's higher entry in the deleted spot */
		rt = &t->cache_rte[rg[nml].rg_hi - 1];
		tcam_rte_shift(dev, rt, index, 1);
		rg[nml].rg_hi--;

		for (i = nml - 1; i > REG_GAP0; i--) {
			if (rg[i].rg_lo != rg[i].rg_hi) {
				/*
				 * Current region is not empty.
				 * Shift its higher entry above.
				 */
				rt = &t->cache_rte[rg[i].rg_hi - 1];
				index = rg[i + 1].rg_hi;
				tcam_rte_shift(dev, rt, index, 1);
			}
			rg[i].rg_lo--;
			rg[i].rg_hi--;
		}
	} else {
		/* Shift the region's lower entry to the deleted spot */
		rt = &t->cache_rte[rg[nml].rg_lo + 1];
		tcam_rte_shift(dev, rt, index, 1);
		rg[nml].rg_lo++;

		for (i = nml + 1; i <= REG_GAP0; i++) {
			if (rg[i].rg_lo != rg[i].rg_hi) {
				/*
				 * Current region is not empty.
				 * Shift its lower entry below.
				 */
				rt = &t->cache_rte[rg[i].rg_lo + 1];
				index = rg[i - 1].rg_lo;
				tcam_rte_shift(dev, rt, index, 1);
			}
			rg[i].rg_lo++;
			rg[i].rg_hi++;
		}
	}
	return 0;
}

/*
 * Main table dumping into TCAM related routines.
 * Inspired by the iproute2 package'source code.
 */

 /* Sets up the routing attributes table */
static int parse_rtattr(struct rtattr *tb[], int max,
			struct rtattr *rta, int len)
{
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= max)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}
#if 0
	if (len)
		printk("parse_rtattr: Deficit %d, rta_len=%d\n",
		       len, rta->rta_len);
	return 0;
#endif
}

/* Add the main table entries one by one in the TCAM */
static void tcam_rte_main_insert(struct toedev *dev,
				 struct nlmsghdr *n,
				 int len)
{
	struct rtattr *t_attr[RTA_MAX+1];
	struct rtmsg *r = NULL;
	int nlen;
	u32 addr;
	struct tcam_rte_entry tcam_rte;

	while (NLMSG_OK(n, len)) {
		memset(&tcam_rte, 0, sizeof(tcam_rte));
		r = NLMSG_DATA(n);
		nlen = len - NLMSG_LENGTH(sizeof(*r));
		memset(t_attr, 0, sizeof(t_attr));
		parse_rtattr(t_attr, RTA_MAX, RTM_RTA(r), nlen);

		if (r->rtm_dst_len) {
			/* Insert the destination address in the TCAM */
			addr = *(u32 *)RTA_DATA(t_attr[RTA_DST]);
			tcam_rte.rte_faddr = addr;
			tcam_rte.rte_netmask = TCAM_NM32(r->rtm_dst_len);
			tcam_rte.rte_nm_len = r->rtm_dst_len;
		}

		if (t_attr[RTA_GATEWAY]) {
			struct tcam_rte_entry tcam_gw;
			addr = *(u32 *)RTA_DATA(t_attr[RTA_GATEWAY]);

			/* Insert the gateway in the TCAM. */
			memset(&tcam_gw, 0, sizeof(tcam_gw));
			tcam_gw.rte_faddr = addr;
			tcam_gw.rte_nm_len = 32;
			tcam_gw.rte_netmask = TCAM_NM32(32);
			tcam_gw.rte_flags |= RTE_GW;
			tcam_rte_insert(dev, &tcam_gw);

			/* Provide info to retrieve the gateway */
			tcam_rte.rte_flags |= RTE_IND;
			tcam_rte.gw_opt_addr = tcam_gw.rte_faddr;
			tcam_rte.gw_opt_nm_len = tcam_gw.rte_nm_len;
		}

		tcam_rte_insert(dev, &tcam_rte);

		n = NLMSG_NEXT(n, len);
	}
}

/* Dumps the main table entries in the TCAM */
static int tcam_rte_dump_main(struct toedev *tdev)
{
	struct sk_buff *skb_re = NULL;
	struct sk_buff *skb_rp = NULL;
	struct nlmsghdr *n;
	struct netlink_callback *cb;
	struct fib_table *tb;
	int len;
	int ret;

	if (!tdev) return -1;

	skb_re = alloc_skb(sizeof(*n), GFP_ATOMIC);
	if (!skb_re) return -ENOBUFS;

	n = (struct nlmsghdr *)skb_re->data;
	memset(n, 0, sizeof(*n));

	cb = (struct netlink_callback *)kmalloc(sizeof(*cb), GFP_ATOMIC);
	if (!cb) {
		ret = -ENOBUFS;
		goto out;
	}

	memset(cb, 0, sizeof(*cb));
	cb->args[0] = RT_TABLE_MAIN;
	cb->skb = skb_re;
	cb->nlh = n;

	tb = fib_get_table(RT_TABLE_MAIN);
	if (!tb || tb != ip_fib_main_table) {
		ret = -1;
		goto out;
	}

	skb_rp = alloc_skb(NLMSG_GOODSIZE, GFP_ATOMIC);
	if (!skb_rp) {
		ret = -ENOBUFS;
		goto out;
	}

	ret = tb->tb_dump(tb, skb_rp, cb);
	if (ret < 0) {
		printk(KERN_ERR "tcam_rte_dump_main: table dump failed\n");
		goto out;
	}

	n = (struct nlmsghdr *)skb_rp->data;
	if (n->nlmsg_type != RTM_NEWROUTE && n->nlmsg_type != RTM_DELROUTE) {
		ret = -1;
		printk(KERN_ERR "tcam_rte_dump_main: wrong message type\n");
		goto out;
	}

	len = ret;
	tcam_rte_main_insert(tdev, n, len);

out:
	if (skb_re)
		kfree(skb_re);
	if (skb_rp)
		kfree(skb_rp);
	if (cb)
		kfree(cb);

	return ret;
}

/*
 * Route add/deletion events notifying routines
 */
static DECLARE_MUTEX(tcam_notify_mutex);
static struct notifier_block *tcam_rte_notify_list;

int register_tcam_rte_notifier(struct notifier_block *nb)
{
	int err;

	down(&tcam_notify_mutex);
	err = notifier_chain_register(&tcam_rte_notify_list, nb);
	up(&tcam_notify_mutex);
	return err;
}

int unregister_tcam_rte_notifier(struct notifier_block *nb)
{
	int err;

	down(&tcam_notify_mutex);
	err = notifier_chain_unregister(&tcam_rte_notify_list, nb);
	up(&tcam_notify_mutex);
	return err;
}

/*
 * Replacement routines for host main routing table's delete and insert methods.
 */

/* Placeholder the original ip_fib_main_table->tb_insert() method */
static int (*tcam_rte_save_fib_insert)(struct fib_table *table, struct rtmsg *r,
				       struct kern_rta *rta, struct nlmsghdr *n,
				       struct netlink_skb_parms *req) = NULL;
/*
 * The new ip_fib_main_table->tb_insert() method.
 * Calls the original one and notifies the event to all TOEs.
 */
static int tcam_fib_hash_insert(struct fib_table *cb, struct rtmsg *r,
				struct kern_rta *rta, struct nlmsghdr *n,
				struct netlink_skb_parms *req)
{
	int ret;

	ret = tcam_rte_save_fib_insert(cb, r, rta, n, req);
	if (!ret) {
		struct tcam_rte_entry tcam_rte, tcam_gw;

		memset(&tcam_rte, 0, sizeof(tcam_rte));
		if (rta->rta_dst) {
			tcam_rte.rte_faddr = *(u32 *)rta->rta_dst;
			tcam_rte.rte_nm_len = r->rtm_dst_len;
			tcam_rte.rte_netmask = TCAM_NM32(r->rtm_dst_len);
		}

		if (rta->rta_gw) {
			 /* Insert the gateway in the TCAM. */
			memset(&tcam_gw, 0, sizeof(tcam_gw));
			tcam_gw.rte_faddr = *(u32 *)rta->rta_gw;
			tcam_gw.rte_nm_len = 32;
			tcam_gw.rte_netmask = TCAM_NM32(32);
			tcam_gw.rte_flags |= RTE_GW;

			down(&tcam_notify_mutex);
			notifier_call_chain(&tcam_rte_notify_list,
					    TCAM_RTE_INSERT,
					    &tcam_gw);
			up(&tcam_notify_mutex);

			/* Provide info to retrieve the gateway */
			tcam_rte.rte_flags |= RTE_IND;
			tcam_rte.gw_opt_addr = tcam_gw.rte_faddr;
			tcam_rte.gw_opt_nm_len = tcam_gw.rte_nm_len;

		}

		down(&tcam_notify_mutex);
		notifier_call_chain(&tcam_rte_notify_list,
				    TCAM_RTE_INSERT,
				    &tcam_rte);
		up(&tcam_notify_mutex);
	}
	return ret;
}

/* Placeholder the original ip_fib_main_table->tb_delete() method */
static int (*tcam_rte_save_fib_delete)(struct fib_table *table, struct rtmsg *r,
				       struct kern_rta *rta, struct nlmsghdr *n,
				       struct netlink_skb_parms *req) = NULL;
/*
 * The new ip_fib_main_table->tb_delete() method.
 * Calls the original one and notifies the event to all TOEs.
 */
static int tcam_fib_hash_delete(struct fib_table *cb, struct rtmsg *r,
				struct kern_rta *rta, struct nlmsghdr *n,
				struct netlink_skb_parms *req)
{
	int ret;

	ret = tcam_rte_save_fib_delete(cb, r, rta, n, req);
	if (!ret) {
		struct tcam_rte_entry tcam_rte;
		if (rta->rta_dst) {
			memset(&tcam_rte, 0, sizeof(tcam_rte));
			tcam_rte.rte_faddr = *(u32 *)rta->rta_dst;
			tcam_rte.rte_nm_len = r->rtm_dst_len;
			tcam_rte.rte_netmask = TCAM_NM32(r->rtm_dst_len);

			down(&tcam_notify_mutex);
			notifier_call_chain(&tcam_rte_notify_list,
					    TCAM_RTE_DELETE,
					    &tcam_rte);
			up(&tcam_notify_mutex);
		}
	}
	return ret;
}

/*
 * Configure TCAM routing entries with available info
 * from the host main routing table.
 * Register for upcoming routing events.
 */
static DECLARE_MUTEX(tcam_rte_mod_main_tb);

int tcam_rte_start(struct toedev *dev)
{
	struct tcam_rte_unit *t;
	struct tcam_rte_entry tcam_rt;
	struct fib_table *tb;
	int ret;

	if (!dev) return -1;
	t = TOM_RTE_DATA(dev);

	if (!t) return -1;

	/* Sets the default entry */
	memset(&tcam_rt, 0, sizeof(tcam_rt));
	tcam_rt.rte_index = t->rte_n_entries - 1;
	tcam_rte_insert(dev, &tcam_rt);

	/* Push the host's main routing table in the TCAM */
	ret = tcam_rte_dump_main(dev);
	if (ret < 0) return -1;

	/* Modify the host's main routing table insert/delete methods */
	tb = fib_get_table(RT_TABLE_MAIN);
	if (!tb || tb != ip_fib_main_table) return -1;

	down(&tcam_rte_mod_main_tb);
	if (tcam_rte_save_fib_insert == NULL) {
		tcam_rte_save_fib_insert = tb->tb_insert;
		tb->tb_insert = tcam_fib_hash_insert;

		tcam_rte_save_fib_delete = tb->tb_delete;
		tb->tb_delete = tcam_fib_hash_delete;
	}
	up(&tcam_rte_mod_main_tb);

	return 0;
}

/* Early initialization of the TCAM routing structures */
struct tcam_rte_unit *init_tcam_rte(struct toe_tid_range *tid,
				    int term_l2t_capacity)
{
	struct tcam_rte_unit *t;
#ifdef CONFIG_PROC_FS
	struct tcam_rte_entry *proc_rt;
#endif
	int size;

	size = sizeof(*t) +
	       (tid->num * (sizeof(struct tcam_rte_entry)));

	t = kmalloc(size, GFP_KERNEL);
	if (!t) goto out;

	memset(t, 0, size);

	t->tcam_rte_base = tid->base;
	t->rte_n_entries = tid->num;
	t->l2t_n_entries = term_l2t_capacity;
	t->cache_rte = (struct tcam_rte_entry *)(t + 1);
	tcam_opt_reg_init(t);
	init_MUTEX(&t->rte_mutex);

	/* The last entry is reserved for default behaviour */
	atomic_set(&t->tcam_rte_cnt, 1);

#ifdef CONFIG_PROC_FS
	size = (tid->num * (sizeof(struct tcam_rte_entry))) +
	       (term_l2t_capacity * (sizeof(struct term_l2t_entry)));
	proc_rt = kmalloc(size, GFP_KERNEL);
	if (!proc_rt) goto out_t_free;

	memset(proc_rt, 0, size);

	t->rte_entry = proc_rt;
	t->l2t_entry = (struct term_l2t_entry *)&proc_rt[t->rte_n_entries];
	atomic_set(&t->pending_reqs, 0);
	init_waitqueue_head(&t->cpl_rte_wait);
	init_MUTEX(&t->read_rte_mutex);
#endif
	goto out;

out_t_free:
	kfree(t);
out:
	return t;
}

void free_tcam_rte(struct tcam_rte_unit *t)
{
#ifdef CONFIG_PROC_FS
	kfree(t->rte_entry);
#endif
	kfree(t);
}

#ifdef CONFIG_PROC_FS
static int tcam_req_rtinfo(struct toedev *dev, u32 base, int i)
{
	struct cpl_rte_read_req *req;
	struct sk_buff *skb;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
	u32 tcam_index;

	if (!t) return -1;

	/* Send a request for the foreign address (selector = 0) */
	skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
	if (!skb) return -1;

	req = (struct cpl_rte_read_req *)__skb_put(skb, sizeof(*req));
	memset(req, 0, sizeof(*req));
	OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_RTE_READ_REQ,
				(RTE_BASE_ATID + 2 * i)));
	tcam_index = base + i;

	req->params = htonl(V_RTE_READ_REQ_LUT_IX(tcam_index));

	atomic_inc(&t->pending_reqs);
	toe_send(dev, skb);

	/* Send a request for the netmask (selector = 1) */
	skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
	if (!skb) return -1;

	req = (struct cpl_rte_read_req *)__skb_put(skb, sizeof(*req));
	memset(req, 0, sizeof(*req));
	OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_RTE_READ_REQ,
				(RTE_BASE_ATID + 2 * i + 1)));
	req->params = htonl(V_RTE_READ_REQ_LUT_IX(tcam_index)
			    | V_RTE_READ_REQ_SELECT);

	atomic_inc(&t->pending_reqs);
	toe_send(dev, skb);

	return 0;
}

/*
 * Process an RTE_READ CPL message.
 */
static int do_rte_read_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_rte_read_rpl *rpl = cplhdr(skb);
    struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
    struct tcam_rte_entry *e;
    tid_t tid = GET_TID(rpl);
    int index;

    if (!t) return -1;

    index = (rpl->select) ? (tid - 1 - RTE_BASE_ATID) / 2
                     : (tid - RTE_BASE_ATID) / 2;
    e = (struct tcam_rte_entry *)&t->rte_entry[index];
    e->rte_index = index;

    if (rpl->select) {
        e->rte_netmask = rpl->addr;
    } else {
        e->rte_status = rpl->status;
        e->rte_l2t_ix = rpl->l2t_idx;
        e->rte_faddr = rpl->addr;
    }

    if (atomic_dec_and_test(&t->pending_reqs)) {
        wake_up(&t->cpl_rte_wait);
    }
    return CPL_RET_BUF_DONE;
}

static int tcam_req_arpinfo(struct toedev *dev, int i)
{
	struct cpl_l2t_read_req *req;
	struct sk_buff *skb;
	struct tcam_rte_unit *t = TOM_RTE_DATA(dev);

	skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
	if (!skb) return -1;

	req = (struct cpl_l2t_read_req *)__skb_put(skb, sizeof(*req));
	OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_L2T_READ_REQ, i));
	req->l2t_idx = i;

	atomic_inc(&t->pending_reqs);
	toe_send(dev, skb);

	return 0;
}

/*
 * Process an L2T_READ CPL message.
 */
static int do_l2t_read_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_l2t_read_rpl *rpl = cplhdr(skb);
    struct tcam_rte_unit *t = TOM_RTE_DATA(dev);
    struct term_l2t_entry *e;
    tid_t tid = GET_TID(rpl);

    if (!t) return -1;

    e = (struct term_l2t_entry *)&t->l2t_entry[tid];
    e->l2t_index = tid;
    e->l2t_status = G_L2TOP_STATUS(ntohl(rpl->params));
    e->l2t_phy_port = G_L2TOP_IFF(ntohl(rpl->params));
    e->l2t_vlan_tag = G_L2TOP_VLAN(ntohl(rpl->params));
    memcpy(e->l2t_dmac, rpl->dst_mac, sizeof(e->l2t_dmac));

    if (atomic_dec_and_test(&t->pending_reqs)) {
        wake_up(&t->cpl_rte_wait);
    }
    return CPL_RET_BUF_DONE;
}

static void *tcam_rte_seq_start(struct seq_file *seq, loff_t *pos)
{
	struct toedev *dev = seq->private;
	struct tcam_rte_unit *t;

	if (!dev) return NULL;

	t = TOM_RTE_DATA(dev);
	if (!t) return NULL;

	return ((*pos < t->rte_n_entries) ? &t->rte_entry[*pos] : NULL);
}

static void *tcam_rte_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	struct tcam_rte_entry *e = v;
	struct toedev *dev = seq->private;
	struct tcam_rte_unit *t;

	t = TOM_RTE_DATA(dev);
	++e;
	++*pos;
	v = (*pos < (t->rte_n_entries)) ? e : NULL;

	return v;
}

static void tcam_rte_seq_stop(struct seq_file *seq, void *v)
{
	struct tcam_rte_unit *t;
	struct toedev *dev = seq->private;

	t = TOM_RTE_DATA(dev);
	up(&t->read_rte_mutex);
}

static int tcam_rte_seq_show(struct seq_file *seq, void *v)
{
	struct tcam_rte_entry *re = v;
	struct term_l2t_entry *le;
	struct tcam_rte_unit *t;
	struct toedev *dev = seq->private;
	char ip_faddr[20];
	char ip_netmask[20];
	char mac_addr[20];

	if (re->rte_index == 0) {
		seq_puts(seq,
			 "Index  IP address       Netmask          "
			 "L2T_IX  Ethernet address\n");
	}

	t = TOM_RTE_DATA(dev);
	if(!re->rte_status) {
		sprintf(ip_faddr, "%u.%u.%u.%u", NIPQUAD(re->rte_faddr));
		sprintf(ip_netmask,"%u.%u.%u.%u", NIPQUAD(re->rte_netmask));
		if (re->rte_l2t_ix) {
			le = &t->l2t_entry[re->rte_l2t_ix];
			sprintf(mac_addr,
				"%02x:%02x:%02x:%02x:%02x:%02x",
				le->l2t_dmac[0], le->l2t_dmac[1],
				le->l2t_dmac[2], le->l2t_dmac[3],
				le->l2t_dmac[4], le->l2t_dmac[5]);
		} else {
			sprintf(mac_addr, "%s", "N/A");
		}
		seq_printf(seq, "%-5d  %-15s  %-15s  %-6d  %-17s\n",
			   re->rte_index, ip_faddr, ip_netmask,
			   re->rte_l2t_ix, mac_addr);
	} else {
		sprintf(ip_faddr, "invalid entry");
		seq_printf(seq, "%-5d %-15s\n",
			   re->rte_index, ip_faddr);
	}
	return 0;
}

static struct seq_operations tcam_rte_seq_ops = {
	.start  = tcam_rte_seq_start,
	.next   = tcam_rte_seq_next,
	.stop   = tcam_rte_seq_stop,
	.show   = tcam_rte_seq_show
};

static int tcam_rte_seq_open(struct inode *inode, struct file *file)
{

	int rc = seq_open(file, &tcam_rte_seq_ops);

	if (!rc) {
		struct proc_dir_entry *dp = PDE(inode);
		struct seq_file *seq = file->private_data;
		struct toedev *dev;
		struct tcam_rte_unit *t;
		int i, error;

		if (!seq || !dp) return -1;

		seq->private = dp->data;

		dev = seq->private;
		if (!dev) return -1;

		t = TOM_RTE_DATA(dev);
		if (!t) return -1;

		/* Serialize monitoring */
		down(&t->read_rte_mutex);

		for (i = 0; i < t->rte_n_entries; i++) {
			if ((error = tcam_req_rtinfo(dev,
						     t->tcam_rte_base,
						     i))) {
				printk(KERN_ERR
					"tcam_rte: rte error on index %d\n", i);
			}
		}
		for (i = 1; i < t->l2t_n_entries; i++) {
			if ((error = tcam_req_arpinfo(dev, i))) {
				printk(KERN_ERR
					"tcam_rte: l2t error on index %d\n", i);
			}
		}
		wait_event(t->cpl_rte_wait, !(atomic_read(&t->pending_reqs)));
	}
	return rc;
}

static struct file_operations tcam_rte_seq_fops = {
	.owner   = THIS_MODULE,
	.open    = tcam_rte_seq_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
};

/*
 * Create the proc entries for the TCAM routing table under dir.
 */
int tcam_rte_proc_setup(struct proc_dir_entry *dir, struct toedev *dev)
{
    struct proc_dir_entry *p;

    if (!dir) return -EINVAL;

    p = create_proc_entry("tcam_rte", S_IRUGO, dir);
    if (!p) return -ENOMEM;

    p->proc_fops = &tcam_rte_seq_fops;
    p->data = dev;

    return 0;
}

void tcam_rte_proc_free(struct proc_dir_entry *dir)
{
	if (dir) {
		remove_proc_entry("tcam_rte", dir);
	}
}
#endif /* CONFIG_PROC_FS */

void __init t1_init_route_cpl_handlers(void)
{
    t1_register_cpl_handler(CPL_RTE_WRITE_RPL, do_rte_write_rpl);
    t1_register_cpl_handler(CPL_RTE_DELETE_RPL, do_rte_delete_rpl);
#ifdef CONFIG_PROC_FS
    t1_register_cpl_handler(CPL_RTE_READ_RPL, do_rte_read_rpl);
    t1_register_cpl_handler(CPL_L2T_READ_RPL, do_l2t_read_rpl);
#endif
}
