/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <linux/toedev.h>
#include <net/offload.h>
#include "../toe_compat.h"
#include "defs.h"
#include "tom.h"
#include "../cpl5_cmd.h"
#include "../toe_ctl_defs.h"
#include "l2t.h"
#include "tcam_rte.h"
#ifdef CONFIG_CHELSIO_T1_ULP
#include "../ulp/ulp.h"
#endif
#ifdef CONFIG_CHELSIO_T1_BONDING
#include "bond.h"
#endif

#ifdef work_struct
#include <linux/tqueue.h>
#else
#include <linux/workqueue.h>
#endif

static const int MAX_ATIDS = 64 * 1024;
static const int ATID_BASE = 0x100000;

static LIST_HEAD(toe_list);
static DECLARE_MUTEX(toe_list_lock);

static struct sk_buff_head deferq;
static struct work_struct deferq_task;

/*
 * Add an skb to the deferred skb queue for processing from process context.
 * This function should be called only from soft IRQ context.
 */
void t1_defer_reply(struct sk_buff *skb, struct toedev *dev,
		    defer_handler_t handler)
{
    DEFERRED_SKB_CB(skb)->handler = handler;
    DEFERRED_SKB_CB(skb)->dev = dev;
    spin_lock(&deferq.lock);
    __skb_queue_tail(&deferq, skb);
    if (skb_queue_len(&deferq) == 1)
	schedule_work(&deferq_task);
    spin_unlock(&deferq.lock);
}

/*
 * Process the defer queue.
 */
static void process_deferq(void *data)
{
    struct sk_buff *skb;

    spin_lock_bh(&deferq.lock);
    while ((skb = __skb_dequeue(&deferq)) != NULL) {
	spin_unlock_bh(&deferq.lock);
	DEFERRED_SKB_CB(skb)->handler(DEFERRED_SKB_CB(skb)->dev, skb);
	spin_lock_bh(&deferq.lock);
    }
    spin_unlock_bh(&deferq.lock);
}

/*
 * Allocate and initialize the TID tables.  Returns 0 on success.
 */
static int init_tid_tabs(struct tid_info *t, unsigned int ntids,
			 unsigned int natids, unsigned int nstids,
			 unsigned int atid_base, unsigned int stid_base)
{
    unsigned long size = ntids * sizeof(*t->tid_tab) +
	natids * sizeof(*t->atid_tab) + nstids * sizeof(*t->stid_tab);

    t->tid_tab = vmalloc(size);
    if (!t->tid_tab) return -ENOMEM;

    size /= sizeof(*t->tid_tab);
    t->mss_tab = vmalloc(size);
    if (!t->mss_tab) {
	vfree(t->tid_tab);
	return -ENOMEM;
    } 

    t->stid_tab = (union listen_entry *)&t->tid_tab[ntids];
    t->atid_tab = (union active_open_entry *)&t->stid_tab[nstids];
    t->ntids = ntids;
    t->nstids = nstids;
    t->stid_base = stid_base;
    t->natids = natids;
    t->atid_base = atid_base;
    t->stids_in_use = t->atids_in_use = 0;
    atomic_set(&t->tids_in_use, 0);

    t->stid_lock = SPIN_LOCK_UNLOCKED;
    t->atid_lock = SPIN_LOCK_UNLOCKED;

    memset(t->tid_tab, 0, ntids * sizeof(*t->tid_tab));

    /*
     * Setup the free lists for stid_tab and atid_tab.
     */
    if (nstids) {
	t->stid_tab[--nstids].next = NULL;
	for ( ; nstids; --nstids)
	    t->stid_tab[nstids - 1].next = &t->stid_tab[nstids];
	t->sfree = t->stid_tab;
    } else
	t->sfree = NULL;

    t->atid_tab[--natids].next = NULL;
    for ( ; natids; --natids)
	t->atid_tab[natids - 1].next = &t->atid_tab[natids];
    t->afree = t->atid_tab;
    return 0;
}

static void free_tid_maps(struct tid_info *t)
{
    vfree(t->mss_tab);
    vfree(t->tid_tab);
}

/*
 * Process a received packet with an unknown/unexpected CPL opcode.
 */
static int do_bad_cpl(struct toedev *dev, struct sk_buff *skb)
{
    printk(KERN_ERR "%s: received bad CPL command %d\n", dev->name,
	   *skb->data);
    return CPL_RET_BUF_DONE | CPL_RET_BAD_MSG;
}

/*
 * Handlers for each CPL opcode
 */
static cpl_handler_t cpl_handlers[NUM_CPL_CMDS];

/*
 * Add a new handler to the CPL dispatch table.  A NULL handler may be supplied
 * to unregister an existing handler.
 */
void t1_register_cpl_handler(unsigned int opcode, cpl_handler_t h)
{
    if (opcode < NUM_CPL_CMDS)
	cpl_handlers[opcode] = h ? h : do_bad_cpl;
    else
	printk(KERN_ERR
	       "Chelsio T2 TOM: handler registration for opcode %u failed\n",
	       opcode);
}

/*
 * TOM's receive method.
 */
static int process_rx(struct toedev *dev, struct sk_buff **skbs, int n)
{
	while (n--) {
		struct sk_buff *skb = *skbs++;
		union opcode_tid *p = cplhdr(skb);
		int ret = cpl_handlers[p->opcode](dev, skb);

#if VALIDATE_TID
		if (ret & CPL_RET_UNKNOWN_TID)
			printk(KERN_ERR "%s: CPL message (opcode %d) had "
			       "unknown TID %u\n", dev->name, p->opcode,
			       G_TID(ntohl(p->opcode_tid)));
#endif
		if (ret & CPL_RET_BUF_DONE)
			kfree_skb(skb);
	}
	return 0;
}

/*
 * Make a preliminary determination if a connection can be offloaded.  It's OK
 * to fail the offload later if we say we can offload here.  For now this
 * always accepts the offload request unless there are IP options.
 */
static int can_offload(struct toedev *dev, struct sock *sk)
{
    struct tom_data *d = TOM_DATA(dev);
    struct tid_info *t = &d->tid_maps;

    /* Do not offload the r* commands until we move TIME_WAIT to the host. */
    if (inet_addr_info(sk)->dport >= htons(512) &&
	inet_addr_info(sk)->dport <= htons(514))
	return 0;

    return inet_sk(sk)->opt == NULL &&
	d->tx_pages - 2 * d->tx_page_users >= d->conf.min_newconn_tx_pages &&
	(d->conf.max_conn < 0 ||
	 atomic_read(&t->tids_in_use) + t->atids_in_use < d->conf.max_conn);
}

/*
 * This is called through a notifier chain when a socket listen event is
 * published.  We iterate through all the TOEs we are handling and establish
 * or close listening servers as appropriate.
 */
static int listen_notify_handler(struct notifier_block *this,
				 unsigned long event, void *data)
{
    struct sock *sk = data;
    struct tom_data *p;

    switch (event) {
    case OFFLOAD_LISTEN_START:
    case OFFLOAD_LISTEN_STOP:
	down(&toe_list_lock);
	list_for_each_entry(p, &toe_list, list_node) {
	    if (event == OFFLOAD_LISTEN_START)
		t1_listen_start(p->dev, sk);
	    else
		t1_listen_stop(p->dev, sk);
	}
	up(&toe_list_lock);
	break;
    }
    return NOTIFY_DONE;
}

/*
 * Update every TCAM when the host main routing table is
 * added or deleted an entry.
 */
static int tcam_rte_notify_handler(struct notifier_block *this,
				   unsigned long event, void *data)
{
    struct tcam_rte_entry *rt = data;
    struct tom_data *p;

    switch (event) {
    case TCAM_RTE_INSERT:
    case TCAM_RTE_DELETE:
	down(&toe_list_lock);
	list_for_each_entry(p, &toe_list, list_node) {
	    if (event == TCAM_RTE_INSERT)
		tcam_rte_insert(p->dev, rt);
	    else
		tcam_rte_delete(p->dev, rt);
	}
	up(&toe_list_lock);
	break;
    }
    return NOTIFY_DONE;
}

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>

static int tom_info_read_proc(char *buf, char **start, off_t offset,
			      int length, int *eof, void *data)
{
    struct tom_data *d = data;
    struct tid_info *t = &d->tid_maps;
    int len;

    len = sprintf(buf, "TID range: 0..%d, in use: %u\n"
		  "STID range: %d..%d, in use: %u\n"
		  "ATID range: %d..%d, in use: %u\n"
		  "TX pages: %u x %uKB, in use: %u by %u connections\n"
		  "MSS: %u\n",
		  t->ntids - 1, atomic_read(&t->tids_in_use), t->stid_base,
		  t->stid_base + t->nstids - 1, t->stids_in_use,
		  t->atid_base, t->atid_base + t->natids - 1, t->atids_in_use,
		  d->tx_pages, d->tx_page_size >> 10,
		  d->tx_pages - d->tx_pages_free, d->tx_page_users,
		  d->tx_max_chunk);
    if (len > length) len = length;
    *eof = 1;
    return len;
}

#if 0
static void tom_info_proc_free(struct proc_dir_entry *dir)
{
    if (dir)
	remove_proc_entry("info", dir);
}
#endif

static int tom_info_proc_setup(struct proc_dir_entry *dir, struct tom_data *d)
{
    struct proc_dir_entry *p;

    if (!dir) return -EINVAL;

    p = create_proc_read_entry("info", 0, dir, tom_info_read_proc, d);
    if (!p) return -ENOMEM;

    p->owner = THIS_MODULE;
    return 0;
}

static void tom_proc_init(struct toedev *dev)
{
    l2t_proc_setup(dev->proc_dir, L2DATA(dev));
    t1_listen_proc_setup(dev->proc_dir, TOM_DATA(dev));
    tom_info_proc_setup(dev->proc_dir, TOM_DATA(dev));
    tcam_rte_proc_setup(dev->proc_dir, dev);
}

#if 0
static void tom_proc_cleanup(struct toedev *dev)
{
    l2t_proc_free(dev->proc_dir);
    tcam_rte_proc_free(dev->proc_dir);
    t1_listen_proc_free(dev->proc_dir);
    tom_info_proc_free(dev->proc_dir);
}
#endif
#else
#define tom_proc_init(dev)
#define tom_proc_cleanup(dev)
#endif

/*
 * Add a TOE to the list of TOEs we are managing.
 */
static void toe_add(struct tom_data *t)
{
    down(&toe_list_lock);
    list_add_tail(&t->list_node, &toe_list);
    up(&toe_list_lock);
}

#if 0
/*
 * Remove a TOE from the list of TOEs we are managing.
 */
static void toe_remove(struct tom_data *t)
{
    down(&toe_list_lock);
    list_del(&t->list_node);
    up(&toe_list_lock);
}
#endif

static int chelsio_toe_attach(struct toedev *dev, const struct toe_id *entry)
{
    struct tom_data *t;
    struct toe_tid_range stid_range, tid_range;
    struct toe_page_info tx_info;
#ifdef CONFIG_CHELSIO_T1_CM
    struct toe_page_info rx_info;
#endif
    int natids, i;
    unsigned int l2t_capacity;
    struct toe_tid_range toe_tcam_info;

    t = kmalloc(sizeof(*t), GFP_KERNEL);
    if (!t) return -ENOMEM;
    memset(t, 0, sizeof(*t));

    if (dev->ctl(dev, TOE_GET_TX_PAGE_INFO, &tx_info) < 0 ||
#ifdef CONFIG_CHELSIO_T1_CM
        dev->ctl(dev, TOE_GET_RX_PAGE_INFO, &rx_info) < 0 ||
#endif
	dev->ctl(dev, TOE_GET_TX_MAX_CHUNK, &t->tx_max_chunk) < 0 ||
	dev->ctl(dev, TOE_GET_L2T_CAPACITY, &l2t_capacity) < 0 ||
	dev->ctl(dev, TOE_GET_TID_RANGE, &tid_range) < 0 ||
	dev->ctl(dev, TOE_GET_STID_RANGE, &stid_range) < 0 ||
	dev->ctl(dev, TOE_GET_RTBL_RANGE, &toe_tcam_info) < 0 ||
	dev->ctl(dev,  TOE_GET_MAX_PORT, &t->max_nports) < 0)
	goto out_free;

    t->lmss = kmalloc(t->max_nports * sizeof(*(t->lmss)), GFP_KERNEL);
    if (!t->lmss) goto out_free;
    for (i = 0; i < t->max_nports; i++) {
        t->lmss[i] = t->tx_max_chunk;
    }

    L2DATA(dev) = init_l2t(l2t_capacity);
    if (!L2DATA(dev)) goto out_free;

    TOM_RTE_DATA(dev) = init_tcam_rte(&toe_tcam_info, l2t_capacity);
    if (!TOM_RTE_DATA(dev)) goto out_free_l2t;

    /* natids = min(ntids / 2, MAX_ATIDS, admin limit) */
    natids = tid_range.num / 2;
    if (natids > MAX_ATIDS) natids = MAX_ATIDS;
    if (init_tid_tabs(&t->tid_maps, tid_range.num, natids, stid_range.num,
		      ATID_BASE, stid_range.base) < 0)
        goto out_free_tcam_rte;

#ifdef CONFIG_CHELSIO_T1_ULP
    ULPDATA(dev) = t1_init_ulp(dev);
    if (!ULPDATA(dev))
        goto out_free_tid_maps;
#endif

    INIT_LIST_HEAD(&t->list_node);
    t->dev = dev;   /* XXX get a reference */
    t->listen_lock = RW_LOCK_UNLOCKED;
    spin_lock_init(&t->tx_page_lock);
    spin_lock_init(&t->synq_lock);
    init_waitqueue_head(&t->tx_waitq);
    t->tx_pages_free = t->tx_pages = tx_info.num;
    t->tx_page_size = tx_info.page_size;
    t->tx_page_mask = tx_info.page_size - 1;
    t->tx_page_shift = ffs(tx_info.page_size) - 1;
#ifdef CONFIG_CHELSIO_T1_CM
    t->tx_page_base = tx_info.pm_base;
    t->rx_page_base = rx_info.pm_base;
    t->rx_page_size = rx_info.page_size;
#endif
    t1_init_tunables(t, &t->conf);

    TOM_DATA(dev) = t;
    dev->recv = process_rx;
    dev->can_offload = can_offload;
    dev->connect = t1_connect;
    dev->neigh_update = l2t_update;

    /* All setup completed, let the driver know. */
    if (dev->open(dev)) goto free_all;

    tcam_rte_start(dev);
    tom_proc_init(dev);
#ifdef CONFIG_SYSCTL
    t->sysctl = t1_sysctl_register(dev, &t->conf);
#endif

    toe_add(t);   /* XXX there may be races here */
    return 0;

 free_all:
    dev->recv = NULL;
    dev->can_offload = NULL;
    dev->connect = NULL;
    dev->neigh_update = NULL;
    TOM_DATA(dev) = NULL;
#ifdef CONFIG_CHELSIO_T1_ULP
    t1_free_ulp(dev);
 out_free_tid_maps:
#endif
    free_tid_maps(&t->tid_maps);
 out_free_tcam_rte:
    free_tcam_rte(TOM_RTE_DATA(dev));
    TOM_RTE_DATA(dev) = NULL;
 out_free_l2t:
    free_l2t(L2DATA(dev));
    L2DATA(dev) = NULL;
 out_free:
    if (t->lmss) kfree(t->lmss);
    kfree(t);
    return -1;
}

static struct notifier_block listen_notifier = {
    .notifier_call = listen_notify_handler
};

static struct notifier_block tcam_rte_notifier = {
    .notifier_call = tcam_rte_notify_handler
};

static struct toe_id chelsio_toe_id_tab[] = {
	{ TOE_ID_CHELSIO_T1, 0 },
	{ TOE_ID_CHELSIO_T1C, 0 },
	{ 0, 0UL } /* END_OF_TOE_ID_TABLE */
};

static struct tom_info chelsio_tom_info = {
    .attach = chelsio_toe_attach,
    .id_table = chelsio_toe_id_tab,
    .name = "Chelsio_T2"
};

/*
 * Initialize the CPL dispatch table.
 */
static void __init init_cpl_handlers(void)
{
    int i;

    for (i = 0; i < NUM_CPL_CMDS; ++i)
	cpl_handlers[i] = do_bad_cpl;

    t1_init_listen_cpl_handlers();
    t1_init_route_cpl_handlers();
}

static int __init chelsio_tom_init(void)
{
    skb_queue_head_init(&deferq);
    INIT_WORK(&deferq_task, process_deferq, NULL);
    init_cpl_handlers();
    if (t1_init_cpl_io() < 0) return -1;
    t1_init_offload_ops();

    /*
     * Register with the TOE database.
     */
    if (register_tom(&chelsio_tom_info) != 0) {
	printk(KERN_ERR "Unable to register Chelsio TCP offload module.\n");
	return -1;
    }

    register_listen_offload_notifier(&listen_notifier);
    register_tcam_rte_notifier(&tcam_rte_notifier);

    return 0;
}

module_init(chelsio_tom_init);

// Local Variables:
// c-basic-offset: 4
// End:
