/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _CHELSIO_TOM_T1_H
#define _CHELSIO_TOM_T1_H

#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/wait.h>
#include "tom_types.h"
#include "tom_sysctl.h"

struct sock;

struct listen_info {
    struct listen_info *next;      /* Link to next entry */
    struct sock *sk;               /* The listening socket */
    tid_t stid;                    /* The server TID */
};

union listen_entry {
    struct sock *sk;
    union listen_entry *next;
};

union active_open_entry {
    struct sock *sk;
    union active_open_entry *next;
};

/*
 * Holds the size, base address, free list start, etc of the TID, server TID,
 * and active-open TID tables for a TOE.  The tables themselves are allocated
 * dynamically.
 */
struct tid_info {
    struct sock **tid_tab;
    unsigned int ntids;
    atomic_t tids_in_use;

    union listen_entry *stid_tab;
    unsigned int nstids;
    unsigned int stid_base;
    unsigned int stids_in_use;

    union active_open_entry *atid_tab;
    unsigned int natids;
    unsigned int atid_base;
    unsigned int atids_in_use;

    /*
     * The following members are accessed R/W so we put them in their own
     * cache lines.
     *
     * XXX We could combine the atid fields above with the lock here since
     * atids are use once (unlike other tids).  OTOH the above fields are
     * usually in cache due to tid_tab.
     */
    spinlock_t atid_lock ____cacheline_aligned;
    union active_open_entry *afree;

    spinlock_t stid_lock ____cacheline_aligned;
    union listen_entry *sfree;

    char *mss_tab;
};

#define LISTEN_INFO_HASH_SIZE 32

struct tom_data {
    struct list_head list_node;
    struct toedev *dev;
    unsigned int tx_pages;        /* # of TX pages */
    unsigned int tx_page_size;    /* page size (power of 2) */
    unsigned int tx_page_mask;    /* page size - 1 */
    unsigned int tx_page_shift;   /* log2(page size) */
    unsigned int tx_max_chunk;    /* max payload for TX_DATA */

#ifdef CONFIG_CHELSIO_T1_CM
    unsigned int rx_page_size;    /* page size (power of 2) */
    unsigned int tx_page_base;    /* pm base */
    unsigned int rx_page_base;   
#endif
    unsigned long pm_pressure_ts; /* latest PM pressure timestamp */
    u8 pm_pressure;               /* true if the timestamp is valid */

    u8 max_nports;
    unsigned int *lmss;
    
    struct tom_tunables conf;
    struct tom_sysctl_table *sysctl;

    struct tid_info tid_maps;

    spinlock_t tx_page_lock ____cacheline_aligned;
    wait_queue_head_t tx_waitq;
    unsigned int tx_pages_free;   /* # of free TX pages */
    unsigned int tx_page_users;

    rwlock_t listen_lock ____cacheline_aligned;
    struct listen_info *listen_hash_tab[LISTEN_INFO_HASH_SIZE];

    /*
     * Synchronizes access to the various SYN queues.  We assume that SYN
     * queue accesses do not cause much contention so that one lock for all
     * the queues suffices.  This is because the primary user of this lock
     * is the TOE softirq, which runs on one CPU and so most accesses should
     * be naturally contention-free.  The only contention can come from
     * listening sockets processing backlogged messages, and that should not
     * be high volume.
     */
    spinlock_t synq_lock ____cacheline_aligned;
};

/*
 * toedev -> tom_data accessor
 */
#define TOM_DATA(dev) (*(struct tom_data **)&(dev)->l4opt)

/*
 * Access a configurable parameter of a TOE device's TOM.
 */
#define TOM_TUNABLE(dev, param) (TOM_DATA(dev)->conf.param)

/*
 * Map an ATID or STID to their entries in the corresponding TID tables.
 */
#define ATID2ENTRY(t, atid) ((t)->atid_tab[(atid) - (t)->atid_base])
#define STID2ENTRY(t, stid) ((t)->stid_tab[(stid) - (t)->stid_base])

/*
 * Find the socket corresponding to a TID.
 */
static inline struct sock *lookup_tid(const struct tid_info *t, tid_t tid)
{
    return tid < t->ntids ? t->tid_tab[tid] : NULL;
}

/*
 * Find the socket corresponding to a server TID.
 */
static inline struct sock *lookup_stid(const struct tid_info *t, tid_t tid)
{
    struct sock *sk;

    if (tid < t->stid_base || tid >= t->stid_base + t->nstids)
	 return NULL;
    sk = STID2ENTRY(t, tid).sk;
    return ((void *)sk < (void *)t->tid_tab ||
	    (void *)sk >= (void *)&t->atid_tab[t->natids]) ? sk : NULL;
}

/*
 * Find the socket corresponding to an active-open TID.
 */
static inline struct sock *lookup_atid(const struct tid_info *t, tid_t tid)
{
    struct sock *sk;

    if (tid < t->atid_base || tid >= t->atid_base + t->natids)
	return NULL;
    sk = ATID2ENTRY(t, tid).sk;
    return ((void *)sk < (void *)t->tid_tab ||
	    (void *)sk >= (void *)&t->atid_tab[t->natids]) ? sk : NULL;
}
#endif

// Local Variables:
// c-basic-offset: 4
// End:
