/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/param.h>
#include <linux/slab.h>
#include <linux/sysctl.h>
#include <linux/toedev.h>
#include "tom_sysctl.h"
#include "tom.h"

/* This belongs in linux/sysctl.h */
#define CTL_TOE 11

/* sysctl ids for tunables */
enum {
    TOE_CONF_MAX_HOST_SNDBUF = 1,
    TOE_CONF_TX_HOLD_THRES,
    TOE_CONF_MAX_TX_PAGES,
    TOE_CONF_RX_CREDIT_THRES,
    TOE_CONF_MIN_NEWCONN_TX_PAGES,
    TOE_CONF_AUTOMODE_ON,
    TOE_CONF_ACTIVE_MSS,
    TOE_CONF_DELACK,
    TOE_CONF_MAX_CONN,
    TOE_SOFT_BACKLOG_LIMIT,
    TOE_CONF_TX_LEASE_TIME,
    TOE_CONF_OVERRIDE_TX_WSCALE,

    TOE_CONF_LAST                 // must be last
};

static struct tom_tunables default_tunable_vals = {
    .max_host_sndbuf = 32 * 1024,
    .tx_hold_thres = 0,
    .max_tx_pages = 2,
    .rx_credit_thres = 4 * 1024,
    .min_newconn_tx_pages = 2,
    .toe_automode_on = 0,
    .mss = 16384,
    .delack = 1,
    .max_conn = -1,
    .soft_backlog_limit = 0,
    .tx_lease_time = (10 * HZ) / 1000,                   // 10ms
    .tx_wscale = 0,
};

static int min_tx_pages = 2;   /* Min # of TX pages for a connection */
static int min_mss = 1;        /* Min length of TX_DATA payload */
static int min_rx_credits = 1; /* Min RX credit threshold */
static int min_tx_wscale;      /* Min TX override window scale factor */
static int max_tx_wscale = 14; /* Max TX override window scale factor */

/* Number of fields in tom_tunables */
#define NUM_TUNABLES (TOE_CONF_LAST - 1)

/*
 * Sysctl table template.  This is cloned for each TOM instance.
 */
struct tom_sysctl_table {
    struct ctl_table_header *sysctl_header;

    char tom_instance_dir_name[TOENAMSIZ + 4];
    ctl_table tunables[NUM_TUNABLES + 1];
    ctl_table tom_instance_dir[2];
    ctl_table root_dir[2];
};

static struct tom_sysctl_table tom_sysctl = {
    .tunables = {
	{
	    .ctl_name     = TOE_CONF_MAX_HOST_SNDBUF,
	    .procname     = "max_host_sndbuf",
	    .data         = &default_tunable_vals.max_host_sndbuf,
	    .maxlen       = sizeof(default_tunable_vals.max_host_sndbuf),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{
	    .ctl_name     = TOE_CONF_TX_HOLD_THRES,
	    .procname     = "tx_hold_thres",
	    .data         = &default_tunable_vals.tx_hold_thres,
	    .maxlen       = sizeof(default_tunable_vals.tx_hold_thres),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{
	    .ctl_name     = TOE_CONF_MAX_TX_PAGES,
	    .procname     = "max_tx_pages",
	    .data         = &default_tunable_vals.max_tx_pages,
	    .maxlen       = sizeof(default_tunable_vals.max_tx_pages),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec_minmax,
	    .strategy     = &sysctl_intvec,
	    .extra1       = &min_tx_pages
	},
	{
	    .ctl_name     = TOE_CONF_RX_CREDIT_THRES,
	    .procname     = "rx_credit_thres",
	    .data         = &default_tunable_vals.rx_credit_thres,
	    .maxlen       = sizeof(default_tunable_vals.rx_credit_thres),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec_minmax,
	    .strategy     = &sysctl_intvec,
	    .extra1       = &min_rx_credits
	},
	{
	    .ctl_name     = TOE_CONF_MIN_NEWCONN_TX_PAGES,
	    .procname     = "min_newconn_tx_pages",
	    .data         = &default_tunable_vals.min_newconn_tx_pages,
	    .maxlen       = sizeof(default_tunable_vals.min_newconn_tx_pages),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{
	    .ctl_name     = TOE_CONF_AUTOMODE_ON,
	    .procname     = "auto_passv_open",
	    .data         = &default_tunable_vals.toe_automode_on,
	    .maxlen       = sizeof(default_tunable_vals.toe_automode_on),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{
	    .ctl_name     = TOE_CONF_ACTIVE_MSS,
	    .procname     = "mss",
	    .data         = &default_tunable_vals.mss,
	    .maxlen       = sizeof(default_tunable_vals.mss),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec_minmax,
	    .strategy     = &sysctl_intvec,
	    .extra1       = &min_mss
	},
	{
	    .ctl_name     = TOE_CONF_DELACK,
	    .procname     = "delayed_ack",
	    .data         = &default_tunable_vals.delack,
	    .maxlen       = sizeof(default_tunable_vals.delack),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{
	    .ctl_name     = TOE_CONF_MAX_CONN,
	    .procname     = "max_conn",
	    .data         = &default_tunable_vals.max_conn,
	    .maxlen       = sizeof(default_tunable_vals.max_conn),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{
	    .ctl_name     = TOE_SOFT_BACKLOG_LIMIT,
	    .procname     = "soft_backlog_limit",
	    .data         = &default_tunable_vals.soft_backlog_limit,
	    .maxlen       = sizeof(default_tunable_vals.soft_backlog_limit),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec
	},
	{   // XXX this needs a strategy
	    .ctl_name     = TOE_CONF_TX_LEASE_TIME,
	    .procname     = "tx_lease_time",
	    .data         = &default_tunable_vals.tx_lease_time,
	    .maxlen       = sizeof(default_tunable_vals.tx_lease_time),
	    .mode         = 0644,
	    .proc_handler = &proc_doulongvec_ms_jiffies_minmax,
	},
	{
	    .ctl_name     = TOE_CONF_OVERRIDE_TX_WSCALE,
	    .procname     = "override_tx_wscale",
	    .data         = &default_tunable_vals.tx_wscale,
	    .maxlen       = sizeof(default_tunable_vals.tx_wscale),
	    .mode         = 0644,
	    .proc_handler = &proc_dointvec_minmax,
	    .strategy     = &sysctl_intvec,
	    .extra1       = &min_tx_wscale,
	    .extra2       = &max_tx_wscale
	},
    },
    .tom_instance_dir = {
	{
	    .ctl_name     = 1,
	    .procname     = tom_sysctl.tom_instance_dir_name,
	    .mode         = 0555,
	    .child        = tom_sysctl.tunables,
	},
    },
    .root_dir = {
	{
	    .ctl_name     = CTL_TOE,
	    .procname     = "toe",
	    .mode         = 0555,
	    .child        = tom_sysctl.tom_instance_dir,
	},
    }
};

/*
 * Register the sysctl table for a TOM instance associated with the supplied
 * TOE device.
 */
struct tom_sysctl_table *t1_sysctl_register(struct toedev *dev,
					    const struct tom_tunables *p)
{
    int i;
    struct tom_sysctl_table *t = kmalloc(sizeof(*t), GFP_KERNEL);

    if (!t) return NULL;

    memcpy(t, &tom_sysctl, sizeof(*t));
    snprintf(t->tom_instance_dir_name, sizeof(t->tom_instance_dir_name),
	     "%s_tom", dev->name);
    for (i = 0; i < NUM_TUNABLES; ++i) {
	t->tunables[i].data += (char *)p - (char *)&default_tunable_vals;
	t->tunables[i].de = NULL;
    }
    t->tunables[TOE_CONF_ACTIVE_MSS - 1].extra2 = &TOM_DATA(dev)->tx_max_chunk;

    t->tom_instance_dir[0].ctl_name = 2 * dev->toe_index - 1;
    t->tom_instance_dir[0].procname = t->tom_instance_dir_name;
    t->tom_instance_dir[0].child = t->tunables;
    t->tom_instance_dir[0].de = NULL;
    t->root_dir[0].child = t->tom_instance_dir;
    t->root_dir[0].de = NULL;

    t->sysctl_header = register_sysctl_table(t->root_dir, 0);
    if (!t->sysctl_header) {
	kfree(t);
	t = NULL;
    }
    return t;
}

void t1_sysctl_unregister(struct tom_sysctl_table *t)
{
    if (t) {
	unregister_sysctl_table(t->sysctl_header);
	kfree(t);
    }
}

void t1_init_tunables(struct tom_data *t, struct tom_tunables *p)
{
    *p = default_tunable_vals;

    /* Now apply device specific fixups. */
    p->mss = t->tx_max_chunk;
}

// Local Variables:
// c-basic-offset: 4
// End:

