/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  trace.c                                                                  *
 *                                                                           *
 * Description:                                                              *
 *  Chelsio 10Gb Trace buffers.                                              *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Felix Marti             <felix@chelsio.com>                      *
 *          Tina Yang               <tainay@chelsio.com>                     *
 *          Felix Marti             <felix@chelsio.com>                      *
 *          Scott Bardone           <sbardone@chelsio.com>                   *
 *          Frank DiMambro          <frank@chelsio.com>                      *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/07/09 02:47:14 $ $RCSfile: trace.c,v $ $Revision: 1.13 $ */

#include "common.h"
#include "trace.h"

#ifdef T1_TRACE
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,23)
#define PDE(inode) ((struct proc_dir_entry *) (inode)->u.generic_ip)
#endif

#include "common.h"

FILE_IDENT("@(#) $Id: trace.c,v 1.13 2005/07/09 02:47:14 sbardone Exp $");

/******************************************************************************
 * SEQ OPS
 **********/

static void *t1_trace_seq_start (struct seq_file *seq, loff_t *pos)
{
        struct t1_traceb *tb = (struct t1_traceb *)seq->private;
        struct t1_traceb_e *e = NULL;
        unsigned int start, count;

        if (tb->count > tb->nume) {
                start = tb->count & (tb->nume - 1);
                count = tb->nume;
        } else {
                start = 0;
                count = tb->count;
        }

        if (*pos < count)
                e = &tb->ep[(start + *pos) & (tb->nume - 1)];

        return e;
}

static void *t1_trace_seq_next (struct seq_file *seq, void *v, loff_t *pos)
{
        struct t1_traceb *tb = (struct t1_traceb *)seq->private;
        struct t1_traceb_e *e = v;
        unsigned int count = min(tb->count, tb->nume);

        if (++*pos < count) {
                e++;
                if ((unsigned long)e >= (unsigned long)&tb->ep[tb->nume])
                        e = tb->ep;
        } else
                e = NULL;

        return e;
}

static void t1_trace_seq_stop (struct seq_file *seq, void *v)
{
}

static int t1_trace_seq_show (struct seq_file *seq, void *v)
{
        struct t1_traceb_e *ep = v;

        seq_printf(seq, "%016llx ", (unsigned long long) ep->tsc);
        seq_printf(seq, ep->fmt, ep->param[0], ep->param[1], ep->param[2],
                ep->param[3], ep->param[4], ep->param[5]);
        seq_printf(seq, "\n");

        return 0;
}

static struct seq_operations t1_trace_seq_ops = {
	.start  = t1_trace_seq_start,
	.next   = t1_trace_seq_next,
	.stop   = t1_trace_seq_stop,
	.show   = t1_trace_seq_show
};


/******************************************************************************
 * FILE OPS
 ***********/

static int t1_trace_seq_open (struct inode *inode, struct file *file)
{
        int rc = seq_open(file, &t1_trace_seq_ops);

        if (!rc) {
                struct proc_dir_entry *dp = PDE(inode);
                struct seq_file *seq = file->private_data;

                seq->private = dp->data;
        }

        return 0;
}

static struct file_operations t1_trace_seq_fops = {
	.owner   = THIS_MODULE,
	.open    = t1_trace_seq_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
};

/******************************************************************************
 * PROC
 **********/

static int t1_trace_proc_create (char *name, struct t1_traceb *tb)
{
        struct proc_dir_entry *p;

        p = create_proc_entry(name, 0, NULL);
        if (!p)
                return -ENOMEM;

        p->proc_fops = &t1_trace_seq_fops;
        p->data = tb;

        return 0;
}

static void t1_trace_proc_remove (char *name)
{
        remove_proc_entry(name, NULL);
}


/******************************************************************************
 * TRACEBUFFER API
 *****************/

struct t1_traceb *t1_trace_alloc (char *name)
{
        struct t1_traceb *tb;
        unsigned int size;
        int ret;

        if (!name)
                goto err_no_name;

        size = sizeof (struct t1_traceb) +
                (T1_TRACE_MAX_TB_LEN * sizeof (struct t1_traceb_e));
        tb = (struct t1_traceb *) kmalloc(size, GFP_KERNEL);
        if (!tb)
                goto err_no_mem;

        tb->nume      = T1_TRACE_MAX_TB_LEN;
        tb->count     = 0;
        strcpy((char *)&tb->name, name);

        ret = t1_trace_proc_create(name, tb);
        if (ret)
                goto err_no_mem;

        return tb;

 err_no_mem:
        t1_trace_free(tb);
 err_no_name:
        return NULL;
}

void t1_trace_free (struct t1_traceb *tb)
{
        if (tb) {
                t1_trace_proc_remove(tb->name);
                kfree(tb);
        }
}
#endif /* T1_TRACE */
