/*
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Karen Xie (kxie@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * ddp.c * Direct Data Placement (DDP) API for Chelsio TOE
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/toedev.h>
#include <linux/errno.h>
#include <linux/tcp.h>
#include <net/offload.h>
#include "../cpl5_cmd.h"
#include "ulp.h"
#include "ddp.h"
#include "ulp_defs.h"
#include "../tom/defs.h"
#include "../toe_compat.h"
#include "../common.h"
#include "../chtoe.h"
#include "../regs.h"

#define PAGE_POD_SIZE_ULP1	64
#define PAGE_POD_SHIFT_ULP1	6
#define TAG_COLOR_SIZE_ULP1	6

#define PAGE_POD_SIZE_ULP0	32
#define PAGE_POD_SHIFT_ULP0	5
#define TAG_COLOR_SIZE_ULP0	5

#define PAGES_PER_PAGE_POD      4

/* ULP0 pagepod manipulation */
#define GET_PAGE_POD_VLD_ULP0(PTR)           (*((u8 *)(PTR)) & 0x80)
#define CLR_PAGE_POD_VLD_ULP0(PTR)           (*((u8 *)(PTR)) &= ~0x80)
#define SET_PAGE_POD_VLD_ULP0(PTR)           (*((u8 *)(PTR)) |= 0x80)
#define GET_PAGE_POD_COLOR_ULP0(PTR)         (*((u8 *)(PTR)) & 0x1f)
#define GET_PAGE_POD_TID_ULP0(PTR)           (htonl(*(u32 *)(((u8 *)(PTR)))) & 0xffffff)
#define SET_PAGE_POD_VLD_AND_COLOR_AND_TID_ULP0(PTR,VC,TID)  (*(u32 *)(((u8 *)(PTR))) = htonl(TID) | VC)

#define GET_PAGE_POD_MAX_ULP0(PTR)           (htonl(*(u32 *)(((u8 *)(PTR))+4)))
#define SET_PAGE_POD_MAX_ULP0(PTR,VAL)       (*(u32 *)(((u8 *)(PTR))+4) = htonl(VAL))

#define GET_PAGE_POD_POFF_ULP0(PTR)          (htonl(*(u32 *)(((u8 *)(PTR))+8)))
#define SET_PAGE_POD_POFF_ULP0(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+8) = htonl(VAL))

#define SET_PAGE_POD_REGION0_AND_PNUM0_ULP0(PTR,REG,PNUM)    (*(u32 *)(((u8 *)(PTR))+12) = htonl(PNUM) | REG)
#define SET_PAGE_POD_REGION1_AND_PNUM1_ULP0(PTR,REG,PNUM)    (*(u32 *)(((u8 *)(PTR))+16) = htonl(PNUM) | REG)
#define SET_PAGE_POD_REGION2_AND_PNUM2_ULP0(PTR,REG,PNUM)    (*(u32 *)(((u8 *)(PTR))+20) = htonl(PNUM) | REG)
#define SET_PAGE_POD_REGION3_AND_PNUM3_ULP0(PTR,REG,PNUM)    (*(u32 *)(((u8 *)(PTR))+24) = htonl(PNUM) | REG)
#define SET_PAGE_POD_REGION4_AND_PNUM4_ULP0(PTR,REG,PNUM)    (*(u32 *)(((u8 *)(PTR))+28) = htonl(PNUM) | REG)

/* ULP1 pagepod manipulation */
#define GET_PAGE_POD_VLD_ULP1(PTR)           (*((u8 *)(PTR)) & 0x1)
#define SET_PAGE_POD_VLD_ULP1(PTR)           (*((u8 *)(PTR)) = 0x1)
#define CLR_PAGE_POD_VLD_ULP1(PTR)           (*((u8 *)(PTR)) &= ~0x1)

#define GET_PAGE_POD_TID_ULP1(PTR)           ((htonl(*(u32 *)(((u8 *)(PTR))+1)) & 0xffffff00)>>8)
#define SET_PAGE_POD_TID_ULP1(PTR,VAL)       (*(u32 *)(((u8 *)(PTR))+1) |= htonl(((VAL<<8)&0xffffff00)))

#define GET_PAGE_POD_COLOR_ULP1(PTR)         (*((u8 *)(((u8 *)(PTR))+7)) & 0x3F)
#define SET_PAGE_POD_COLOR_ULP1(PTR,VAL)     (*((u8 *)(((u8 *)(PTR))+7)) |= (VAL&0x3F))

#define SET_PAGE_POD_PAGESIZE_ULP1(PTR,VAL)  (*(u8 *)(((u8 *)(PTR))+4) |= (((VAL<<6)&0xc0)))

#define GET_PAGE_POD_TAG_ULP1(PTR)           ((htonl(*(u32 *)(((u8 *)(PTR))+4)) & 0x3fffffc0)>>6)
#define SET_PAGE_POD_TAG_ULP1(PTR,VAL)       (*(u32 *)(((u8 *)(PTR))+4) |= htonl(((VAL<<6)&0x3fffffc0)))

#define SET_PAGE_POD_TAG_COLOR_ULP1(PTR,VAL) (*(u32 *)(((u8 *)(PTR))+4) |= htonl(VAL))

#define GET_PAGE_POD_MAX_ULP1(PTR)           ((htonl(*(u32 *)((u8 *)(PTR)+8))))
#define SET_PAGE_POD_MAX_ULP1(PTR,VAL)       (*(u32 *)(((u8 *)(PTR))+8) |= htonl(VAL))

#define GET_PAGE_POD_POFF_ULP1(PTR)          ((htonl(*(u32 *)((u8 *)(PTR)+12))))
#define SET_PAGE_POD_POFF_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+12) |= htonl(VAL))

#define SET_PAGE_POD_PAH0_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+24) |= htonl(VAL))
#define SET_PAGE_POD_PAL0_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+28) |= htonl(VAL))
#define SET_PAGE_POD_PAH1_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+32) |= htonl(VAL))
#define SET_PAGE_POD_PAL1_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+36) |= htonl(VAL))
#define SET_PAGE_POD_PAH2_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+40) |= htonl(VAL))
#define SET_PAGE_POD_PAL2_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+44) |= htonl(VAL))
#define SET_PAGE_POD_PAH3_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+48) |= htonl(VAL))
#define SET_PAGE_POD_PAL3_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+52) |= htonl(VAL))
#define SET_PAGE_POD_PAH4_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+56) |= htonl(VAL))
#define SET_PAGE_POD_PAL4_ULP1(PTR,VAL)      (*(u32 *)(((u8 *)(PTR))+60) |= htonl(VAL))

// CPL_PCMD pcmd_cmd fields
#define         PM_WRITE_CMD                    1
#define         PM_TID_CMD                      4
#define         PM_ULP_CMD                      7

#define SET_CPL_PCMD_MODE_AND_FBCK(PTR,VAL)     (*((__u8 *)(PTR)) = (VAL))
#define SET_CPL_PCMD_CMD(PTR,VAL)               (*(((__u8 *)(PTR))+1) = ((VAL<<4)&0xF0)|((*(((__u8 *)(PTR))+1))&0xF))
#define SET_CPL_PCMD_MEM(PTR)                   (*(((__u8 *)(PTR))+1) |= 0x8)
#define CLR_CPL_PCMD_MEM(PTR)                   (*(((__u8 *)(PTR))+1) &= ~(0x8))
#define SET_CPL_PCMD_BYPASS(PTR)                (*(((__u8 *)(PTR))+1) |= 0x4)
#define CLR_CPL_PCMD_BYPASS(PTR)                (*(((__u8 *)(PTR))+1) &= ~(0x4))
#define SET_CPL_PCMD_EOP(PTR)                   (*(((__u8 *)(PTR))+1) |= 0x2)
#define CLR_CPL_PCMD_EOP(PTR)                   (*(((__u8 *)(PTR))+1) &= ~(0x2))
#define SET_CPL_PCMD_EOI(PTR)                   (*(((__u8 *)(PTR))+1) |= 0x1)
#define CLR_CPL_PCMD_EOI(PTR)                   (*(((__u8 *)(PTR))+1) &= ~(0x1))
#define SET_CPL_PCMD_ADDRESS(PTR,VAL)           (*(__u32 *)(((__u8 *)(PTR))+2) = htonl(VAL))
#define SET_CPL_PCMD_LEN(PTR,VAL)               (*(__u16 *)(((__u8 *)(PTR))+6) = htons(VAL))

/* local function declarations */
static int dmt_update(struct toedev *tdev, chtoe_ddp_map_table *dmt, int idx, int tlen);
static int setup_pagepod_ulp0(struct toedev *tdev, chtoe_ddp_map_table *dmt, 
	int sgcnt, struct scatterlist *sg, __u32 offset, __u32 tid, 
        int maplen, __u8 psize,  __u32 *tag);
static int setup_pagepod_ulp1(struct toedev *tdev, chtoe_ddp_map_table *dmt, 
	int sgcnt, struct scatterlist *sg, __u32 offset, __u32 tid, 
        int maplen, __u8 psize,  __u32 *tag);

/* only support DDP page size of 4K, 8K, 16K */
__u32 ddp_page_size[3] = {4096, 8192, 16384};
__u8  ddp_page_code[3] = {PAGE_POD_PAGE_SIZE_4K, PAGE_POD_PAGE_SIZE_8K, PAGE_POD_PAGE_SIZE_16K};
__u8  ddp_page_shift[3] = {12, 13, 14};	
/*
 * init_ddp: allocate and initialize software structures for a DDP map table
 */
void* t1_init_ddp_table(struct ulp_info *ulp)
{
    chtoe_ddp_map_table *dmt = NULL;
    __u32		nbytes = 65536;
    __u32		mask;
    int 		i;

    dmt = kmalloc(sizeof(chtoe_ddp_map_table), GFP_KERNEL);
    if (!dmt) return NULL;

    if (ulp->version == 0) {
	dmt->pp_size = PAGE_POD_SIZE_ULP0;
	dmt->pp_shift = PAGE_POD_SHIFT_ULP0;
        dmt->setup_pagepod = setup_pagepod_ulp0;
    } else {
	dmt->pp_size = PAGE_POD_SIZE_ULP1;
	dmt->pp_shift = PAGE_POD_SHIFT_ULP1;
        dmt->setup_pagepod = setup_pagepod_ulp1;
    }

    /* fill in the fields */
    dmt->dmt_size = nbytes/dmt->pp_size;
    dmt->dmt_tag_id_mask = (dmt->dmt_size << 1) - 1;
    dmt->dmt_tag_color_mask = (1 << dmt->pp_shift) - 1;
    dmt->dmt_tag_mask = (dmt->dmt_tag_id_mask << dmt->pp_shift) | dmt->dmt_tag_color_mask;
    for (i = 0, mask = dmt->dmt_tag_id_mask; mask; i++) {
        mask &= ~(1 << i);
    }
    dmt->tag_id_bits = i;
    dmt->tag_color_bits = dmt->pp_shift;

    dmt->dmt_off_map_tag = 0;
    dmt->dmt_map_tag_invalid = nbytes;

    /* clear/invalidate memory region table */
    for (i=0; i < MAX_MEMORY_REGIONS; i++)
	dmt->dmt_memregion[i] = 1;

    /* allocate page pod table */
    dmt->dmt_table = kmalloc(nbytes, GFP_KERNEL);
    if (!dmt->dmt_table) {
	kfree(dmt);
	return NULL;
    }
    memset(dmt->dmt_table, 0, nbytes);

    /* allocate page pod color table */
    dmt->dmt_color = kmalloc(dmt->dmt_size, GFP_KERNEL);
    if (!dmt->dmt_color) {
	kfree(dmt->dmt_table);
	kfree(dmt);
	return NULL;
    }
    memset(dmt->dmt_color, 0, dmt->dmt_size);

    spin_lock_init(&dmt->dmt_lock);

    return dmt;
}

/*
 * free_ddp: free software structures for a DDP map table
 */
void t1_free_ddp(void *arg)
{
    chtoe_ddp_map_table *dmt = arg;
    if (!dmt)	return;
    if (dmt->dmt_color) kfree(dmt->dmt_color);
    if (dmt->dmt_table) kfree(dmt->dmt_table);
    kfree(dmt);
}

static int ddp_find_memory_region(chtoe_ddp_map_table *dmt, __u32 memreg)
{
    int i;
    unsigned long flags;
    spin_lock_irqsave(&dmt->dmt_lock, flags);
    for (i=0; i < MAX_MEMORY_REGIONS; i++) {
        if (dmt->dmt_memregion[i] == memreg) {
            spin_unlock_irqrestore(&dmt->dmt_lock, flags);
	    return i;
        }
    }
    spin_unlock_irqrestore(&dmt->dmt_lock, flags);
    return -1;
}

static int ddp_add_memory_region(struct toedev *tdev, chtoe_ddp_map_table *dmt, 
		__u32 memreg)
{
    int 		i;
    unsigned long	flags;
#if 0 /* ulp0: PCMD */
    struct sk_buff *skb;
    struct cpl_pcmd	*pcmd;
    __u8		*p;
#endif
    struct adapter	*adapter = tdev->priv;
	
    spin_lock_irqsave(&dmt->dmt_lock, flags);
    for (i=0; i < MAX_MEMORY_REGIONS; i++)
        if (dmt->dmt_memregion[i] == 1) break;

    if (i == MAX_MEMORY_REGIONS) {
        spin_unlock_irqrestore(&dmt->dmt_lock, flags);
        return -1;
    }

    dmt->dmt_memregion[i] = memreg;
    spin_unlock_irqrestore(&dmt->dmt_lock, flags);

    /* use PIOs for setting up the memory region */
    t1_write_reg_4(adapter, A_ULP_HREG_INDEX, i);
    t1_write_reg_4(adapter, A_ULP_HREG_DATA, memreg);

#if 0
    /* ulp0: PCMD can also be used for setting up memory region */
    skb = alloc_skb_nofail(sizeof(struct cpl_pcmd));
    pcmd = (struct cpl_pcmd *)__skb_put(skb, sizeof(struct cpl_pcmd));
    /* memset(pcmd, 0, sizeof(struct cpl_pcmd)); */
    pcmd->dlen_in = 0;
    pcmd->dlen_out = 0;
    pcmd->pcmd_parm[0] = 0;
    pcmd->pcmd_parm[1] = 0;
    OPCODE_TID(pcmd) = htonl(MK_OPCODE_TID(CPL_PCMD, 0));
    p = (__u8 *)&(pcmd->pcmd_parm[0]);
    SET_CPL_PCMD_CMD(p, PM_ULP_CMD);
    /* CLR_CPL_PCMD_MEM(p); */
    SET_CPL_PCMD_EOP(p);
    SET_CPL_PCMD_EOI(p);
    SET_CPL_PCMD_ADDRESS(p,memreg);
    SET_CPL_PCMD_LEN(p,i);
     
    toe_send(tdev, skb);
#endif

    return 0;
}

static int setup_pagepod_ulp0(struct toedev *tdev, chtoe_ddp_map_table *dmt, 
	int sgcnt, struct scatterlist *sg, __u32 offset, __u32 tid, 
        int maplen, __u8 psize,  __u32 *tag)
{
    int		i, rv, index;
    unsigned long flags;
    __u32 	npods, color, pshift = ddp_page_shift[psize];
    __u32 	memreg, prev_memreg = 1, memregnum, pnum = 0;
    __u8	*p;

    /* search table for contiguous span of free entries */
    npods = (sgcnt + PAGES_PER_PAGE_POD - 1) / PAGES_PER_PAGE_POD;
    spin_lock_irqsave(&dmt->dmt_lock, flags);
    for (index = 0; index <= (dmt->dmt_size - npods); ) {
        p = &(dmt->dmt_table[index * PAGE_POD_SIZE_ULP0]);
	for (i=0; i < npods; i++, p += PAGE_POD_SIZE_ULP0)
            /* if entry is used */
            if (GET_PAGE_POD_VLD_ULP0(p)) break;
	if (i == npods) break;
	index += i + 1;
    }
    if (index > (dmt->dmt_size - npods)) {
        /* can't find contiguous entries */
        spin_unlock_irqrestore(&dmt->dmt_lock, flags);
        return -ENOSPC;
    }
    dmt->dmt_off_map_tag--;
    p = &dmt->dmt_table[index * PAGE_POD_SIZE_ULP0];
    for (i=0; i < npods; i++, p += PAGE_POD_SIZE_ULP0) {
	memset (p, 0, PAGE_POD_SIZE_ULP0);
        SET_PAGE_POD_VLD_ULP0(p);
    }
    spin_unlock_irqrestore(&dmt->dmt_lock, flags);

    /* update software page pods */
    color = dmt->dmt_color[index];
    color = (color + 1) & ((1 << TAG_COLOR_SIZE_ULP0) - 1);
    dmt->dmt_color[index] = color;

    *tag = (((__u32)index) << TAG_COLOR_SIZE_ULP0) | color;

#ifdef __ULP_DEBUG__
    printk(KERN_INFO "ulp0_ddp_map: tid 0x%x, color 0x%x, index 0x%x -> tag 0x%x.\n", tid, color, index, *tag); 
#endif

    prev_memreg = 1;
    memregnum = -1;

    p = &dmt->dmt_table[index * PAGE_POD_SIZE_ULP0];
    for (i=0; i < npods; i++, p += PAGE_POD_SIZE_ULP0) {
	SET_PAGE_POD_VLD_AND_COLOR_AND_TID_ULP0(p, (0x80 | color), tid);
        SET_PAGE_POD_MAX_ULP0(p, maplen);
        SET_PAGE_POD_POFF_ULP0(p, offset);

#define SET_MEMREGNUM_AND_PNUM(X) \
    do { \
	if ((X) >= sgcnt) { \
	    memreg = 1; \
	    memregnum = 0; \
	    pnum = 1; \
	    break; \
	} \
	memreg = 0; \
	if (memreg != prev_memreg) { \
	    memregnum = ddp_find_memory_region(dmt, memreg); \
	    if (memregnum < 0) { \
                /* XXX error */ \
            } \
	} \
	/* XXX page shift depends on variable page size */ \
	pnum = sg[X].dma_address >> pshift; \
    } while (0)

        SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 0);
        SET_PAGE_POD_REGION0_AND_PNUM0_ULP0(p, memregnum, pnum);

        SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 1);
        SET_PAGE_POD_REGION1_AND_PNUM1_ULP0(p, memregnum, pnum);

        SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 2);
        SET_PAGE_POD_REGION2_AND_PNUM2_ULP0(p, memregnum, pnum);

        SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 3);
        SET_PAGE_POD_REGION3_AND_PNUM3_ULP0(p, memregnum, pnum);

        SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 4);
        SET_PAGE_POD_REGION4_AND_PNUM4_ULP0(p, memregnum, pnum);

#undef SET_MEMREGNUM_AND_PNUM
    }

    rv = dmt_update(tdev, dmt, index, PAGE_POD_SIZE_ULP0 * npods);

    return rv;
}

static int setup_pagepod_ulp1(struct toedev *tdev, chtoe_ddp_map_table *dmt, 
	int sgcnt, struct scatterlist *sg, __u32 offset, __u32 tid, 
        int maplen, __u8 psize, __u32 *tag)
{
    int		i, rv, index;
    unsigned long flags;
    __u32 	npods, color, psize_code = ddp_page_code[psize];
    __u32 	pnum = 0;
    __u8	*p;

    /* search table for contiguous span of free entries */
    npods = (sgcnt + PAGES_PER_PAGE_POD - 1) / PAGES_PER_PAGE_POD;
    spin_lock_irqsave(&dmt->dmt_lock, flags);
    for (index = 0; index <= (dmt->dmt_size - npods); ) {
	p = &(dmt->dmt_table[index * PAGE_POD_SIZE_ULP1]);
	for (i=0; i < npods; i++, p += PAGE_POD_SIZE_ULP1)
            /* if entry is used */
            if (GET_PAGE_POD_VLD_ULP1(p)) break;
	if (i == npods) break;
	index += i + 1;
    }
    if (index > (dmt->dmt_size - npods)) {
        /* can't find contiguous entries */
        spin_unlock_irqrestore(&dmt->dmt_lock, flags);
        return -ENOSPC;
    }
    p = &dmt->dmt_table[index * PAGE_POD_SIZE_ULP1];
    for (i=0; i < npods; i++, p += PAGE_POD_SIZE_ULP1) {
        memset (p, 0, PAGE_POD_SIZE_ULP1);
        SET_PAGE_POD_VLD_ULP1(p);
    }
    dmt->dmt_off_map_tag--;
    spin_unlock_irqrestore(&dmt->dmt_lock, flags);

    /* update software page pods */
    color = dmt->dmt_color[index];
    color = (color + 1) & ((1 << TAG_COLOR_SIZE_ULP1) - 1);
    dmt->dmt_color[index] = color;

    *tag = ((((__u32)index) << TAG_COLOR_SIZE_ULP1) & (~((1 << dmt->pp_shift)-1)) ) | color;
#ifdef __ULP_DEBUG__
    printk(KERN_INFO "ulp1_ddp_map: tid 0x%x, color 0x%x, index 0x%x -> tag 0x%x, poff 0x%x, maplen 0x%x.\n", 
       tid, color, index, *tag, offset, maplen);
#endif

    p = &dmt->dmt_table[index * PAGE_POD_SIZE_ULP1];
    for (i=0; i < npods; i++, p += PAGE_POD_SIZE_ULP1) {
        SET_PAGE_POD_TID_ULP1(p, tid);
        SET_PAGE_POD_PAGESIZE_ULP1(p, psize_code);
        SET_PAGE_POD_TAG_ULP1(p, index);
        SET_PAGE_POD_COLOR_ULP1(p, color);
        SET_PAGE_POD_POFF_ULP1(p, offset);
        SET_PAGE_POD_MAX_ULP1(p, maplen);

/* need modification for 64bit machine */
#define SET_MEMREGNUM_AND_PNUM(X) \
                do { \
                        if ((X) >= sgcnt) { \
                                pnum = 0; \
                                break; \
                                } \
                        pnum = sg[X].dma_address; \
                } while (0)

         SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 0);
         /*SET_PAGE_POD_PAH0_ULP1(p, memregnum);*/
         SET_PAGE_POD_PAL0_ULP1(p, pnum);

         SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 1);
         /*SET_PAGE_POD_PAH1_ULP1(p, memregnum);*/
         SET_PAGE_POD_PAL1_ULP1(p, pnum);

         SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 2);
         /*SET_PAGE_POD_PAH1_ULP1(p, memregnum); */
         SET_PAGE_POD_PAL2_ULP1(p, pnum);

         SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 3);
         /*SET_PAGE_POD_PAH1_ULP1(p, memregnum); */
         SET_PAGE_POD_PAL3_ULP1(p, pnum);

         SET_MEMREGNUM_AND_PNUM((i * PAGES_PER_PAGE_POD) + 4);
         /*SET_PAGE_POD_PAH1_ULP1(p, memregnum); */
         SET_PAGE_POD_PAL4_ULP1(p, pnum);
    }
#undef SET_MEMREGNUM_AND_PNUM
    rv = dmt_update(tdev, dmt, index, PAGE_POD_SIZE_ULP1 * npods);

    return rv;
}

/* map the given scattergather array to a DDP tag */
int t1_tom_ddp_map(struct sock *sk, int sgcnt, struct scatterlist *sg, 
	__u32 offset, __u32 *tag, __u8 psize, void *priv)
{
    struct toedev	*tdev = TOE_DEV(tcp_sk(sk));
    struct ulp_info	*ulp;
    chtoe_ddp_map_table *dmt;
    struct adapter	*adapter;
    tid_t		tid = TID(tcp_sk(sk));
    __u32 		memreg, prev_memreg = 1;
    int 		i, maplen = 0;
    int 		rv;
    __u32		pmask;

    if (!sk || !tdev) {
        /* printk(KERN_ERR "t1_ddp_map: sk 0x%p, tdev 0x%p!\n", sk, tdev); */
        return -EINVAL;
    }
    
    /* only support ddp page size of 4K, 8K, and 16K */
    if (psize > 2) {
        printk(KERN_ERR "t1_ddp_map: sk 0x%p, psize %u > 2!\n", sk, psize);
        return -EINVAL;
    }
    pmask = ddp_page_size[psize] - 1;

    ulp = ULPDATA(tdev);
    dmt = ulp->dmt;
    adapter = tdev->priv;

    *tag = ((dmt->dmt_off_map_tag++ & dmt->dmt_tag_mask) | dmt->dmt_map_tag_invalid);

    if ((sgcnt <= 0) || !sg) return -EINVAL;

    if (pci_map_sg(adapter->pdev, sg, sgcnt, PCI_DMA_FROMDEVICE) != sgcnt)
        return -EINVAL;

    /* confirm sg list OK */
    for (i=0; i < sgcnt; i++) {
        /* all addrs except first need to be page-aligned */
        if (((__u32)sg[i].dma_address & pmask)) { 
            printk(KERN_ERR "ddp_map: non-initial sg entry address not page-aligned 0x%llx\n",
		(unsigned long long) sg[i].dma_address); 
            return -EINVAL; 
        } 
        /* all lengths except last need to end page-aligned */ 
        if ((i>0 && i<(sgcnt-1)) && (sg[i].length & pmask)) {
            printk(KERN_ERR "ddp_map: non-final sg entry length not page-aligned 0x%x\n", sg[i].length);
            return -EINVAL;
        }
        maplen += sg[i].length;

        /* allocate any new region table entries */
        if (ulp->version == 0) {
            memreg = 0; 
            if ( (memreg != prev_memreg) &&
	  	 (ddp_find_memory_region(dmt, memreg) < 0) &&
		 (ddp_add_memory_region(tdev, dmt, memreg) < 0)) {
                printk(KERN_ERR "ddp_map: unable to add new memory region 0x%x\n", memreg);
                return -ENOSPC;
           }
           prev_memreg = memreg;
        }
    }

    rv = dmt->setup_pagepod(tdev, dmt, sgcnt, sg, offset, tid, maplen, psize, tag);
    return(rv);
}

static int dmt_update(struct toedev *tdev, chtoe_ddp_map_table *dmt, int idx, int tlen)
{
    __u8 	*sp,*dp;
    __u8	*dmtp;
    int i, pcmd_cnt;
    int index, maplen;

    dmtp = dmt->dmt_table;

#define MAX_PCMD_SHIFT	13
#define MAX_PCMD_SIZE	0x2000
#define MAX_PCMD_MASK   0x1fff

    pcmd_cnt = (tlen>>MAX_PCMD_SHIFT) + ((tlen&MAX_PCMD_MASK)? 1 : 0);
    for (i = 0; i < pcmd_cnt; i++) {
	struct sk_buff *skb;
	struct cpl_pcmd	*pcmd;
        __u8	*p;

        index = idx + ((i << MAX_PCMD_SHIFT) >> dmt->pp_shift);
	maplen = min(tlen, MAX_PCMD_SIZE);
	tlen -= maplen;

	skb = alloc_skb_nofail(sizeof(struct cpl_pcmd)+maplen);
	pcmd = (struct cpl_pcmd *)__skb_put(skb, sizeof(struct cpl_pcmd));
        pcmd->dlen_out = 0;
        pcmd->pcmd_parm[0] = 0;
        pcmd->pcmd_parm[1] = 0;
        /* memset(pcmd, 0, sizeof(struct cpl_pcmd)); */
	OPCODE_TID(pcmd) = htonl(MK_OPCODE_TID(CPL_PCMD, 0));
	pcmd->dlen_in = htons((__u16)maplen);	
        p = (__u8 *)pcmd->pcmd_parm;
        SET_CPL_PCMD_CMD(p,PM_WRITE_CMD);
        SET_CPL_PCMD_MEM(p);
        SET_CPL_PCMD_EOP(p);
        SET_CPL_PCMD_EOI(p);
        SET_CPL_PCMD_ADDRESS(p,index * dmt->pp_size);
        SET_CPL_PCMD_LEN(p,maplen);

        /* send dmt table */
	sp = &dmtp[(index) * dmt->pp_size];
  	dp = (__u8 *)__skb_put(skb, maplen);
	memcpy (dp, sp, maplen);

	toe_send(tdev, skb);
    }
    return 0;
}

/*
 * unmap the scattergather memory for connection tid mapped to tag
 */
int t1_tom_ddp_unmap(struct sock *sk, __u32 tag, __u32 psize, struct ulpdev_data *udata)
{
    struct toedev	*tdev;
    struct ulp_info	*ulp;
    chtoe_ddp_map_table *dmt;
    tid_t		tid = 0;
    int			i;
    unsigned long 	flags;

    __u32 index, color, color_pp, tid_pp,poff, maplen;
    __u32 npages, npods;
    __u8 *p;

    if (!udata || !udata->privdata) {
        printk(KERN_ERR "ddp_unmap: sk 0x%p, data 0x%p, NULL.\n", sk, udata);
        return -EINVAL;
    }

    ulp = udata->privdata;
    tid = udata->tid;
    tdev = ulp->tdev;

    dmt = ulp->dmt;
    if (!dmt) {
        printk(KERN_ERR "ddp_unmap: sk 0x%p, tdev=0x%p. tid=0x%x, cannot find dmt.\n", sk, tdev, tid);
        return -EINVAL;
    }
    /* mask out alias bits */
    tag &= dmt->dmt_tag_mask;

    index = (tag & (~((1 << dmt->pp_shift) - 1)) ) >> dmt->pp_shift;
    color = tag & ((1 << dmt->pp_shift) - 1);

    /* confirm tag OK */
    if (index > dmt->dmt_size) { 
        printk(KERN_ERR "ddp_unmap: index out of range 0x%x, tag=0x%x\n", index, tag);
	return -EINVAL;
    }
  
    p = &dmt->dmt_table[index * dmt->pp_size];
    if (ulp->version) {
        /* ulp 1 */
        tid_pp = GET_PAGE_POD_TID_ULP1(p);
        color_pp = GET_PAGE_POD_COLOR_ULP1(p);
        maplen = GET_PAGE_POD_MAX_ULP1(p);
        poff = GET_PAGE_POD_POFF_ULP1(p);
    } else {
        /* ulp 0 */
        tid_pp = GET_PAGE_POD_TID_ULP0(p);
        color_pp = GET_PAGE_POD_COLOR_ULP0(p);
        maplen = GET_PAGE_POD_MAX_ULP0(p);
        poff = GET_PAGE_POD_POFF_ULP0(p);
    }

    npages = (poff + maplen + psize - 1)/psize;
    npods = (npages + PAGES_PER_PAGE_POD - 1) / PAGES_PER_PAGE_POD;

    if (color != color_pp) {
        printk(KERN_ERR "ddp_unmap: color mismatch, exp 0x%x, tag 0x%x, tid 0x%x, color 0x%x, maplen 0x%x, poff 0x%x, index 0x%x.\n",
		color, tag, tid_pp, color_pp, maplen, poff, index);
	return -EINVAL;
    }
    if (tid != tid_pp) {
        printk(KERN_ERR "ddp_unmap: tid mismatch, exp 0x%x, tag 0x%x, tid 0x%x, color 0x%x, maplen 0x%x, poff 0x%x, index 0x%x.\n",
		tid, tag, tid_pp, color_pp, maplen, poff, index);
	return -EINVAL;
    }

    /* free up table entries */
    if ((index + npods) > dmt->dmt_size) {
	printk(KERN_ERR "ddp_unmap: mapping extends beyond table? 0x%x 0x%x\n",
			index, npods);
	/* just free what's in the table... */
	npods = dmt->dmt_size - index;
    }

    spin_lock_irqsave(&dmt->dmt_lock, flags);
    for (i=0; i < npods; i++) {
	/* XXX debug check poff/max/color/tid */
	p = &dmt->dmt_table[(index+i) * dmt->pp_size];
        memset (p, 0, dmt->pp_size);
	/*CLR_PAGE_POD_VLD(p);*/
    }
    spin_unlock_irqrestore(&dmt->dmt_lock, flags);

    /* send cpl commands to update page pod table entries */
    dmt_update(tdev, dmt, index, dmt->pp_size * npods);

#ifdef __ULP_DEBUG__
    printk(KERN_INFO "ddp_unmap: tag 0x%x, color 0x%x. index %u, tid 0x%x.\n", tag, color, index, tid); 
#endif

    return 0;
}

int t1_tom_ddp_dump(void *priv, __u32 index, __u32 npods) 
{
    struct ulp_info	*ulp;
    struct toedev	*tdev;
    chtoe_ddp_map_table *dmt;
    int 		i, j;
    char		buffer[512], *buf;

    if (!priv) {
        printk(KERN_ERR "ddp_dump: data NULL.\n");
        return -EINVAL;
    }
    ulp = (struct ulp_info *)priv;
    tdev = ulp->tdev;
    dmt = ulp->dmt;

    /* confirm tag OK */
    if (index > dmt->dmt_size) { 
        printk(KERN_ERR "ddp_dump: index out of range 0x%x\n", index);
	return -EINVAL;
    }

    printk(KERN_INFO "ddp_dump: index=0x%x, npods=%u.\n", index, npods);

    for (i = 0; i < npods; i++) {
        __u32 *p;
	p = (__u32 *)&dmt->dmt_table[(index+i) * dmt->pp_size];
        buf = buffer;
        for (j = 0; j < (dmt->pp_size/4); p++, buf += 11) {
            sprintf (buf, "0x%08x ", *p);
        }
        *buf = '\0';
        printk(KERN_INFO "%s\n", buffer);
    }

    return 0;
}

/*
 * return tag value with alias bits cleared
 */
__u32 chtoe_ddp_tag_no_alias(struct sock *sk, __u32 tag)
{
    struct toedev	*tdev = TOE_DEV(tcp_sk(sk));
    struct ulp_info	*ulp;
    chtoe_ddp_map_table *dmt;

    if (!tdev) {
        /* connection gone */
        /* printk(KERN_ERR "ddp_tag_next_alias: sk=0x%p, tdev 0x%p, 0x%p.\n", sk, tdev, priv); */
        return -EINVAL;
    }
    ulp = ULPDATA(tdev);
    dmt = ulp->dmt;
    return (tag & dmt->dmt_tag_mask);
}

/*
 * return tag value with alias bits "incremented"
 */
int t1_tom_ddp_tag_nextalias(struct sock *sk, __u32 *tag)
{
    struct toedev	*tdev = TOE_DEV(tcp_sk(sk));
    struct ulp_info	*ulp;
    chtoe_ddp_map_table *dmt;

    if (!tdev) {
        /* connection gone */
        /* printk(KERN_ERR "ddp_tag_next_alias: sk=0x%p, tdev 0x%p, 0x%p.\n", sk, tdev, priv); */
        return -EINVAL;
    }
    ulp = ULPDATA(tdev);
    dmt = ulp->dmt;
    *tag = *tag + (dmt->dmt_tag_mask + 1);
    return 0;
}


EXPORT_SYMBOL(t1_tom_ddp_dump);
EXPORT_SYMBOL(t1_tom_ddp_map);
EXPORT_SYMBOL(t1_tom_ddp_unmap);
EXPORT_SYMBOL(t1_tom_ddp_tag_nextalias);
