/*
 * This file implements the Chelsio CPL5 message processing.
 *
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Karen Xie (kxie@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/toedev.h>
#include <net/tcp.h>
#include <net/offload.h>
#include <net/route.h>
#include <asm/atomic.h>
#include <asm/unaligned.h>
#include "../common.h"
#include "../chtoe.h"
#include "../toe_compat.h"
#include "../tom/defs.h"
#include "../tom/tom.h"
#include "../cpl5_cmd.h"
#include "../toe_ctl_defs.h"
#include "../tom/cpl5_defs.h"
#include "../tom/tcb.h"
#include "../tom/l2t.h"
#include "ddp.h"
#include "ulp.h"
#include "ulp_defs.h"

/* do not access the cpl msg part in this function! */
void ulp_new_rx_data (struct sock *sk, struct sk_buff *skb)
{
    struct toedev       *tdev = TOE_DEV(tcp_sk(sk));
    struct ulp_info     *ulp = ULPDATA(tdev);
    struct tcp_opt 	*tp = tcp_sk(sk);
    struct sk_buff	*skb_udata = ULP_DATASKB(tp);
    struct skb_ulpinfo  *skcb = (struct skb_ulpinfo *)SKB_ULPINFO(skb);

    /* connection transit from ULP mode to TOE mode */
    if (ulp->version) {
        skcb->mode = 0;
        skcb->data_ready = 1;
        ULP_MODE(tp) = ULP_MODE_NONE;
        if (skb_udata) {
            __skb_queue_tail(&sk->sk_receive_queue, skb_udata);
        }
        __skb_queue_tail(&sk->sk_receive_queue, skb);
        ULP_DATASKB(tp) = NULL;
        ULP_DATACNT(tp) = 0;
    } else {
        ULP_DATACNT(tp)++;
        if (skb_udata) {
            __skb_queue_tail(&sk->sk_receive_queue, skb_udata);
        }
        ULP_DATASKB(tp) = skb;
        skcb->data_ready = 0;
    } 
}

/*
 * Handler for RX_DATA_DDP CPL messages.
 */
static void ulp_proc_ddp_status (struct sk_buff *skb, 
				struct cpl_rx_data_ddp *cpl, 
				__u8 data_rcvd, __u8 ulp_ver)
{
    __u32 		val;
    struct skb_ulpinfo	*skcb = (struct skb_ulpinfo *)SKB_ULPINFO(skb);

#define RX_DDP_STATUS_PAD_ERR                0x8
#define RX_DDP_STATUS_HDR_CRC_ERR            0x10
#define RX_DDP_STATUS_DATA_CRC_ERR           0x20
#define RX_DDP_STATUS_TAG_ERR                0x40
#define RX_DDP_STATUS_ULIMIT_ERR             0x80
#define RX_DDP_STATUS_OFFSET_ERR             0x100
#define RX_DDP_STATUS_COLOR_ERR              0x200
#define RX_DDP_STATUS_TID_ERR                0x400
#define RX_DDP_STATUS_PAGEPOD_ERR            0x800

#define RX_DDP_STATUS_DDP_ERROR_MASK_ULP1 	\
	(RX_DDP_STATUS_ULIMIT_ERR | RX_DDP_STATUS_OFFSET_ERR |\
         RX_DDP_STATUS_COLOR_ERR | RX_DDP_STATUS_TID_ERR | \
         RX_DDP_STATUS_PAGEPOD_ERR | RX_DDP_STATUS_TAG_ERR)

#define RX_DDP_STATUS_DDP_ERROR_MASK_ULP0 	\
	(RX_DDP_STATUS_ULIMIT_ERR | RX_DDP_STATUS_OFFSET_ERR |\
         RX_DDP_STATUS_COLOR_ERR | RX_DDP_STATUS_TID_ERR | \
         RX_DDP_STATUS_PAGEPOD_ERR)

    /* save the ddp status */
    val = ntohs(get_unaligned(&cpl->ddp_status));

    skcb->ddp_status = 0; 
    if ( val & RX_DDP_STATUS_HDR_CRC_ERR) 
        skcb->ddp_status |= DDP_STATUS_HCRC_ERROR;
    if ( val & RX_DDP_STATUS_DATA_CRC_ERR) 
        skcb->ddp_status |= DDP_STATUS_DCRC_ERROR;

    if (ulp_ver) {
        if ( val & RX_DDP_STATUS_PAD_ERR) 
            skcb->ddp_status |= DDP_STATUS_PAD_ERROR;
    }

#ifdef __ULP_DEBUG__
    if (skcb->ddp_status) {
        printk (KERN_INFO "ddp status 0x%x -> 0x%x, skb 0x%p.\n", val, skcb->ddp_status, skb);
    }
#endif
    if (!data_rcvd) skcb->ddp_status |= DDP_STATUS_DATA_DDPED;

    /* save the data crc */
    skcb->data_digest = ntohl(get_unaligned(&cpl->ulp_crc));
    /* save the pdu len */
    skcb->pdu_length = ntohl(get_unaligned(&cpl->len));

    skcb->mode = 1;
    skcb->data_ready = 1;

    skcb->data_gap = 0;
}

void ulp_new_rx_data_ddp(struct sock *sk, struct sk_buff *skb)
{
    struct toedev       	*tdev = TOE_DEV(tcp_sk(sk));
    struct ulp_info     	*ulp = ULPDATA(tdev);
    struct tcp_opt 		*tp = tcp_sk(sk);
    struct cpl_rx_data_ddp 	*hdr = cplhdr(skb);
    struct sk_buff		*skb1 = ULP_DATASKB(tp);

    if ( skb1 ) {
        ulp_proc_ddp_status (skb1, hdr, (ULP_DATACNT(tp) > 1), ulp->version);
        tp->rcv_nxt = ntohl(get_unaligned(&hdr->seq)) + 
                      ntohl(get_unaligned(&hdr->len));
        tp->ack.lrcvtime = tcp_time_stamp;
        __skb_queue_tail(&sk->sk_receive_queue, skb1);
	ULP_DATASKB(tp) = NULL;
        ULP_DATACNT(tp) = 0;
        sk->sk_data_ready(sk, 0);
    } else {
        printk(KERN_ERR 
               "RX_DATA_DDP ERR! sk 0x%p, flag 0x%x, prev NULL.\n", 
               sk, ULP_DATACNT(tp));
    }
    __kfree_skb(skb);
}

/*
 * RX_ISCSI_HDR CPL messages.
 */
void ulp_new_rx_iscsi_hdr(struct sock *sk, struct sk_buff *skb)
{
    struct tcp_opt 		*tp = tcp_sk(sk);
    struct cpl_rx_data 		*header_cpl;
    __u32			header_len, total_len;
    struct skb_ulpinfo  	*skcb = (struct skb_ulpinfo *)SKB_ULPINFO(skb);
    struct sk_buff      	*skb_udata = ULP_DATASKB(tp);

    /* 
       RX_ISCSI_HDR can be coalesced.
       An coalesced RX_ISCSI_HDR msg could contain either
       1). RX_ISCSI_HDR + RX_ISCSI_HDR + RX_DATA_DDP or
       2). RX_ISCSI_HDR + RX_DATA_DDP
       RX_ISCSI_HDR are the same as RX_DATA except the cpl opcode 
    */

    header_cpl = (struct cpl_rx_data *)cplhdr(skb);
    /* pdu header length */
    header_len = ntohl(get_unaligned(&header_cpl->len));
    __skb_pull(skb, sizeof(struct cpl_rx_data));
    total_len = skb->len;

    /* msg coalesce is on */
    if (total_len > header_len) {
        struct cpl_rx_data_ddp 	*status_cpl;

        status_cpl = (struct cpl_rx_data_ddp *)
			(skb->tail - sizeof(struct cpl_rx_data_ddp));
        total_len -= sizeof(struct cpl_rx_data_ddp);
        /* there is more data than the pdu header */
        if (total_len > header_len) {
            total_len -= sizeof(struct cpl_rx_data);
            ulp_proc_ddp_status(skb, status_cpl, 1, 1);
            /* this is a coalesced msg, so between the header and data, there is
               16 bytes gap (for 2nd cpl_rx_iscsi_hdr).  */
            skcb->data_gap = sizeof(struct cpl_rx_data);
        } else {
            ulp_proc_ddp_status(skb, status_cpl, 0, 1);
        }
                                                                                
        tp->rcv_nxt = ntohl(get_unaligned(&(status_cpl->seq))) +
                      ntohl(get_unaligned(&(status_cpl->len)));
        tp->ack.lrcvtime = tcp_time_stamp;
                                                                                
        __skb_trim(skb, total_len);
        __skb_queue_tail(&sk->sk_receive_queue, skb);
        sk->sk_data_ready(sk, 0);
    } else {
        /* msg coalesce is off */
        ULP_DATACNT(tp)++;
        if (skb_udata) {
            __skb_queue_tail(&sk->sk_receive_queue, skb_udata);
        }
        ULP_DATASKB(tp) = skb;
        skcb->data_ready = 0;
        //tp->rcv_nxt += header_len;
        tp->ack.lrcvtime = tcp_time_stamp;
    }

    return;
}

extern int do_rx_iscsi_hdr(struct toedev *dev, struct sk_buff *skb);
extern int do_rx_data_ddp(struct toedev *dev, struct sk_buff *skb);
void * t1_init_ulp(struct toedev *dev)
{
    struct adapter *adapter = dev->priv;
    struct ulp_info *ulp = NULL;
    struct toe_page_info tx_info, rx_info;
    unsigned int max_pdu;

    if (dev->ctl(dev, TOE_GET_TX_PAGE_INFO, &tx_info) < 0 ||
        dev->ctl(dev, TOE_GET_RX_PAGE_INFO, &rx_info) < 0 ) {
	printk(KERN_ERR "Chelsio ULP offload: get page info failed!\n");
        goto out;
    }

    t1_register_cpl_handler(CPL_RX_DATA_DDP, do_rx_data_ddp);
    t1_register_cpl_handler(CPL_RX_ISCSI_HDR, do_rx_iscsi_hdr);

    max_pdu = min(adapter->params.sge.large_buf_capacity,
                  adapter->params.tp.rx_coalescing_size);
    if (max_pdu > 16224) max_pdu = 16224;

    ulp = kmalloc(sizeof(struct ulp_info), GFP_KERNEL);
    if (!ulp) goto out;
    ulp->tdev = dev;
    if (dev->ttid == TOE_ID_CHELSIO_T1) {
        ulp->version = 0;
    } else {
        ulp->version = 1;
        max_pdu -= 2 * sizeof(struct cpl_rx_data) + sizeof(struct cpl_rx_data_ddp);
    }
    ulp->dmt = t1_init_ddp_table(ulp);
    if (!ulp->dmt) {
        kfree(ulp);
        ulp = NULL;
        goto out;
    }

    ulp->max_tx_pkt_size = min(tx_info.page_size, max_pdu);
    ulp->max_rx_pkt_size = min(rx_info.page_size, max_pdu);

#ifdef __ULP_DEBUG__
    printk(KERN_INFO "Initializing ULP %d, 0x%p, max pkt: %u, %u\n", ulp->version, ulp, ulp->max_tx_pkt_size, ulp->max_rx_pkt_size);
#endif

out:  
    return (ulp);
}

void t1_free_ulp(struct toedev *dev)
{
    struct ulp_info *ulp;
    ulp = dev->ulp;
    t1_free_ddp(ulp->dmt);
}

int t1_tom_udev_data(struct sock *sk, struct ulpdev_data *dp)
{
    struct tcp_opt	*tp = tcp_sk(sk);
    struct toedev	*tdev = TOE_DEV(tp);
    struct l2t_entry 	*e = L2T_ENTRY(tp);
    struct tom_data	*d;
    struct ulp_info     *ulp;
    chtoe_ddp_map_table *dmt;
    u32			mss;

    if (!tdev || !e || !dp ) {
        printk (KERN_INFO "%s: sk 0x%p, dp 0x%p, tdev 0x%p, e 0x%p.\n", 
                __FUNCTION__, sk, dp, tdev, e); 
        return -EINVAL;
    }
    d = TOM_DATA(tdev);
    if (!d) {
        printk (KERN_INFO "%s: sk 0x%p, tdev 0x%p, tom NULL.\n", 
                __FUNCTION__, sk, tdev); 
        return -EINVAL;
    }
    mss = min(tp->mss_cache, d->lmss[e->real_dev->if_port]);

    ulp = ULPDATA(tdev);
    if (!ulp) {
        printk (KERN_ERR "%s: sk 0x%p, dp 0x%p, tdev 0x%p, ulp NULL.\n",
                __FUNCTION__, sk, dp, tdev); 
        return -EINVAL;
    }

    dmt = ulp->dmt;

    dp->privdata = (void *)ulp;
    dp->max_rx_pkt_size = min(ulp->max_rx_pkt_size, mss);
    dp->max_tx_pkt_size = min(ulp->max_tx_pkt_size, mss);
    dp->tid = TID(tp);
    dp->pad_included = ulp->version ? 0 : 1;
    dp->tag_id_bits = dmt->tag_id_bits;
    dp->tag_color_bits = dmt->tag_color_bits;

    return 0;
}

/* set the tcb for ulp mode and the sock ulpmode (not the submode), */
/* the tx side will be controlled via sendskb */
int t1_tom_set_conn_rx_ulp_mode(struct sock *sk, unsigned char mode, 
    unsigned char hcrc, unsigned char dcrc, unsigned long pagesize)
{
    __u32		submode, pg_encode;
    struct toedev       *tdev = TOE_DEV(tcp_sk(sk));
    struct ulp_info     *ulp = ULPDATA(tdev);
    struct tcp_opt 	*tp = tcp_sk(sk);
    struct sk_buff 	*skb1 = alloc_skb_nofail(sizeof(struct cpl_set_tcb_field));
    struct sk_buff 	*skb2 = alloc_skb_nofail(sizeof(struct cpl_set_tcb_field));
    struct sk_buff 	*skb3 = alloc_skb_nofail(sizeof(struct cpl_set_tcb_field));
    struct cpl_set_tcb_field *req;

    if (mode != ULP_MODE_ISCSI) {
        printk (KERN_ERR "t1_set_ulp_mode: UNKNOWN mode %u.\n", mode);
        return -EINVAL;
    }
    ULP_MODE(tp) = ULP_MODE_ISCSI;

    submode = ( (hcrc ? 1 : 0) | (dcrc ? 2 : 0) );

    switch (pagesize) {
        case 0x1000:    pg_encode = PAGE_POD_PAGE_SIZE_4K;
            break;
        case 0x2000:    pg_encode = PAGE_POD_PAGE_SIZE_8K;
            break;
        case 0x4000:    pg_encode = PAGE_POD_PAGE_SIZE_16K;
            break;
        case 0x1000000: 
            pg_encode = ulp->version ? PAGE_POD_PAGE_SIZE_16M_ULP1 : 
                                       PAGE_POD_PAGE_SIZE_16M_ULP0;
            break;
        default:
            pg_encode = PAGE_POD_PAGE_SIZE_4K;
            printk(KERN_ERR "UNKNOWN ddp page size =0x%lx, default to 4K.\n", 
                            pagesize);
    }

    /* set up ulp mode and submode */
    req = (struct cpl_set_tcb_field *)__skb_put(skb1, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SET_TCB_FIELD, TID(tp)));
    req->offset = W_TCB_ULP_MODE;
    req->rsvd[0] = 0;
    req->rsvd[1] = 0;
    req->rsvd[2] = 0;
    req->mask = htonl( (M_TCB_ULP_SUBMODE << S_TCB_ULP_SUBMODE) |
                       (M_TCB_ULP_PAGE_SIZE << S_TCB_ULP_PAGE_SIZE) |
                       (M_TCB_ULP_MODE << S_TCB_ULP_MODE) );
    req->val = htonl( (mode << S_TCB_ULP_MODE) |
                      (submode << S_TCB_ULP_SUBMODE) |
                      (pg_encode << S_TCB_ULP_PAGE_SIZE) );
    skb1->priority = CPL_PRIORITY_DATA;

    /* header length */
    req = (struct cpl_set_tcb_field *)__skb_put(skb2, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SET_TCB_FIELD, TID(tp)));
    req->offset = 28;
    req->rsvd[0] = 0;
    req->rsvd[1] = 0;
    req->rsvd[2] = 0;
    req->mask = htonl(0x000fe700);
    req->val = htonl(0x000d6000);
    skb2->priority = CPL_PRIORITY_DATA;

    /* turn off rx coalescing */
    req = (struct cpl_set_tcb_field *)__skb_put(skb3, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SET_TCB_FIELD, TID(tp)));
    req->offset = 18;
    req->rsvd[0] = 0;
    req->rsvd[1] = 0;
    req->rsvd[2] = 0;
    req->mask = htonl(0x00000400);
    req->val = htonl(0x00000000);
    skb3->priority = CPL_PRIORITY_DATA;

    if (sk->sk_state == TCP_SYN_SENT) {
        __skb_queue_tail(&tp->out_of_order_queue, skb1);
        __skb_queue_tail(&tp->out_of_order_queue, skb2);  
        __skb_queue_tail(&tp->out_of_order_queue, skb3);  
    } else {
	l2t_send(tdev, skb1, L2T_ENTRY(tp));
	l2t_send(tdev, skb2, L2T_ENTRY(tp));
	l2t_send(tdev, skb3, L2T_ENTRY(tp));
    }

    return 0;
}


#ifdef __ULP_DEBUG__
void dbg_dump_byte_str (char *str, u8 *byte_str, u32 start, u32 len)
{
   u8     *ptr;
   u32    i;
   char         buffer[60], *buf;

   if (str) {
      printk ("%s: %u -- %u: \n", str, start, (start+len-1));
   }
   ptr = byte_str + start;
   buf = buffer;
   for (i = 0; i < len; i++, ptr++)
   {
      if (i && (i % 16 == 0))   {
         printk ("%s\n", buffer);
         buf = buffer;
      }
      sprintf (buf, "%02x ", *ptr);
      buf += 3;
   }

   if (buf != buffer) {
      printk ("%s\n", buffer);
      buf = buffer;
   }
}
#endif

EXPORT_SYMBOL(t1_tom_udev_data);
EXPORT_SYMBOL(t1_tom_set_conn_rx_ulp_mode);


// Local Variables:
// c-basic-offset: 4
// End:
