/*
 * Copyright 2003 Digi International (www.digi.com)
 *	Scott H Kilau <Scott_Kilau at digi dot com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *	NOTE TO LINUX KERNEL HACKERS:  DO NOT REFORMAT THIS CODE!
 *
 *	This is shared code between Digi's CVS archive and the
 *	Linux Kernel sources.
 *	Changing the source just for reformatting needlessly breaks
 *	our CVS diff history.
 *
 *	Send any bug fixes/changes to:  Eng.Linux at digi dot com.
 *	Thank you.
 *
 * $Id: jsm_driver.c,v 1.57 2004/08/31 19:41:34 scottk Exp $
 *
 */


#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/pci.h>

#include "jsm_driver.h"
#include "jsm_pci.h"
#include "jsm_proc.h"
#include "dpacompat.h"
#include "jsm_mgmt.h"
#include "jsm_tty.h"
#include "jsm_trace.h"
#include "jsm_neo.h"

/*
 * Because this driver is supported on very old 2.4 versions of the
 * Linux kernel, we still need this ifdef check.
 */
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

MODULE_AUTHOR("Digi International, http://www.digi.com");
MODULE_DESCRIPTION("Driver for the Digi International Neo and Classic PCI based product line");
MODULE_SUPPORTED_DEVICE("jsm");


#define JSM_DRIVER_NAME "jsm"
#define NR_PORTS       32 
#define JSM_MINOR_START        0 

struct uart_driver jsm_uart_driver = {
        .owner = THIS_MODULE,
        .driver_name = JSM_DRIVER_NAME,
        .dev_name = "ttyn", 
        .major = 253,
        .minor = JSM_MINOR_START, 
        .nr = NR_PORTS,
        .cons = NULL,
};      
/*
 * insmod command line overrideable parameters
 *
 * NOTE: we use a set of macros to create the variables, which allows
 * us to specify the variable type, name, initial value, and description.
 */
PARM_INT(debug,		0x00,		"Driver debugging level");
PARM_INT(rawreadok,	1,		"Bypass flip buffers on input");
PARM_INT(trcbuf_size,	0x100000,	"Debugging trace buffer size. 1M ");
int jsm_debug, jsm_rawreadok, jsm_trcbuf_size;

/**************************************************************************
 *
 * protos for this file
 *
 */
static int		jsm_start(void);
static int		jsm_finalize_board_init(struct board_t *brd);
static void		jsm_init_globals(void);
static int		jsm_found_board(struct pci_dev *pdev, int id);
static void		jsm_cleanup_board(struct board_t *brd);
static void		jsm_poll_handler(ulong dummy);
static int		jsm_init_one(struct pci_dev *pdev, const struct pci_device_id *ent);
static void		jsm_remove_one(struct pci_dev *dev);
static int		jsm_probe1(struct pci_dev *pdev, int card_type);
static void		jsm_mbuf(struct board_t *brd, const char *fmt, ...);

/* Driver load/unload functions */
int		jsm_init_module(void);
void		jsm_cleanup_module(void);


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)

module_init(jsm_init_module);
module_exit(jsm_cleanup_module);

#else

/*
 * Older 2.4 kernels do not support module_init/exit calls.
 * So we do it this way for now.
 */
int		init_module(void);
void		cleanup_module(void);
#define jsm_init_module init_module   
#define jsm_cleanup_module cleanup_module

#endif


/*
 * File operations permitted on Control/Management major.
 */
static struct file_operations jsm_BoardFops =
{
	.owner		=	THIS_MODULE,
	.ioctl		=	jsm_mgmt_ioctl,
	.open		=	jsm_mgmt_open,
	.release	=	jsm_mgmt_close
};


/*
 * Globals
 */
uint			jsm_NumBoards;
struct board_t		*jsm_Board[MAXBOARDS];
spinlock_t		jsm_global_lock = SPIN_LOCK_UNLOCKED;
int			jsm_driver_state = DRIVER_INITIALIZED;
ulong			jsm_poll_counter;
uint			jsm_Major;

/*
 * Static vars.
 */
static uint		jsm_Major_Control_Registered = FALSE;
static uint		jsm_driver_start = FALSE;

/*
 * Poller stuff
 */
static spinlock_t	jsm_poll_lock = SPIN_LOCK_UNLOCKED;	/* Poll scheduling lock */
static ulong		jsm_poll_time;				/* Time of next poll */
static ulong		jsm_poll_tick = 20;			/* Poll interval - 20 ms */
static uint		jsm_poll_stop;				/* Used to tell poller to stop */
static struct timer_list jsm_poll_timer =
{
	.function =	jsm_poll_handler
};


static struct pci_device_id jsm_pci_tbl[] = {
	{	DIGI_VID, PCI_DEVICE_CLASSIC_4_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	0 },
	{	DIGI_VID, PCI_DEVICE_CLASSIC_4_422_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	1 },
	{	DIGI_VID, PCI_DEVICE_CLASSIC_8_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	2 },
	{	DIGI_VID, PCI_DEVICE_CLASSIC_8_422_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	3 },
	{	DIGI_VID, PCI_DEVICE_NEO_4_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,		4 },
	{	DIGI_VID, PCI_DEVICE_NEO_8_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,		5 },
	{	DIGI_VID, PCI_DEVICE_NEO_2DB9_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	6 },
	{	DIGI_VID, PCI_DEVICE_NEO_2DB9PRI_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	7 },
	{	DIGI_VID, PCI_DEVICE_NEO_2RJ45_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	8 },
	{	DIGI_VID, PCI_DEVICE_NEO_2RJ45PRI_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	9 },
	{	DIGI_VID, PCI_DEVICE_NEO_1_422_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	10 },
	{	DIGI_VID, PCI_DEVICE_NEO_1_422_485_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	11 },
	{	DIGI_VID, PCI_DEVICE_NEO_2_422_485_DID, PCI_ANY_ID, PCI_ANY_ID, 0, 0,	12 },
	{0,}						/* 0 terminated list. */
};
MODULE_DEVICE_TABLE(pci, jsm_pci_tbl);

struct board_id {
	uchar *name;
	uint maxports;
};

static struct board_id jsm_Ids[] =
{	
	{	PCI_DEVICE_CLASSIC_4_PCI_NAME,		4	},
	{	PCI_DEVICE_CLASSIC_4_422_PCI_NAME,	4	},
	{	PCI_DEVICE_CLASSIC_8_PCI_NAME,		8	},
	{	PCI_DEVICE_CLASSIC_8_422_PCI_NAME,	8	},
	{	PCI_DEVICE_NEO_4_PCI_NAME,		4	},
	{	PCI_DEVICE_NEO_8_PCI_NAME,		8	},
	{	PCI_DEVICE_NEO_2DB9_PCI_NAME,		2	},
	{	PCI_DEVICE_NEO_2DB9PRI_PCI_NAME,	2	},
	{	PCI_DEVICE_NEO_2RJ45_PCI_NAME,		2	},
	{	PCI_DEVICE_NEO_2RJ45PRI_PCI_NAME,	2	},
	{	PCI_DEVICE_NEO_1_422_PCI_NAME,		1	},
	{	PCI_DEVICE_NEO_1_422_485_PCI_NAME,	1	},
	{	PCI_DEVICE_NEO_2_422_485_PCI_NAME,	2	},
	{	NULL,					0	}
};

static struct pci_driver jsm_driver = {
	.name		= "jsm",
	.probe		= jsm_init_one,
	.id_table       = jsm_pci_tbl,
	.remove		= __devexit_p(jsm_remove_one),
};


char *jsm_state_text[] = {
	"Board Failed",
	"Board Found",
	"Board READY",
};

char *jsm_driver_state_text[] = {
	"Driver Initialized",
	"Driver Ready."
};



/************************************************************************
 *
 * Driver load/unload functions
 *
 ************************************************************************/

/*
 * init_module()
 *
 * Module load.  This is where it all starts.
 */
int jsm_init_module(void)
{
	int rc = 0;

	APR(("%s, Digi International Part Number %s\n", DG_NAME, DG_PART));

	/*
	 * Initialize global stuff
	 */
	rc = jsm_start();

	if (rc < 0) 
		return(rc);

	 rc = uart_register_driver(&jsm_uart_driver);
	return pci_module_init(&jsm_driver);
}


/*
 * Start of driver.
 */
static int jsm_start(void)
{
	int rc = 0;
	unsigned long flags;

	if (jsm_driver_start == FALSE) {

		jsm_driver_start = TRUE;

		/* make sure that the globals are init'd before we do anything else */
		jsm_init_globals();

		jsm_NumBoards = 0;

		APR(("For the tools package or updated drivers please visit http://www.digi.com\n"));

		/*
		 * Register our base character device into the kernel.
		 * This allows the download daemon to connect to the downld device
		 * before any of the boards are init'ed.
		 */
		if (!jsm_Major_Control_Registered) {
			/*
			 * Register management/dpa devices
			 */
			rc = register_chrdev(0, "jsm", &jsm_BoardFops);
			if (rc <= 0) {
				APR(("Can't register jsm driver device (%d)\n", rc));
				rc = -ENXIO;
				return(rc);
			}
			jsm_Major = rc;
			jsm_Major_Control_Registered = TRUE;
		}

		/*
		 * Register our basic stuff in /proc/jsm
		 */
		jsm_proc_register_basic_prescan();

		/*
		 * Init any global tty stuff.
		 */
		rc = jsm_tty_preinit();

		if (rc < 0) {
			APR(("tty preinit - not enough memory (%d)\n", rc));
			return(rc);
		}

		/* Start the poller */
		JSM_LOCK(jsm_poll_lock, flags);
		jsm_poll_time = jiffies + jsm_jiffies_from_ms(jsm_poll_tick);
		jsm_poll_timer.expires = jsm_poll_time;
		JSM_UNLOCK(jsm_poll_lock, flags);

		add_timer(&jsm_poll_timer);

		jsm_driver_state = DRIVER_READY;
	}

	return(rc);
}


/* returns count (>= 0), or negative on error */
static int jsm_init_one(struct pci_dev *pdev, const struct pci_device_id *ent)
{       
	int rc;

	rc = pci_enable_device(pdev);
	if (rc) {
		dev_err(&pdev->dev, "Device enable FAILED\n");
                return rc;
        } 

/*	if ( (rc = pci_request_regions(pdev, "jsm"))) {
                 dev_err(&pdev->dev, "pci_request_region FAILED\n");
                 pci_disable_device(pdev);
                 return rc;
         }
*/

	if ( (rc = jsm_probe1(pdev, ent->driver_data))) {
                dev_err(&pdev->dev, "jsm_probe1 FAILED\n");
                pci_disable_device(pdev);
	 	return rc;
	}
	return rc;
}

static int jsm_probe1(struct pci_dev *pdev, int card_type)
{
	return jsm_found_board(pdev, card_type);
}


static void jsm_remove_one(struct pci_dev *dev)
{
	int i;

	for (i = 0; i < jsm_NumBoards; ++i) {
		jsm_remove_uart_port(jsm_Board[i]);
		jsm_cleanup_board(jsm_Board[i]);
		jsm_tty_uninit(jsm_Board[i]);
	}
}

/*
 * jsm_cleanup_module()
 *
 * Module unload.  This is where it all ends.
 */
void jsm_cleanup_module(void)
{
	ulong lock_flags;

	JSM_LOCK(jsm_poll_lock, lock_flags);
	jsm_poll_stop = 1;
	JSM_UNLOCK(jsm_poll_lock, lock_flags);

	del_timer_sync(&jsm_poll_timer);

	jsm_proc_unregister_all();

	if (jsm_Major_Control_Registered)
		unregister_chrdev(jsm_Major, "jsm");

	jsm_tty_post_uninit();

#if defined(JSM_TRACER)
	jsm_tracer_free();
#endif
	if (jsm_NumBoards) {
		pci_unregister_driver(&jsm_driver);
	}
	uart_unregister_driver(&jsm_uart_driver);

}


/*
 * jsm_cleanup_board()
 *
 * Free all the memory associated with a board
 */
static void jsm_cleanup_board(struct board_t *brd)
{
	int i = 0;

        if(!brd || brd->magic != JSM_BOARD_MAGIC)
                return;

	switch (brd->device) {
	case PCI_DEVICE_CLASSIC_4_DID:
	case PCI_DEVICE_CLASSIC_8_DID:
	case PCI_DEVICE_CLASSIC_4_422_DID:
	case PCI_DEVICE_CLASSIC_8_422_DID:

		/* Tell card not to interrupt anymore. */
		outb(0, brd->iobase + 0x4c);
		break;

	default:
		break;
	}

	if (brd->irq)
		free_irq(brd->irq, brd);

	tasklet_kill(&brd->helper_tasklet);

	if (brd->re_map_membase) {
		iounmap(brd->re_map_membase);
		brd->re_map_membase = NULL;
	}

	if (brd->msgbuf_head) {
		unsigned long flags;

		JSM_LOCK(jsm_global_lock, flags);
		brd->msgbuf = NULL;
		printk(brd->msgbuf_head);
		kfree(brd->msgbuf_head);
		brd->msgbuf_head = NULL;
		JSM_UNLOCK(jsm_global_lock, flags);
        }

	/* Free all allocated channels structs */
	for (i = 0; i < MAXPORTS ; i++) {
		if (brd->channels[i]) {
			if (brd->channels[i]->ch_rqueue)
				kfree(brd->channels[i]->ch_rqueue);
			if (brd->channels[i]->ch_equeue)
				kfree(brd->channels[i]->ch_equeue);
			if (brd->channels[i]->ch_wqueue)
				kfree(brd->channels[i]->ch_wqueue);

			kfree(brd->channels[i]);
			brd->channels[i] = NULL;
		}
	}

	if (brd->flipbuf)
		kfree(brd->flipbuf);

	jsm_Board[brd->boardnum] = NULL;

        kfree(brd);
}


/*
 * jsm_found_board()
 *
 * A board has been found, init it.
 */
static int jsm_found_board(struct pci_dev *pdev, int id)
{
        struct board_t *brd;
	unsigned int pci_irq;
	int i = 0;
	int rc = 0;
	unsigned long flags;

        /* get the board structure and prep it */
        brd = jsm_Board[jsm_NumBoards] =
        (struct board_t *) jsm_driver_kzmalloc(sizeof(struct board_t), GFP_KERNEL);
	if (!brd) {
		APR(("memory allocation for board structure failed\n"));
		return(-ENOMEM);
	}
	memset(brd,0,sizeof(struct board_t));

        /* make a temporary message buffer for the boot messages */
        brd->msgbuf = brd->msgbuf_head =
                (char *) jsm_driver_kzmalloc(sizeof(char) * 8192, GFP_KERNEL);
        if (!brd->msgbuf) {
		kfree(brd);
		APR(("memory allocation for board msgbuf failed\n"));
		return(-ENOMEM);
        }
        
	/* store the info for the board we've found */
	brd->magic = JSM_BOARD_MAGIC;
	brd->boardnum = jsm_NumBoards;
	brd->vendor = jsm_pci_tbl[id].vendor;
	brd->device = jsm_pci_tbl[id].device;
	brd->pci_bus = pdev->bus->number;
	brd->pci_slot = PCI_SLOT(pdev->devfn);
	brd->pci_dev = pdev;
	brd->name = jsm_Ids[id].name;
	brd->maxports = jsm_Ids[id].maxports;
	brd->dpastatus = BD_NOFEP;
	init_waitqueue_head(&brd->state_wait);

	JSM_SPINLOCK_INIT(brd->bd_lock);
	JSM_SPINLOCK_INIT(brd->bd_intr_lock);

	brd->state = BOARD_FOUND;

	for (i = 0; i < MAXPORTS; i++) {
		brd->channels[i] = NULL;
	}

	/* store which card & revision we have */
	pci_read_config_word(pdev, PCI_SUBSYSTEM_VENDOR_ID, &brd->subvendor);
	pci_read_config_word(pdev, PCI_SUBSYSTEM_ID, &brd->subdevice);
	pci_read_config_byte(pdev, PCI_REVISION_ID, &brd->rev);

	pci_irq = pdev->irq;
	brd->irq = pci_irq;


	switch(brd->device) {

	case PCI_DEVICE_NEO_4_DID:
	case PCI_DEVICE_NEO_8_DID:
	case PCI_DEVICE_NEO_2DB9_DID:
	case PCI_DEVICE_NEO_2DB9PRI_DID:
	case PCI_DEVICE_NEO_2RJ45_DID:
	case PCI_DEVICE_NEO_2RJ45PRI_DID:
	case PCI_DEVICE_NEO_1_422_DID:
	case PCI_DEVICE_NEO_1_422_485_DID:
	case PCI_DEVICE_NEO_2_422_485_DID:

		/*
		 * This chip is set up 100% when we get to it.
		 * No need to enable global interrupts or anything. 
		 */
		brd->dpatype = T_NEO | T_PCIBUS;

                DPR_INIT(("jsm_found_board - NEO.\n"));

		/* get the PCI Base Address Registers */
		brd->membase     = pci_resource_start(pdev, 0);
		brd->membase_end = pci_resource_end(pdev, 0);

		if (brd->membase & 1)
			brd->membase &= ~3;
		else
			brd->membase &= ~15;

		/* Assign the board_ops struct */
		brd->bd_ops = &jsm_neo_ops;

		brd->bd_uart_offset = 0x200;
		brd->bd_dividend = 921600;

		brd->re_map_membase = ioremap(brd->membase, 0x1000);
		DPR_INIT(("remapped mem: 0x%p\n", brd->re_map_membase));
		if (!brd->re_map_membase) {
			kfree(brd->msgbuf);
			kfree(brd);
			APR(("card has no PCI Memory resources, failing board.\n"));
			return -ENOMEM;
		}
		break;

	default:
		APR(("Did not find any compatible Neo or Classic PCI boards in system.\n"));
		kfree(brd->msgbuf);
		kfree(brd);
		return(-ENXIO);

	}

	/*
	 * Do tty device initialization.
	 */


	rc = jsm_finalize_board_init(brd);
	if (rc < 0) {
		APR(("Can't finalize board init (%d)\n", rc));
		brd->state = BOARD_FAILED;
		brd->dpastatus = BD_NOFEP;
		goto failed;

	}



	rc = jsm_tty_register(brd);
	if (rc < 0) {
		jsm_tty_uninit(brd);
		APR(("Can't register tty devices (%d)\n", rc));
		brd->state = BOARD_FAILED;
		brd->dpastatus = BD_NOFEP;
		goto failed;
	}

	rc = jsm_tty_init(brd);
	if (rc < 0) {
		jsm_tty_uninit(brd);
		APR(("Can't init tty devices (%d)\n", rc));
		brd->state = BOARD_FAILED;
		brd->dpastatus = BD_NOFEP;
		goto failed;
	}

	rc = jsm_uart_port_init(brd);
	if (rc < 0) {
		goto failed;
	}

	brd->state = BOARD_READY;
	brd->dpastatus = BD_RUNNING;

	/* init our poll helper tasklet */
	tasklet_init(&brd->helper_tasklet, brd->bd_ops->tasklet, (unsigned long) brd);

	 /* Log the information about the board */
	jsm_mbuf(brd, DRVSTR": board %d: %s (rev %d), irq %d\n",
		jsm_NumBoards, brd->name, brd->rev, brd->irq);

	DPR_INIT(("jsm_scan(%d) - printing out the msgbuf\n", i));
	JSM_LOCK(jsm_global_lock, flags);
	brd->msgbuf = NULL;
	printk(brd->msgbuf_head);
	kfree(brd->msgbuf_head);
	brd->msgbuf_head = NULL;
	JSM_UNLOCK(jsm_global_lock, flags);

	/*
	 * allocate flip buffer for board.
	 *
	 * Okay to malloc with GFP_KERNEL, we are not at interrupt
	 * context, and there are no locks held.
	 */
	brd->flipbuf = jsm_driver_kzmalloc(MYFLIPLEN, GFP_KERNEL);

	jsm_proc_register_basic_postscan(jsm_NumBoards);

	wake_up_interruptible(&brd->state_wait);

	jsm_NumBoards++;

        return(0);

failed:
	kfree(brd->msgbuf);
	kfree(brd);
	iounmap((void *) brd->re_map_membase);
	return (-ENXIO);

}


static int jsm_finalize_board_init(struct board_t *brd) {
	int rc = 0;

        DPR_INIT(("jsm_finalize_board_init() - start\n"));

	if (!brd || brd->magic != JSM_BOARD_MAGIC)
                return(-ENODEV);

        DPR_INIT(("jsm_finalize_board_init() - start #2\n"));

	if (brd->irq) {
		rc = request_irq(brd->irq, brd->bd_ops->intr, SA_INTERRUPT|SA_SHIRQ, "JSM", brd);

		if (rc) {
			printk("Failed to hook IRQ %d\n",brd->irq);
			brd->state = BOARD_FAILED;
			brd->dpastatus = BD_NOFEP;
			rc = -ENODEV;
		} else {
			DPR_INIT(("Requested and received usage of IRQ %d\n", brd->irq));
		}
	}
	return(rc);
}


/*****************************************************************************
*
* Function:
*                                       
*    jsm_poll_handler
*
* Author:
*
*    Scott H Kilau
*       
* Parameters:
*
*    dummy -- ignored                    
*
* Return Values:
*
*    none
*
* Description:   
*                                       
*    As each timer expires, it determines (a) whether the "transmit"
*    waiter needs to be woken up, and (b) whether the poller needs to
*    be rescheduled.
*
******************************************************************************/

static void jsm_poll_handler(ulong dummy)
{
        struct board_t *brd;
        unsigned long lock_flags;
	int i;

	jsm_poll_counter++;

	/*
	 * Do not start the board state machine until
	 * driver tells us its up and running, and has
	 * everything it needs.
	 */
	if (jsm_driver_state != DRIVER_READY) {
		goto schedule_poller;
	}

	/* Go thru each board, kicking off a tasklet for each if needed */
	for (i = 0; i < jsm_NumBoards; i++) {
		brd = jsm_Board[i];

		JSM_LOCK(brd->bd_lock, lock_flags);

		/* If board is in a failed state, don't bother scheduling a tasklet */
		if (brd->state == BOARD_FAILED) {
			JSM_UNLOCK(brd->bd_lock, lock_flags);
			continue;
		}

		/* Schedule a poll helper task */
		tasklet_schedule(&brd->helper_tasklet);

		JSM_UNLOCK(brd->bd_lock, lock_flags);
	}

schedule_poller:

	/*
	 * Schedule ourself back at the nominal wakeup interval.
	 */
	if (jsm_NumBoards > 0) {
		ulong time;

		JSM_LOCK(jsm_poll_lock, lock_flags);
		jsm_poll_time += jsm_jiffies_from_ms(jsm_poll_tick);

		time = jsm_poll_time - jiffies;

		if ((ulong) time >= 2 * jsm_poll_tick) {
			jsm_poll_time = jiffies +  jsm_jiffies_from_ms(jsm_poll_tick);
		}

		jsm_poll_timer.expires = jsm_poll_time;
		JSM_UNLOCK(jsm_poll_lock, lock_flags);

		if (!jsm_poll_stop)
			add_timer(&jsm_poll_timer);
	}

}

/*
 * jsm_init_globals()
 *
 * This is where we initialize the globals from the static insmod
 * configuration variables.  These are declared near the head of
 * this file.
 */
static void jsm_init_globals(void)
{
	int i = 0;

	jsm_rawreadok		= rawreadok;
        jsm_trcbuf_size	= trcbuf_size;
	jsm_debug		= debug;

	for (i = 0; i < MAXBOARDS; i++) {
		jsm_Board[i] = NULL;
	}

	init_timer(&jsm_poll_timer);
}


/************************************************************************
 *
 * Utility functions
 *
 ************************************************************************/


/*
 * jsm_driver_kzmalloc()
 *
 * Malloc and clear memory,
 */
void *jsm_driver_kzmalloc(size_t size, int priority)
{
 	void *p = kmalloc(size, priority);
	if(p)
		memset(p, 0, size);
	return(p);
}


/*
 * jsm_mbuf()
 *
 * Used to print to the message buffer during board init.
 */
static void jsm_mbuf(struct board_t *brd, const char *fmt, ...) {
	va_list		ap;
	char		buf[1024];
	int		i;
	unsigned long	flags;

	JSM_LOCK(jsm_global_lock, flags);

	/* Format buf using fmt and arguments contained in ap. */
	va_start(ap, fmt);
	i = vsprintf(buf, fmt,  ap);
	va_end(ap);

	DPR((buf));

	if (!brd || !brd->msgbuf) {
		printk(buf);
		JSM_UNLOCK(jsm_global_lock, flags);
		return;
	}

	memcpy(brd->msgbuf, buf, strlen(buf));
	brd->msgbuf += strlen(buf);
	*brd->msgbuf = (char) NULL;

	JSM_UNLOCK(jsm_global_lock, flags);
}


/*
 * jsm_ms_sleep()
 *
 * Put the driver to sleep for x ms's
 *
 * Returns 0 if timed out, !0 (showing signal) if interrupted by a signal.
 */
int jsm_ms_sleep(ulong ms)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout((ms * HZ) / 1000);
	return (signal_pending(current));
}



/*
 *      jsm_ioctl_name() : Returns a text version of each ioctl value.
 */
char *jsm_ioctl_name(int cmd)
{
	switch(cmd) {

	case TCGETA:		return("TCGETA");
	case TCGETS:		return("TCGETS");
	case TCSETA:		return("TCSETA");
	case TCSETS:		return("TCSETS");
	case TCSETAW:		return("TCSETAW");
	case TCSETSW:		return("TCSETSW");
	case TCSETAF:		return("TCSETAF");
	case TCSETSF:		return("TCSETSF");
	case TCSBRK:		return("TCSBRK");
	case TCXONC:		return("TCXONC");
	case TCFLSH:		return("TCFLSH");
	case TIOCGSID:		return("TIOCGSID");

	case TIOCGETD:		return("TIOCGETD");
	case TIOCSETD:		return("TIOCSETD");
	case TIOCGWINSZ:	return("TIOCGWINSZ");
	case TIOCSWINSZ:	return("TIOCSWINSZ");

	case TIOCMGET:		return("TIOCMGET");
	case TIOCMSET:		return("TIOCMSET");
	case TIOCMBIS:		return("TIOCMBIS");
	case TIOCMBIC:		return("TIOCMBIC");

	/* from digi.h */
	case DIGI_SETA:		return("DIGI_SETA");
	case DIGI_SETAW:	return("DIGI_SETAW");
	case DIGI_SETAF:	return("DIGI_SETAF");
	case DIGI_SETFLOW:	return("DIGI_SETFLOW");
	case DIGI_SETAFLOW:	return("DIGI_SETAFLOW");
	case DIGI_GETFLOW:	return("DIGI_GETFLOW");
	case DIGI_GETAFLOW:	return("DIGI_GETAFLOW");
	case DIGI_GETA:		return("DIGI_GETA");
	case DIGI_GEDELAY:	return("DIGI_GEDELAY");
	case DIGI_SEDELAY:	return("DIGI_SEDELAY");
	case DIGI_GETCUSTOMBAUD: return("DIGI_GETCUSTOMBAUD");
	case DIGI_SETCUSTOMBAUD: return("DIGI_SETCUSTOMBAUD");
	case TIOCMODG:		return("TIOCMODG");
	case TIOCMODS:		return("TIOCMODS");
	case TIOCSDTR:		return("TIOCSDTR");
	case TIOCCDTR:		return("TIOCCDTR");

	default:		return("unknown");
	}
}


