/*****************************************************************************
 *                                                                           *
 * File:                                                                     *
 *  toedev.c                                                                 *
 *                                                                           *
 * Description:                                                              *
 *  TOE device support infrastructure.                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License, version 2, as       *
 * published by the Free Software Foundation.                                *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED    *
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF      *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                     *
 *                                                                           *
 * http://www.chelsio.com                                                    *
 *                                                                           *
 * Copyright (c) 2003 - 2005 Chelsio Communications, Inc.                    *
 * All rights reserved.                                                      *
 *                                                                           *
 * Maintainers: maintainers@chelsio.com                                      *
 *                                                                           *
 * Authors: Dimitrios Michailidis   <dm@chelsio.com>                         *
 *                                                                           *
 * History:                                                                  *
 *                                                                           *
 ****************************************************************************/
/* $Date: 2005/11/18 07:48:51 $ $RCSfile: toedev.c,v $ $Revision: 1.24 $ */

#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/toedev.h>
#include <net/neighbour.h>
#include <asm/semaphore.h>

/* 2.4 compatibility */
#ifndef subsys_initcall
#define subsys_initcall(fn) module_init(fn)

static int boot_phase = 1;
#else
#define boot_phase 0
#endif

#ifndef __raise_softirq_irqoff
#define __raise_softirq_irqoff(nr) __cpu_raise_softirq(smp_processor_id(), nr)
#endif

static DECLARE_MUTEX(toedev_db_lock);
static LIST_HEAD(toedev_list);
static LIST_HEAD(tom_list);

static int toedev_init(void);

/*
 * Returns the entry in the TOE id table 'table' that has a given id, or NULL
 * if the id is not found.
 */
static const struct toe_id *id_find(unsigned int id,
				    const struct toe_id *table)
{
	const struct toe_id *p;

	for (p = table; p->id; ++p)
		if (p->id == id)
			return p;
	return NULL;
}

/*
 * Returns true if a TOE device is presently attached to an offload module.
 */
static inline int is_attached(const struct toedev *dev)
{
	return dev->offload_mod != NULL;
}

/*
 * Try to attach a new TOE device to an existing TCP offload module that can
 * handle the device's TOE id.  Returns 0 if it succeeds.
 *
 * Must be called with the toedev_db_lock held.
 */
static int toedev_attach(struct toedev *dev)
{
	struct tom_info *t;

	list_for_each_entry(t, &tom_list, list_node) {
		const struct toe_id *entry;

		entry = id_find(dev->ttid, t->id_table);
		if (entry && t->attach(dev, entry) == 0) {
			dev->offload_mod = t;
			return 0;
		}
	}
	return -ENOPROTOOPT;
}
			
/*
 * Register a TCP Offload Module (TOM).
 */
int register_tom(struct tom_info *t)
{
	down(&toedev_db_lock);
	list_add(&t->list_node, &tom_list);
	up(&toedev_db_lock);
	return 0;
}

/*
 * Unregister a TCP Offload Module (TOM).  Note that this does not affect any
 * TOE devices to which the TOM is already attached.
 */
int unregister_tom(struct tom_info *t)
{
	down(&toedev_db_lock);
	list_del(&t->list_node);
	up(&toedev_db_lock);
	return 0;
}

/*
 * Find a TOE device by name.  Must be called with toedev_db_lock held.
 */
static struct toedev *__find_toedev_by_name(const char *name)
{
	struct toedev *dev;

	list_for_each_entry(dev, &toedev_list, toe_list) {
		if (!strncmp(dev->name, name, TOENAMSIZ))
			return dev;
	}
	return NULL;
}

#if 0
/*
 * Find a TOE device by name.
 */
static struct toedev *find_toedev_by_name(const char *name)
{
	struct toedev *dev;

	down(&toedev_db_lock);
	dev = __find_toedev_by_name(name);
	if (dev)
		toedev_hold(dev);
	up(&toedev_db_lock);
	return dev;
}
#endif

/*
 * Find a TOE device by index.  Must be called with toedev_db_lock held.
 */
static struct toedev *__find_toedev_by_index(int index)
{
	struct toedev *dev;

	list_for_each_entry(dev, &toedev_list, toe_list) {
		if (dev->toe_index == index)
			return dev;
	}
	return NULL;
}

/*
 * Return true if a TOE device is already registered.
 * Must be called with the toedev_db_lock held.
 */
static int toedev_registered(const struct toedev *dev)
{
	struct toedev *d;

	list_for_each_entry(d, &toedev_list, toe_list) {
	if (d == dev)
		return 1;
	}
	return 0;
}

/*
 * Finalize the name of a TOE device by assigning values to any format strings
 * in its name.
 */
static int toedev_assign_name(struct toedev *dev, const char *name, int limit)
{
	int i;

	for (i = 0; i < limit; ++i) {
		char s[TOENAMSIZ];

		snprintf(s, sizeof(s), name, i);
		if (!__find_toedev_by_name(s)) {
			strcpy(dev->name, s);
			return 0;
		}
	}
	return -1;
}

/*
 * Allocate a unique index for a TOE device.  We keep the index within 30 bits
 * to allow it to be used as a sysctl index, which uses signed IDs.
 *
 * We don't handle index exhaustion.  Guess why.
 */
static int toedev_new_index(void)
{
	static int toe_index;

	for (;;) {
		if (++toe_index & 0xc0000000)
			toe_index = 1;
		if (!__find_toedev_by_index(toe_index))
			return toe_index;
	}
}

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>

static struct proc_dir_entry *toedev_proc_root;

/* XXX This doesn't handle module unloading properly.  Do we need to? */

static int devices_read_proc(char *buf, char **start, off_t offset,
			     int length, int *eof, void *data)
{
	int len;
	struct toedev *dev;
	struct net_device *ndev;

	len = sprintf(buf,
		"Device           Offload Module       Interfaces\n");

	down(&toedev_db_lock);
	list_for_each_entry(dev, &toedev_list, toe_list) {
		len += sprintf(buf + len, "%-16s %-20s", dev->name,
			is_attached(dev) ? dev->offload_mod->name : "<None>");
		read_lock(&dev_base_lock);
		for (ndev = dev_base; ndev; ndev = ndev->next) {
			if (TOEDEV(ndev) == dev)
				len += sprintf(buf + len, " %s", ndev->name);
		}
		read_unlock(&dev_base_lock);
		len += sprintf(buf + len, "\n");
		if (len >= length)
			break;
	}
	up(&toedev_db_lock);

	if (len > length)
		len = length;
	*eof = 1;
	return len;
}

static void toe_proc_cleanup(void)
{
	remove_proc_entry("devices", toedev_proc_root);
	remove_proc_entry("net/toe", NULL);
	toedev_proc_root = NULL;
}

static struct proc_dir_entry *create_toe_proc_dir(const char *name)
{
	struct proc_dir_entry *d;

	if (!toedev_proc_root)
		return NULL;

	d = proc_mkdir(name, toedev_proc_root);
	if (d)
		d->owner = THIS_MODULE;
	return d;
}

#if 0
static void delete_toe_proc_dir(struct toedev *dev)
{
	if (dev->proc_dir) {
		remove_proc_entry(dev->name, toedev_proc_root);
		dev->proc_dir = NULL;
	}
}
#endif

static int __init toe_proc_init(void)
{
	struct proc_dir_entry *d;

	toedev_proc_root = proc_mkdir("net/toe", NULL);
	if (!toedev_proc_root)
		return -ENOMEM;
	toedev_proc_root->owner = THIS_MODULE;

	d = create_proc_read_entry("devices", 0, toedev_proc_root, 
			       devices_read_proc, NULL);
	if (!d)
		goto cleanup;
	d->owner = THIS_MODULE;
	return 0;

cleanup:
	toe_proc_cleanup();
	return -ENOMEM;
}
#else
#define toe_proc_init() 0
#define create_toe_proc_dir(name) NULL
#define delete_toe_proc_dir(dev)
#endif /* CONFIG_PROC_FS */

/*
 * Register a TOE device and try to attach an appropriate TCP offload module
 * to it.  'name' is a template that may contain at most one %d format
 * specifier.
 */
int register_toedev(struct toedev *dev, const char *name)
{
	int ret;
	char *p;

	if (boot_phase)
		toedev_init();

	/* Validate the name template.  Only one %d allowed. */
	if (!name || !*name)
		return -EINVAL;
	p = strchr(name, '%');
	if (p && (p[1] != 'd' || strchr(p + 2, '%')))
		return -EINVAL;

	down(&toedev_db_lock);
	if (toedev_registered(dev)) {   /* device already registered */
		ret = -EEXIST;
		goto out;
	}

	if ((ret = toedev_assign_name(dev, name, 32)) != 0)
		goto out;

	dev->proc_dir = create_toe_proc_dir(dev->name);
	dev->toe_index = toedev_new_index();
	dev->offload_mod = NULL;
	list_add_tail(&dev->toe_list, &toedev_list);
	toedev_hold(dev);
out:
	up(&toedev_db_lock);
	return ret;
}

/*
 * Allocate and initialize a toedev structure.
 */
struct toedev *alloc_toedev(void)
{
	struct toedev *dev = kmalloc(sizeof(struct toedev), GFP_KERNEL);

	if (dev) {
		memset(dev, 0, sizeof(struct toedev));
		atomic_set(&dev->refcnt, 0);
		INIT_LIST_HEAD(&dev->toe_list);
	}
	return dev;
}

/*
 * Activate a TOE device.
 */
int activate_toedev(struct toedev *dev)
{
	int ret = 0;

	down(&toedev_db_lock);
	if (!toedev_registered(dev))
		ret = -ENODEV;
	else if (!is_attached(dev))
		ret = toedev_attach(dev);
	up(&toedev_db_lock);
	return ret;
}
	
/*
 * Set the link-layer device associated with a TOE.  For sniffing purposes any
 * messages sent to/received from the TOE will be associated with this device.
 */
void toe_set_lldev(struct toedev *dev, struct net_device *lldev)
{
	struct net_device *olddev = dev->lldev;

	if (lldev)
		dev_hold(lldev);
	dev->lldev = lldev;
	if (olddev)
		dev_put(olddev);
}

/*
 * Sends an sk_buff to a TOE driver after dealing with any active network taps.
 */
int toe_send(struct toedev *dev, struct sk_buff *skb)
{
	int r;

	local_bh_disable();
	if (unlikely(netdev_nit)) {     /* deal with active taps */
		skb->nh.raw = skb->data;
		skb->dev = dev->lldev;
		dev_queue_xmit_nit(skb, skb->dev);
	}
	r = dev->send(dev, skb);
	local_bh_enable();
	return r;
}

/**
 *	toe_receive_skb - process n received TOE packets
 *	@dev: the toe device
 *	@skb: an array of offload packets
 *	@n: the number of offload packets
 *
 *	Process an array of ingress offload packets.  Each packet is forwarded
 *	to any active network taps and then passed to the toe device's receive
 *	method.  We optimize passing packets to the receive method by passing
 *	it the whole array at once except when there are active taps.
 */
int toe_receive_skb(struct toedev *dev, struct sk_buff **skb, int n)
{
	int i;

	if (likely(!netdev_nit))
		return dev->recv(dev, skb, n);

	for (i = 0; i < n; ++i) {
		struct sk_buff *p = *skb++;

		p->dev = dev->lldev;
		skb_get(p);
		netif_receive_skb(p);
		p->dev = NULL;
		dev->recv(dev, &p, 1);
	}
	return 0;
}

void toe_neigh_update(struct neighbour *neigh, int flags)
{
	struct net_device *dev = neigh->dev;
	
	if (dev && (dev->features & NETIF_F_TCPIP_OFFLOAD)) {
		struct toedev *tdev = TOEDEV(dev);

		if (tdev && tdev->neigh_update)
			tdev->neigh_update(tdev, neigh, flags, neigh->dev);
	}
}

static int __init toedev_init(void)
{
#ifndef boot_phase
	if (!boot_phase)
		return 0;     /* duplicate call --- 2.4 only */
	boot_phase = 0;
#endif

	/* We tolerate proc failures */
	if (toe_proc_init())
		printk(KERN_WARNING "Unable to create /proc/net/toe entries\n");

	return 0;
}

subsys_initcall(toedev_init);

EXPORT_SYMBOL(register_tom);
EXPORT_SYMBOL(unregister_tom);
EXPORT_SYMBOL(register_toedev);
EXPORT_SYMBOL(alloc_toedev);
EXPORT_SYMBOL(activate_toedev);
EXPORT_SYMBOL(toe_set_lldev);
EXPORT_SYMBOL(toe_send);
EXPORT_SYMBOL(toe_receive_skb);
