/*
 * This file implements the Chelsio CPL5 message processing.
 *
 * Copyright (C) 2003 - 2004 Chelsio Communications.  All rights reserved.
 *
 * Written by Dimitris Michailidis (dm@chelsio.com)
 *
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 
 * Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/ip.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/toedev.h>
#include <net/tcp.h>
#include <net/offload.h>
#include <net/route.h>
#include <asm/atomic.h>
#include <asm/unaligned.h>
#include "../toe_compat.h"
#include "defs.h"
#include "tom.h"
#include "../cpl5_cmd.h"
#include "cpl5_defs.h"
#include "l2t.h"
#include "tcb.h"
#include "../toe_ctl_defs.h"


#ifdef CONFIG_CHELSIO_T1_ULP
#include "../ulp/ulp.h"
#endif

#ifdef CONFIG_CHELSIO_T1_BONDING
#include "bond.h"
#ifdef CONFIG_CHELSIO_T1_CM
#include "../cml/cml_osdep.h"
#endif
#endif

extern void __tcp_v4_hash(struct sock *sk, int listen_possible);
extern void __tcp_inherit_port(struct sock *sk, struct sock *child);

extern struct proto chelsio_t1_prot;
extern struct toe_funcs chelsio_t1_toe_funcs;

void t1_set_tx_wscale(struct sock *sk, int wscale);

/*
 * This sk_buff holds a fake header-only TCP segment that we use whenever we
 * need to exploit SW TCP functionality that expects TCP headers, such as
 * tcp_create_openreq_child().  It's a RO buffer that may be used by multiple
 * CPUs without locking.
 */
static struct sk_buff *tcphdr_skb;

#define SET_BLOG_CPL_HANDLER(skb, hnd) BLOG_SKB_CB(skb)->backlog_rcv = (hnd)

/*
 * Top-level CPL message processing used by most CPL messages that
 * pertain to connections.
 */
#define PROCESS_CPL_MSG(fn, sk, skb) \
    bh_lock_sock(sk); \
    if (sock_owned_by_user(sk)) { \
	SET_BLOG_CPL_HANDLER(skb, fn); \
    	sk_add_backlog(sk, skb); \
    } else \
	fn(sk, skb); \
    bh_unlock_sock(sk)

/*
 * Similar to PROCESS_CPL_MSG() but takes an extra socket reference around the
 * call to the handler.  Should be used if the handler may drop a socket
 * reference.
 */
#define PROCESS_CPL_MSG_REF(fn, sk, skb) \
    sock_hold(sk); \
    PROCESS_CPL_MSG(fn, sk, skb); \
    sock_put(sk);

/*
 * Returns an sk_buff for a reply CPL message of size len.  If the input
 * sk_buff has no other users it is trimmed and reused, otherwise a new buffer
 * is allocated.  The input skb must be of size at least len.  Note that this
 * operation does not destroy the original skb data even if it decides to reuse
 * the buffer.
 */
struct sk_buff *get_cpl_reply_skb(struct sk_buff *skb, size_t len, int gfp)
{
    /* XXX We never get shared skbs, but check to make sure. */
    if (skb_shared(skb))
	printk(KERN_ERR "Received shared skb!\n");

    if (!skb_cloned(skb)) {
	__skb_trim(skb, len);
	skb_get(skb);
    } else {
	skb = alloc_skb(len, gfp);
	if (skb) __skb_put(skb, len);
    }
    return skb;
}

/*
 * Like get_cpl_reply_skb() but the returned buffer starts out empty.
 */
struct sk_buff *__get_cpl_reply_skb(struct sk_buff *skb, size_t len, int gfp)
{
    if (!skb_cloned(skb) && ((unsigned long)skb->data & 3) == 0) {
	__skb_trim(skb, 0);
	skb_get(skb);
    } else
	skb = alloc_skb(len, gfp);
    return skb;
}

/*
 * Returns a new active-open TID or -1 if all TIDs are taken.
 */
static int alloc_atid(struct tid_info *t)
{
    int atid = -1;

    spin_lock_bh(&t->atid_lock);
    if (t->afree) {
	union active_open_entry *p = t->afree;

	atid = (p - t->atid_tab) + t->atid_base;
	t->afree = p->next;
	p->sk = NULL;
	t->atids_in_use++;
    }
    spin_unlock_bh(&t->atid_lock);
    return atid;
}

/*
 * Free an active-open TID.  We do not release the socket reference
 * automatically as the common case is that we continue to use the socket.
 */
static void free_atid(struct tid_info *t, tid_t atid)
{
    union active_open_entry *p = &ATID2ENTRY(t, atid);

    if (p->sk) sock_put(p->sk);
    spin_lock_bh(&t->atid_lock);
    p->next = t->afree;
    t->afree = p;
    t->atids_in_use--;
    spin_unlock_bh(&t->atid_lock);
}

/*
 * Insert a socket to the TID table.
 */
static inline void __sk_insert_tid(struct tom_data *d, struct sock *sk,
				   tid_t tid)
{
    struct tid_info *t = &d->tid_maps;

    t->tid_tab[tid] = sk;
    t->mss_tab[tid] = 0x3f;
    atomic_inc(&t->tids_in_use);
}

/*
 * Insert a socket to the TID table and take an extra reference.
 */
static inline void sk_insert_tid(struct tom_data *d, struct sock *sk,
				 tid_t tid)
{
    sock_hold(sk);
    __sk_insert_tid(d, sk, tid);
}

/*
 * Remove a socket from the TID table.  Be a bit paranoid.
 */
static inline void sk_remove_tid(struct tom_data *d, struct sock *sk,
				 tid_t tid)
{
    struct tid_info *t = &d->tid_maps;

    if (tid < t->ntids && t->tid_tab[tid] == sk) {
	t->tid_tab[tid] = NULL;
	atomic_dec(&t->tids_in_use);
	sock_put(sk);
    }
}

/*
 * Combination lookup when the type of the TID is unknown.
 */
struct sock *lookup_tid_any(const struct tid_info *t, tid_t tid)
{
    if (tid < t->ntids)
	return t->tid_tab[tid];
    if (tid >= t->stid_base && tid < t->stid_base + t->nstids)
	return STID2ENTRY(t, tid).sk;
    if (tid >= t->atid_base && tid < t->atid_base + t->natids)
	return ATID2ENTRY(t, tid).sk;
    return NULL;
}

#if VALIDATE_TID
/*
 * Returns true if a connection TID is in range and currently unused.
 */
int valid_new_tid(const struct tid_info *t, tid_t tid)
{
    return tid < t->ntids && !t->tid_tab[tid];
}

#define VALIDATE_SOCK(d, sk) \
    do { \
	if (!(sk)) \
	    return CPL_RET_UNKNOWN_TID | CPL_RET_BUF_DONE; \
    } while (0)
#else
#define VALIDATE_SOCK(d, sk) do {} while (0)
#endif

/*
 * Called when we receive the last message from HW for a connection.  A
 * connection cannot transition to TCP_CLOSE prior to this event.
 * Resources related to the offload state of a connection (e.g., L2T entries)
 * must have been relinquished prior to calling this.
 */
static void connection_done(struct sock *sk)
{
    struct tcp_opt *tp = tcp_sk(sk);

#if 0
    printk("tcp_done: state: %d, dead %d, refs %d\n", sk->sk_state,
	   sock_flag(sk, SOCK_DEAD), atomic_read(&sk->sk_refcnt));
#endif
#ifdef CONFIG_CHELSIO_T1_CM
    if (sock_flag(sk, SOCK_MIGRATE)) {
        t1_cml_sk_cleanup(sk, 0);
    }
#endif

    if (CLOSE_FLAGS(tp) & ABORT_RPL_PENDING) {
	CLOSE_FLAGS(tp) &= ~ABORT_RPL_PENDING;
	if (sk->sk_sleep && waitqueue_active(sk->sk_sleep))
	    wake_up_all(sk->sk_sleep);
    }

    tcp_done(sk);
}

/* Max receive window allowed by HW */
#define MAX_RCV_WND ((M_RCV_BUFSIZ << 10) | 0x3ff)

/*
 * Min receive window.  We want it to be large enough to accommodate receive
 * coalescing, handle jumbo frames, and not trigger sender SWS avoidance.
 */
#define MIN_RCV_WND (24 * 1024U)

/*
 * Determine the receive window scaling factor given a target max
 * receive window.
 */
static inline int select_rcv_wscale(int space, int wscale_ok)
{
    int wscale = 0;

    if (space > MAX_RCV_WND)
	space = MAX_RCV_WND;

    if (wscale_ok)
	for (; space > 65535 && wscale < 14; space >>= 1, ++wscale) ;
    return wscale;
}

/* Returns bits 2:7 of a socket's TOS field */
#define SK_TOS(sk) ((inet_sk(sk)->tos >> 2) & M_TOS)

/*
 * The next two functions calculate the option 0 value for a socket.
 */
static inline int calc_opt0h(struct sock *sk, struct tcp_opt *tp)
{
    return ((tp->nonagle & TCP_NAGLE_OFF) ? 0 : V_NAGLE) |
	(TOM_TUNABLE(TOE_DEV(tp), delack) ? V_DELACK : 0) |
	(sock_flag(sk, SOCK_KEEPOPEN) ? V_KEEP_ALIVE : 0) | V_TCAM_BYPASS |
	V_WND_SCALE(tp->recv_wscale);
}

static inline int calc_opt0l(struct sock *sk, struct tcp_opt *tp)
{
    return V_TOS(SK_TOS(sk)) | V_RCV_BUFSIZ(tp->rcv_wnd >> 10);
}

/*
 * Returns the TX window size.
 */
static inline u32 tx_window_size(struct tcp_opt *tp)
{
    return SND_MAX(tp) - tp->snd_una;
}

/*
 * Returns the available portion of the TX window, i.e., excluding any
 * currently unacked data.
 */
static inline u32 tx_avail_win(struct tcp_opt *tp)
{
    return SND_MAX(tp) - tp->snd_nxt;
}

/*
 * Calculate the max sequence number that can be accommodated by the number of
 * TX pages allocated to a connection.  Do not call this if there are no pages
 * allocated.
 */
static inline void update_sndmax(struct tcp_opt *tp, struct tom_data *d)
{
    SND_MAX(tp) = (tp->snd_una + ((TX_PAGES_ALLOCED(tp) - 1) << d->tx_page_shift)) | d->tx_page_mask;
}

/*
 * Returns the number of TX pages used by a connection.
 */
static inline int tx_pages_used(struct tcp_opt *tp, struct tom_data *d)
{
    return ((tp->snd_nxt - (tp->snd_una & ~d->tx_page_mask)) >>
	    d->tx_page_shift) + (tp->snd_una != tp->snd_nxt);
}

/*
 * Returns true if an sk_buff has been marked ready for transmission.
 */
static inline int skb_tx_ready(struct sk_buff *skb)
{
    return TCP_SKB_CB(skb)->flags & TCPCB_FLAG_PSH;
}

/*
 * Returns true if an sk_buff carries urgent data.
 */
static inline int skb_urgent(struct sk_buff *skb)
{
    return (TCP_SKB_CB(skb)->flags & TCPCB_FLAG_URG) != 0;
}

/*
 * Returns true if an sk_buff is ready to be sent to the TOE.
 */
static inline int snd_check(struct tcp_opt *tp, struct sk_buff *skb)
{
    return skb->len <= tx_avail_win(tp) && skb_tx_ready(skb);
}

/*
 * Generic ARP failure handler that discards the buffer.
 */
static void arp_failure_discard(struct toedev *dev, struct sk_buff *skb)
{
    kfree_skb(skb);
}

/*
 * Prepends CPL_TX_DATA or CPL_CLOSE_CON_REQ headers to buffers waiting in a
 * socket's send queue and sends them on to the TOE.  Must be called with the
 * socket lock held.  Returns the amount of send buffer space that was freed
 * as a result of sending queued data to the TOE.
 */
int t1_push_frames(struct sock *sk)
{
    int total_size = 0;
    struct tcp_opt *tp = tcp_sk(sk);
    struct toedev *dev;
    struct tom_data *d;

    if (unlikely(sk->sk_state == TCP_CLOSE)) return 0;

    dev = TOE_DEV(tp);
    if (unlikely(!dev)) {
	printk(KERN_ERR "%s: sk 0x%p, state 0x%x tdev NULL.\n",
			 __FUNCTION__, sk, sk->sk_state);
	WARN_ON(1);
        return 0;
    }
    d = TOM_DATA(dev);

    /*
     * This shouldn't really happen but just in case make sure we don't send
     * any data after an abort.
     */
    if (unlikely(CLOSE_FLAGS(tp) & ABORT_SHUTDOWN)) return 0;
    if (unlikely(CLOSE_FLAGS(tp) & TX_HALT)) return 0;

    while (skb_queue_len(&sk->sk_write_queue)) {
	struct sk_buff *skb = sk->sk_write_queue.next;
	int len = skb->len;                        /* length before skb_push */

	if (!snd_check(tp, skb))
	    break;

	__skb_unlink(skb, skb->list);
	skb->priority = CPL_PRIORITY_DATA;
	SET_ARP_FAILURE_HANDLER(skb, arp_failure_discard);

	if (unlikely(TCP_SKB_CB(skb)->flags & TCPCB_FLAG_FIN)) {
	    struct cpl_close_con_req *req;

	    req = (struct cpl_close_con_req *)__skb_put(skb, sizeof(*req));
	    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_CLOSE_CON_REQ, TID(tp)));
	    CLOSE_FLAGS(tp) |= CLOSE_CON_REQUESTED;
	} else {
	    struct cpl_tx_data *hdr;
	    hdr = (struct cpl_tx_data *)__skb_push(skb, sizeof(*hdr));
	    OPCODE_TID(hdr) = htonl(MK_OPCODE_TID(CPL_TX_DATA, TID(tp)));
	    hdr->len = htonl(len);
	    hdr->urg = htons(skb_urgent(skb));
	    hdr->flags = htons((d->max_nports > 1 ? V_TX_SHOVE : 0) |
			       V_TX_ULP_SUBMODE(skb_ulp_mode(skb)));
	    BUG_TRAP(TX_PAGES_ALLOCED(tp));
	}

	tp->lsndtime = tcp_time_stamp;
	tp->snd_nxt += len;
	total_size += skb->truesize;

#if VALIDATE_SEQ
	if (!between(tp->snd_nxt, tp->snd_una, SND_MAX(tp)))
	    printk(KERN_ERR
		   "TID %u window violation: snd_una %u, snd_max %u, snd_nxt %u, caller %p\n",
		   TID(tp), tp->snd_una, SND_MAX(tp), tp->snd_nxt,
		   __builtin_return_address(0));
#endif
	l2t_send(TOE_DEV(tp), skb, L2T_ENTRY(tp));
    }
    sk->sk_wmem_queued -= total_size;
    return total_size;
}

/* How long we remember PM pressure events */
#define PM_PRESSURE_HYSTERISIS (10 * HZ)

/*
 * True if PM memory is under pressure and needs to be conserved.  To avoid
 * wild swings in allocation decisions PM pressure events are remembered for
 * PM_PRESSURE_HYSTERISIS jiffies after they occur.
 */
static inline int pm_pressure(struct tom_data *d)
{
    if (d->tx_pages_free < (d->tx_pages >> 1)) {
	d->pm_pressure_ts = jiffies;
	d->pm_pressure = 1;
    } else if (time_after(jiffies, d->pm_pressure_ts + PM_PRESSURE_HYSTERISIS))
	d->pm_pressure = 0;
    return d->pm_pressure;
}

/*
 * Allocate up to n new pages for a connection.  Returns the number of pages
 * allocated.
 */
int t1_alloc_tx_pages(struct sock *sk, unsigned int n)
{
    struct tcp_opt *tp = tcp_sk(sk);
    struct toedev *dev = TOE_DEV(tp);
    struct tom_data *d = TOM_DATA(dev);

    if (!n || TX_PAGES_ALLOCED(tp) + n > d->conf.max_tx_pages)
	n = d->conf.max_tx_pages - TX_PAGES_ALLOCED(tp);
    if (!n) return 0;

    spin_lock_bh(&d->tx_page_lock);
    if (pm_pressure(d)) {
	n = TX_PAGES_ALLOCED(tp) ? 0 : 2;
	if (!n) goto unlock;
    }

    if (d->tx_pages_free >= n) {
	d->tx_pages_free -= n;
	if (!TX_PAGES_ALLOCED(tp)) d->tx_page_users++;
    } else if (!TX_PAGES_ALLOCED(tp) && d->tx_pages_free >= 2) {
	d->tx_pages_free -= 2;
	d->tx_page_users++;
	n = 2;
    } else
	n = 0;
 unlock:
    spin_unlock_bh(&d->tx_page_lock);

    if (n) {
	TX_PAGES_ALLOCED(tp) += n;
	update_sndmax(tp, d);
    }
    return n;
}

/*
 * Relinquish n TX pages from a given connection.
 */
static void free_tx_pages(struct tcp_opt *tp, unsigned int n)
{
    struct tom_data *d = TOM_DATA(TOE_DEV(tp));

    TX_PAGES_ALLOCED(tp) -= n;
    spin_lock_bh(&d->tx_page_lock);
    if (!TX_PAGES_ALLOCED(tp)) {
	SND_MAX(tp) = tp->snd_una;
	d->tx_page_users--;
    }
    d->tx_pages_free += n;
    spin_unlock_bh(&d->tx_page_lock);
}

/*
 * Release resources held by an offload connection (TID, L2T entry, etc.)
 */
void t1_release_offload_resources(struct sock *sk)
{
    struct tom_data *d;
    struct tcp_opt *tp = tcp_sk(sk);
    struct toedev *dev = TOE_DEV(tp);
    tid_t tid = TID(tp);

    if (!dev) return;

    if (TX_PAGES_ALLOCED(tp))
	free_tx_pages(tp, TX_PAGES_ALLOCED(tp));

    if (L2T_ENTRY(tp)) {
	l2t_release(L2DATA(dev), L2T_ENTRY(tp));
	L2T_ENTRY(tp) = NULL;
    }
    d = TOM_DATA(dev);
    if (sk->sk_state == TCP_SYN_SENT) {            // we have ATID
	BUG_TRAP(tid >= d->tid_maps.atid_base);
	free_atid(&d->tid_maps, tid);
	__skb_queue_purge(&tp->out_of_order_queue);
    } else                                         // we have TID
	sk_remove_tid(d, sk, tid);

    TOE_DEV(tp) = NULL;
}

/*
 * Called for each sk_buff in a socket's receive backlog during
 * backlog processing.
 */
static int t1_backlog_rcv(struct sock *sk, struct sk_buff *skb)
{
    if (likely(BLOG_SKB_CB(skb)->backlog_rcv != NULL))
	BLOG_SKB_CB(skb)->backlog_rcv(sk, skb);
    else {
	printk(KERN_ERR "no skb backlog handler in %s\n", __FUNCTION__);
	kfree_skb(skb);
    }
    return 0;
}

/*
 * Switch a socket to the offload protocol operations.  Note that the offload
 * operations do not contain the offload backlog handler, we install that
 * directly to the socket.
 */
static inline void install_offload_ops(struct sock *sk)
{
    sk->sk_prot = &chelsio_t1_prot;
    sk->sk_backlog_rcv = t1_backlog_rcv;
    sk->sk_write_space = t1_write_space;
    tcp_sk(sk)->toe_specific = &chelsio_t1_toe_funcs;
}

/*
 * Convert an ACT_OPEN_RPL status to a Linux errno.
 */
static int act_open_rpl_status_to_errno(int status)
{
    switch (status) {
	case CPL_ERR_CONN_RESET:
	    return ECONNREFUSED;
	case CPL_ERR_ARP_MISS:
	    return EHOSTUNREACH;
	case CPL_ERR_CONN_TIMEDOUT:
	    return ETIMEDOUT;
	case CPL_ERR_TCAM_FULL:
	    return ENOMEM;
	case CPL_ERR_CONN_EXIST:
	    printk(KERN_ERR "ACTIVE_OPEN_RPL: 4-tuple in use\n");
	    return EADDRINUSE;
	default:
	    return EIO;
    }
}

/*
 * Handle active open failures.
 */
static void active_open_failed(struct sock *sk, struct sk_buff *skb)
{
    struct cpl_act_open_rpl *rpl = cplhdr(skb);

    sk->sk_err = act_open_rpl_status_to_errno(rpl->status);
    sk->sk_error_report(sk);
    t1_release_offload_resources(sk);
    connection_done(sk);
    __kfree_skb(skb);
    TCP_INC_STATS_BH(TCP_MIB_ATTEMPTFAILS);
}

#ifdef CONFIG_CHELSIO_T1_CM
static void process_active_open_cm(struct sock *sk, struct sk_buff *skb)
{
    struct cpl_act_open_rpl *rpl = cplhdr(skb);
    tid_t atid = GET_TID(rpl);
    struct tcp_opt *tp = tcp_sk(sk);
    struct tom_data *d = TOM_DATA(TOE_DEV(tp));
    tid_t tid = G_PASS_OPEN_TID(ntohl(rpl->new_tid));
                                                                                
    if (!rpl->status) {
        /* some book-keeping */
        TID(tcp_sk(sk)) = tid;
        sk_insert_tid(d, sk, tid);
        free_atid(&d->tid_maps, atid);
        dst_confirm(sk->sk_dst_cache);
        t1_cm_rx_cpl(sk, skb);
    } else {
        t1_cml_sk_abort(sk, 0);
        active_open_failed(sk, skb);
    }
}
#endif

/*
 * Process an ACT_OPEN_RPL CPL message.
 */
static int do_act_open_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_act_open_rpl *rpl = cplhdr(skb);
    tid_t atid = GET_TID(rpl);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_atid(&d->tid_maps, atid);

    VALIDATE_SOCK(d, sk);

#ifdef CONFIG_CHELSIO_T1_CM
    if (sock_flag(sk, SOCK_MIGRATE)) {
        PROCESS_CPL_MSG_REF(process_active_open_cm, sk, skb);
        return 0;
    }
#endif

    PROCESS_CPL_MSG_REF(active_open_failed, sk, skb);
    return 0;
}

/*
 * Handle an ARP failure for an active open.   XXX purge ofo queue
 *
 * XXX badly broken for crossed SYNs as the ATID is no longer valid.
 * XXX crossed SYN errors should be generated by PASS_ACCEPT_RPL which should
 * check SOCK_DEAD or sk->sk_sock.  Or maybe generate the error here but don't
 * to free the atid.  Hmm.
 */
static void act_open_req_arp_failure(struct toedev *dev, struct sk_buff *skb)
{
    struct sock *sk = skb->sk;

    sock_hold(sk);
    bh_lock_sock(sk);
    if (sk->sk_state == TCP_SYN_SENT || sk->sk_state == TCP_SYN_RECV) {
	if (!sock_owned_by_user(sk)) {
	    sk->sk_err = EHOSTUNREACH;
	    sk->sk_error_report(sk);
	    t1_release_offload_resources(sk);
	    connection_done(sk);
	    __kfree_skb(skb);
	    TCP_INC_STATS_BH(TCP_MIB_ATTEMPTFAILS);
	} else {
	    /*
	     * Smart solution: Synthesize an ACTIVE_OPEN_RPL in the existing
	     * sk_buff and queue it to the backlog.  We are certain the sk_buff
	     * is not shared.  We also don't bother trimming the buffer.
	     */
	    struct cpl_act_open_rpl *rpl = cplhdr(skb);

	    rpl->ot.opcode = CPL_ACT_OPEN_RPL;
	    rpl->status = CPL_ERR_ARP_MISS;
	    SET_BLOG_CPL_HANDLER(skb, active_open_failed);
	    sk_add_backlog(sk, skb);

	    /*
	     * XXX Make sure a PASS_ACCEPT_RPL behind us doesn't destroy the
	     * socket.  Unfortunately we can't go into SYN_SENT because we
	     * don't have an atid.  Needs more thought.
	     */
	}
    }
    bh_unlock_sock(sk);
    sock_put(sk);
}

/*
 * Determine the receive window size for a socket.
 */
static unsigned int select_rcv_wnd(struct sock *sk)
{
    unsigned int wnd = tcp_full_space(sk);

    return wnd > MAX_RCV_WND ? MAX_RCV_WND : wnd;
}

/*
 * Send an active open request.
 */
int t1_connect(struct toedev *dev, struct sock *sk)
{
    struct sk_buff *skb;
    struct l2t_entry *e;
    struct tom_data *d = TOM_DATA(dev);
    struct tcp_opt *tp = tcp_sk(sk);
    struct cpl_act_open_req *req;
    struct net_device *lldev;

    int atid = alloc_atid(&d->tid_maps);
    if (atid < 0) goto out_err;

    e = l2t_get(dev, __sk_dst_get(sk)->neighbour);
    if (!e) goto free_tid;

#ifdef CONFIG_CHELSIO_T1_CM
    if (sock_flag(sk, SOCK_MIGRATE)) {
        skb = alloc_skb(sizeof(*req), GFP_ATOMIC);
        if (!skb) goto free_tid;
    } else
#endif
    skb = alloc_skb_nofail(sizeof(*req));
    skb->priority = CPL_PRIORITY_SETUP;
    skb->sk = sk;
    SET_ARP_FAILURE_HANDLER(skb, act_open_req_arp_failure);

    sock_hold(sk);
    ATID2ENTRY(&d->tid_maps, atid).sk = sk;

    install_offload_ops(sk);

    tp->mss_cache = d->conf.mss;
    L2T_ENTRY(tp) = e;
    TOE_DEV(tp) = dev;
    TID(tp) = atid;
    TX_PAGES_ALLOCED(tp) = 0;
    CLOSE_FLAGS(tp) = 0;
    tp->rcv_wnd = max(select_rcv_wnd(sk), MIN_RCV_WND);
    tp->recv_wscale = select_rcv_wscale(tcp_full_space(sk),
				        sysctl_tcp_window_scaling);
    sk->sk_err = 0;
    sock_reset_flag(sk, SOCK_DONE);
    TCP_INC_STATS(TCP_MIB_ACTIVEOPENS);

    req = (struct cpl_act_open_req *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_ACT_OPEN_REQ, atid));
    req->local_port = inet_addr_info(sk)->sport;
    req->peer_port = inet_addr_info(sk)->dport;
    req->local_ip = inet_addr_info(sk)->saddr;
    req->peer_ip = inet_addr_info(sk)->daddr;
    lldev = __sk_dst_get(sk)->dev;
    req->opt0h = htonl(calc_opt0h(sk, tp) |
			V_MSS_IDX_OVERRIDE |
			V_MSS_IDX(best_mtu_index(dev, lldev->mtu)) |
			V_L2T_IDX(l2e_idx(L2DATA(dev), e)));
    req->opt0l = htonl(calc_opt0l(sk, tp));
#ifdef CONFIG_CHELSIO_T1_CM
    /* set non-offload bit and timer inject bit */
    if (sock_flag(sk, SOCK_MIGRATE)) {
        req->opt0l |= htonl(0x80 | 0x40);
    }
#endif
    req->iff_vlantag = htonl(0);

    l2t_send(dev, skb, e);
    return 0;

 free_tid:
    free_atid(&d->tid_maps, atid);
 out_err:
    return -1;
}

/*
 * Handle an ARP failure for a CPL_ABORT_REQ.  Change it into a no RST variant
 * and send it along.
 */
static void abort_arp_failure(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_abort_req *req = cplhdr(skb);

    req->cmd = CPL_ABORT_NO_RST;
    toe_send(dev, skb);
}

/*
 * Send an ABORT_REQ message.  Cannot fail.  This routine makes sure we do
 * not send multiple ABORT_REQs for the same connection and also that we do
 * not try to send a message after the connection has closed.  Returns 1 if
 * an ABORT_REQ wasn't generated after all, 0 otherwise.
 */
int t1_send_reset(struct sock *sk, int mode, struct sk_buff *skb)
{
    struct cpl_abort_req *req;
    struct tcp_opt *tp = tcp_sk(sk);

    if (unlikely((CLOSE_FLAGS(tp) & ABORT_SHUTDOWN) || !TOE_DEV(tp))) {
	if (skb) __kfree_skb(skb);
	return 1;
    }

    CLOSE_FLAGS(tp) |= ABORT_SHUTDOWN | ABORT_RPL_PENDING;

    /* Purge the send queue so we don't send anything after an abort. */
    t1_purge_write_queue(sk);

    if (CLOSE_FLAGS(tp) & CLOSE_CON_REQUESTED)
	mode |= CPL_ABORT_POST_CLOSE_REQ;

    if (!skb) skb = alloc_skb_nofail(sizeof(*req));
    skb->priority = CPL_PRIORITY_DATA;
    SET_ARP_FAILURE_HANDLER(skb, abort_arp_failure);

    req = (struct cpl_abort_req *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_ABORT_REQ, TID(tp)));
    req->cmd = mode;
    if (sk->sk_state == TCP_SYN_SENT)
	__skb_queue_tail(&tp->out_of_order_queue, skb);  // defer
    else
	l2t_send(TOE_DEV(tp), skb, L2T_ENTRY(tp));
    return 0;
}

/*
 * Reset a connection that is on a listener's SYN queue or accept queue,
 * i.e., one that has not had a struct socket associated with it.
 */
static void reset_listen_child(struct sock *child)
{
    struct sk_buff *skb = alloc_skb_nofail(sizeof(struct cpl_abort_req));

    sock_hold(child);       // need to survive past tcp_destroy_sock()
    local_bh_disable();
    bh_lock_sock(child);

    t1_send_reset(child, CPL_ABORT_SEND_RST, skb);
    sock_orphan(child);
    atomic_inc(&tcp_orphan_count);  // tcp_destroy_sock() decrements it
    if (child->sk_state == TCP_CLOSE)
	tcp_destroy_sock(child);

    bh_unlock_sock(child);
    local_bh_enable();
    sock_put(child);
}

/*
 * The reap list is the list of passive open sockets that were orphaned when
 * their listening parent went away and wasn't able to nuke them for whatever
 * reason.  These sockets are terminated through a work request from process
 * context.
 */
static struct sock *reap_list;
static spinlock_t reap_list_lock = SPIN_LOCK_UNLOCKED;

/*
 * Process the reap list.
 */
static void process_reap_list(void *data)
{
    spin_lock_bh(&reap_list_lock);
    while (reap_list) {
	struct sock *sk = reap_list;

	reap_list = sk->sk_pair;
	sk->sk_pair = NULL;
	spin_unlock_bh(&reap_list_lock);
	reset_listen_child(sk);
	spin_lock_bh(&reap_list_lock);
    }
    spin_unlock_bh(&reap_list_lock);
}

static DECLARE_WORK(reap_task, process_reap_list, NULL);

/*
 * Add a socket to the reap list and schedule a work request to process it.
 * We thread sockets through their sk_pair pointers.  Should be called only
 * from softirq context and any associated open request should have already
 * been freed.
 */
static void add_to_reap_list(struct sock *sk)
{
    BUG_TRAP(in_softirq());
    BUG_TRAP(!sk->sk_pair);
    BUG_TRAP(!sk->sk_user_data);

    if (is_bound(sk))    // release the port immediately --- it may be reused
	tcp_put_port(sk);

    spin_lock(&reap_list_lock);
    sk->sk_pair = reap_list;
    reap_list = sk;
    if (!sk->sk_pair)
	schedule_work(&reap_task);
    spin_unlock(&reap_list_lock);
}

#define MK_SET_TCB_FIELD_REQ(skb, tid, field, value) { \
    struct cpl_set_tcb_field *req; \
    req = (struct cpl_set_tcb_field *)__skb_put(skb, sizeof(*req)); \
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SET_TCB_FIELD, tid)); \
    req->offset = W_TCB_ ## field; \
    req->mask = htonl(M_TCB_ ## field << S_TCB_ ## field); \
    req->val = htonl((value) << S_TCB_ ## field); \
}

/*
 * Send a SET_TCB_FIELD CPL message to change a connection's Nagle setting.
 */
void t1_set_nagle(struct sock *sk)
{
    struct tcp_opt *tp = tcp_sk(sk);
    struct sk_buff *skb = alloc_skb_nofail(sizeof(struct cpl_set_tcb_field));

    MK_SET_TCB_FIELD_REQ(skb, TID(tp), NAGLE, !(tp->nonagle & TCP_NAGLE_OFF));
    skb->priority = CPL_PRIORITY_CONTROL;
    if (sk->sk_state == TCP_SYN_SENT)
	__skb_queue_tail(&tp->out_of_order_queue, skb);  // defer
    else
	toe_send(TOE_DEV(tp), skb);
}

/*
 * Send a SET_TCB_FIELD CPL message to change a connection's TOS setting.
 */
void t1_set_tos(struct sock *sk)
{
    struct tcp_opt *tp = tcp_sk(sk);
    struct sk_buff *skb = alloc_skb_nofail(sizeof(struct cpl_set_tcb_field));

    MK_SET_TCB_FIELD_REQ(skb, TID(tp), TOS, SK_TOS(sk));
    skb->priority = CPL_PRIORITY_CONTROL;
    if (sk->sk_state == TCP_SYN_SENT)
	__skb_queue_tail(&tp->out_of_order_queue, skb);  // defer
    else
	toe_send(TOE_DEV(tp), skb);
}

/*
 * Send a SET_TCB_FIELD CPL message to change a connection's keepalive setting.
 */
void t1_set_keepalive(struct sock *sk, int on_off)
{
    struct tcp_opt *tp = tcp_sk(sk);
    struct sk_buff *skb = alloc_skb_nofail(sizeof(struct cpl_set_tcb_field));

    MK_SET_TCB_FIELD_REQ(skb, TID(tp), KEEPALIVE, on_off);
    skb->priority = CPL_PRIORITY_CONTROL;
    if (sk->sk_state == TCP_SYN_SENT)
	__skb_queue_tail(&tp->out_of_order_queue, skb);  // defer
    else
	toe_send(TOE_DEV(tp), skb);
}

/*
 * Send a SET_TCB_FIELD CPL message to change a connection's TX window scale.
 */
void t1_set_tx_wscale(struct sock *sk, int wscale)
{
    struct tcp_opt *tp = tcp_sk(sk);
    struct sk_buff *skb;

    skb = alloc_skb(sizeof(struct cpl_set_tcb_field), gfp_any());
    if (!skb) return;   // this call is best-effort
    
    MK_SET_TCB_FIELD_REQ(skb, TID(tp), SEND_WSCALE, wscale);
    skb->priority = CPL_PRIORITY_DATA;
    if (sk->sk_state == TCP_SYN_SENT)
	__skb_queue_tail(&tp->out_of_order_queue, skb);  // defer
    else
	toe_send(TOE_DEV(tp), skb);
}

/*
 * Send a SET_TCB_FIELD CPL message to change a connection's ULP mode.
 */
void t1_set_ulp_mode(struct sock *sk, int page_size)
{
    struct cpl_set_tcb_field *req;
    struct tcp_opt *tp = tcp_sk(sk);
    struct sk_buff *skb = alloc_skb_nofail(sizeof(*req));

    req = (struct cpl_set_tcb_field *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_SET_TCB_FIELD, TID(tp)));
    req->offset = W_TCB_ULP_MODE;
    req->mask = htonl((M_TCB_ULP_SUBMODE << S_TCB_ULP_SUBMODE) |
		      (M_TCB_ULP_PAGE_SIZE << S_TCB_ULP_PAGE_SIZE) |
		      (M_TCB_ULP_MODE << S_TCB_ULP_MODE));
    req->val = htonl((ULP_MODE(tp) << S_TCB_ULP_MODE) |
		     (ULP_SUBMODE(tp) << S_TCB_ULP_SUBMODE) |
		     (page_size << S_TCB_ULP_PAGE_SIZE));
    skb->priority = CPL_PRIORITY_DATA;
    if (sk->sk_state == TCP_SYN_SENT)
	__skb_queue_tail(&tp->out_of_order_queue, skb);  // defer
    else
	toe_send(TOE_DEV(tp), skb);
}

/*
 * Send RX credits through an RX_DATA_ACK CPL message.  If nofail is 0 we are
 * permitted to return without sending the message in case we cannot allocate
 * an sk_buff.  Returns the number of credits sent.
 */
u32 t1_send_rx_credits(struct toedev *dev, tid_t tid, u32 credits, int nofail)
{
    struct cpl_rx_data_ack *req;
    struct sk_buff *skb;

    skb = nofail ? alloc_skb_nofail(sizeof(*req)) : alloc_skb(sizeof(*req),
							      GFP_ATOMIC);
    if (!skb) return 0;
    req = (struct cpl_rx_data_ack *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_RX_DATA_ACK, tid));
    req->credit = htonl((1 << 28) | credits);
    skb->priority = CPL_PRIORITY_ACK;
    toe_send(dev, skb);
    return credits;
}

/*
 * Returns true if a socket cannot accept new RX data.
 */
static inline int sk_no_receive(struct sock *sk)
{
    return ((1 << sk->sk_state) & (TCPF_FIN_WAIT1 | TCPF_FIN_WAIT2) &&
	    (sk->sk_shutdown & RCV_SHUTDOWN));
}

/*
 * Returns true if we need to explicitly request RST when we receive new data
 * on an RX-closed connection.
 */
static inline int need_rst_on_excess_rx(struct tcp_opt *tp)
{
    return 1;
}

/*
 * Process new data received for a connection.
 */
#ifndef CONFIG_CHELSIO_T1_CM
static 
#endif
void new_rx_data(struct sock *sk, struct sk_buff *skb)
{
    struct tcp_opt *tp = tcp_sk(sk);

    if (unlikely(sk_no_receive(sk))) {
	if (need_rst_on_excess_rx(tp) && !(CLOSE_FLAGS(tp) & ABORT_SHUTDOWN)) {
	    struct sk_buff *abort_skb;

	    abort_skb = __get_cpl_reply_skb(skb, sizeof(struct cpl_abort_req),
					    GFP_ATOMIC);
	    if (abort_skb)
		t1_send_reset(sk, CPL_ABORT_SEND_RST, abort_skb);
	}
	kfree_skb(skb);
	return;
    }

#if VALIDATE_SEQ
    if (TCP_SKB_CB(skb)->seq != tp->rcv_nxt) {
	printk(KERN_ERR
	       "%s: TID %u: Bad sequence number, got %u, expected %u\n",
	       TOE_DEV(tp)->name, TID(tp), TCP_SKB_CB(skb)->seq, tp->rcv_nxt);
	__kfree_skb(skb);
	return;
    }
#endif

#ifndef CONFIG_CHELSIO_T1_CM
    __skb_pull(skb, sizeof(struct cpl_rx_data));
#endif
    tp->rcv_nxt += skb->len;
    tp->ack.lrcvtime = tcp_time_stamp;
#ifdef CONFIG_CHELSIO_T1_ULP
    if (unlikely(ULP_MODE(tp))) {
	ulp_new_rx_data(sk, skb);
    } else 
#endif
    {
        __skb_queue_tail(&sk->sk_receive_queue, skb);
	sk->sk_data_ready(sk, 0);
    }
}

/*
 * Handler for RX_DATA CPL messages.  Note that RX_DATA header may occasionally
 * be only 2-byte aligned.
 */
static int do_rx_data(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_rx_data *hdr = cplhdr(skb);
    tid_t tid = G_TID(ntohl(get_unaligned(&OPCODE_TID(hdr))));  // open code
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

#if VALIDATE_LEN
    if (unlikely(skb->len != ntohl(get_unaligned(&hdr->len)) + sizeof(*hdr))) {
	printk(KERN_ERR "%s: TID %u: Wrong RX_DATA length, CPL %lu, skb %u\n",
	       dev->name, tid, ntohl(get_unaligned(&hdr->len)) + sizeof(*hdr),
	       skb->len);
	return CPL_RET_BUF_DONE;
    }
#endif

    /* We may get RX_DATA for a bad TID, e.g., in auto mode */
    if (!sk) return CPL_RET_BUF_DONE;

    TCP_SKB_CB(skb)->seq = ntohl(get_unaligned(&hdr->seq));
    skb->h.th = tcphdr_skb->h.th;

#ifdef CONFIG_CHELSIO_T1_CM
    __skb_pull(skb, sizeof(struct cpl_rx_data));
#endif
    PROCESS_CPL_MSG(new_rx_data, sk, skb);
    return 0;
}

#ifdef CONFIG_CHELSIO_T1_ULP
/*
 * Handler for RX_ISCSI_HDR CPL messages.
 */
int do_rx_iscsi_hdr(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_rx_data *hdr = cplhdr(skb);
    tid_t tid = G_TID(ntohl(get_unaligned(&OPCODE_TID(hdr))));
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    if (!sk) return CPL_RET_BUF_DONE;

    TCP_SKB_CB(skb)->seq = ntohl(get_unaligned(&hdr->seq));
    skb->h.th = tcphdr_skb->h.th;

    PROCESS_CPL_MSG(ulp_new_rx_iscsi_hdr, sk, skb);

    return 0;
}

int do_rx_data_ddp(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_rx_data_ddp *hdr = cplhdr(skb);
    struct tom_data *d = TOM_DATA(dev);
    tid_t tid = G_TID(ntohl(get_unaligned(&OPCODE_TID(hdr))));
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    if (!sk) return CPL_RET_BUF_DONE; 

    PROCESS_CPL_MSG(ulp_new_rx_data_ddp, sk, skb);

    return 0;
}
#endif

/*
 * Handle a peer FIN.
 */
static void do_peer_fin(struct sock *sk, struct sk_buff *skb)
{
    struct tcp_opt *tp = tcp_sk(sk);
    int dead = sock_flag(sk, SOCK_DEAD);

    sk->sk_shutdown |= RCV_SHUTDOWN;
    sock_set_flag(sk, SOCK_DONE);

    switch (sk->sk_state) {
	case TCP_SYN_RECV:
	case TCP_ESTABLISHED:
	    tcp_set_state(sk, TCP_CLOSE_WAIT);
	    break;
	case TCP_FIN_WAIT1:
	    tcp_set_state(sk, TCP_CLOSING);
	    break;
	case TCP_FIN_WAIT2:
	    /* XXX enter TIME_WAIT through tcp_time_wait */
	    if (!(CLOSE_FLAGS(tp) & ABORT_RPL_PENDING)) {
		t1_release_offload_resources(sk);
		connection_done(sk);
	    }
	    break;
	default:
	    printk(KERN_ERR "%s: TID %u received PEER_CLOSE in bad state %d\n",
		   TOE_DEV(tp)->name, TID(tp), sk->sk_state);
    }

    if (!dead) {
	sk->sk_state_change(sk);

	/* Do not send POLL_HUP for half duplex close. */
	if ((sk->sk_shutdown & SEND_SHUTDOWN) || sk->sk_state == TCP_CLOSE)
	    sk_wake_async(sk, 1, POLL_HUP);
	else
	    sk_wake_async(sk, 1, POLL_IN);
    }
    __kfree_skb(skb);
}

/*
 * Handler for PEER_CLOSE CPL messages.
 */
static int do_peer_close(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_peer_close *req = cplhdr(skb);
    tid_t tid = GET_TID(req);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    /* In auto mode there may not be a socket */
    if (!sk) return CPL_RET_BUF_DONE;

    VALIDATE_SOCK(d, sk);

    PROCESS_CPL_MSG_REF(do_peer_fin, sk, skb);
    return 0;
}

/*
 * Process a peer ACK to our FIN.
 */
static void process_close_con_rpl(struct sock *sk, struct sk_buff *skb)
{
    switch (sk->sk_state) {
	case TCP_CLOSING:
	    // XXX tcp_time_wait but for now just destroy
	case TCP_LAST_ACK:
	    if (!(CLOSE_FLAGS(tcp_sk(sk)) & ABORT_RPL_PENDING)) {
		t1_release_offload_resources(sk);
		connection_done(sk);
	    }
	    break;
	case TCP_FIN_WAIT1:
	    tcp_set_state(sk, TCP_FIN_WAIT2);
	    sk->sk_shutdown |= SEND_SHUTDOWN;
	    dst_confirm(sk->sk_dst_cache);

	    if (!sock_flag(sk, SOCK_DEAD))
		sk->sk_state_change(sk);    // Wake up lingering close()
	    else {
		// XXX if linger2 < 0 send best effort RST
	    }
	    break;
	default: 
	    printk(KERN_ERR
		   "%s: TID %u received CLOSE_CON_RPL in bad state %d\n",
		   TOE_DEV(tcp_sk(sk))->name, TID(tcp_sk(sk)), sk->sk_state);
    }
    __kfree_skb(skb);
}

/*
 * Handler for CLOSE_CON_RPL CPL messages.
 */
static int do_close_con_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_close_con_rpl *rpl = cplhdr(skb);
    tid_t tid = GET_TID(rpl);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    VALIDATE_SOCK(d, sk);

    PROCESS_CPL_MSG_REF(process_close_con_rpl, sk, skb);
    return 0;
}

/*
 * Process abort replies.  We only process these messages if we anticipate
 * them as the coordination between SW and HW in this area is somewhat lacking
 * and sometimes we get ABORT_RPLs after we are done with the connection that
 * originated the ABORT_REQ.
 */
static void process_abort_rpl(struct sock *sk, struct sk_buff *skb)
{
    int free_skb = 1;
#ifdef CONFIG_CHELSIO_T1_CM
    /* update cm state machine */
    if (sock_flag(sk, SOCK_MIGRATE)) {
       t1_cm_rx_cpl(sk, skb);
       free_skb = 0;
    }
#endif

    if (CLOSE_FLAGS(tcp_sk(sk)) & ABORT_RPL_PENDING) {
	t1_release_offload_resources(sk);
	connection_done(sk);
    }
    if (free_skb) __kfree_skb(skb);
}

/*
 * Handle an ABORT_RPL CPL message.
 */
static int do_abort_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_abort_rpl *rpl = cplhdr(skb);
    tid_t tid = GET_TID(rpl);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    /*
     * It's OK if we don't find a socket, sometimes we may not want to wait
     * for the abort reply.
     */
    if (!sk) return CPL_RET_BUF_DONE;

    PROCESS_CPL_MSG_REF(process_abort_rpl, sk, skb);
    return 0;
}

/*
 * Convert the status code of an ABORT_REQ into a Linux error code.  Also
 * indicate whether RST should be sent in response.
 */
static int abort_status_to_errno(struct sock *sk, int abort_reason,
				 int *need_rst)
{
    switch (abort_reason) {
	case CPL_ERR_BAD_SYN:
	    NET_INC_STATS_BH(LINUX_MIB_TCPABORTONSYN);  // fall through
	case CPL_ERR_CONN_RESET:
	    // XXX need to handle SYN_RECV due to crossed SYNs
	    return sk->sk_state == TCP_CLOSE_WAIT ? EPIPE : ECONNRESET;
	case CPL_ERR_XMIT_TIMEDOUT:
	case CPL_ERR_PERSIST_TIMEDOUT:
	case CPL_ERR_FINWAIT2_TIMEDOUT:
	case CPL_ERR_KEEPALIVE_TIMEDOUT:
	    NET_INC_STATS_BH(LINUX_MIB_TCPABORTONTIMEOUT); // XXX need_rst?
	    return ETIMEDOUT;
	default:
	    return EIO;
    }
}

static void send_deferred_abort_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_abort_rpl *rpl;
    struct cpl_abort_req *req = cplhdr(skb);
    tid_t tid = GET_TID(req);
    struct sk_buff *reply_skb = alloc_skb_nofail(sizeof(*rpl));

    reply_skb->priority = CPL_PRIORITY_DATA;
    __skb_put(reply_skb, sizeof(*rpl));
    rpl = cplhdr(reply_skb);
    OPCODE_TID(rpl) = htonl(MK_OPCODE_TID(CPL_ABORT_RPL, tid));
    rpl->status = req->cmd;
    toe_send(dev, reply_skb);      // XXX not right, need to sync with ARP
    kfree_skb(skb);
}

/*
 * Process abort requests.  If we are waiting for an ABORT_RPL we ignore this
 * request except that we need to reply to it.
 */
static void process_abort_req(struct sock *sk, struct sk_buff *skb)
{
    struct sk_buff *reply_skb;
    struct cpl_abort_rpl *rpl;
    struct cpl_abort_req *req = cplhdr(skb);
    tid_t tid = GET_TID(req);
    struct tcp_opt *tp = tcp_sk(sk);
    struct toedev *dev = TOE_DEV(tp);
    int rst_status = CPL_ABORT_NO_RST;

    if (!(CLOSE_FLAGS(tp) & ABORT_RPL_PENDING)) {
	sk->sk_err = abort_status_to_errno(sk, req->cmd, &rst_status);
	if (!sock_flag(sk, SOCK_DEAD))
	    sk->sk_error_report(sk);
	t1_release_offload_resources(sk);
	connection_done(sk);
    }

    reply_skb = get_cpl_reply_skb(skb, sizeof(*rpl), GFP_ATOMIC);
    if (!reply_skb) {
	/* Defer the reply.  Stick rst_status into req->cmd. */
	req->cmd = rst_status;
	t1_defer_reply(skb, TOE_DEV(tp), send_deferred_abort_rpl);
	return;
    }

    kfree_skb(skb);                             // can't use __kfree_skb here
    reply_skb->priority = CPL_PRIORITY_DATA;
    rpl = cplhdr(reply_skb);
    OPCODE_TID(rpl) = htonl(MK_OPCODE_TID(CPL_ABORT_RPL, tid));
    rpl->status = rst_status;
    toe_send(dev, reply_skb); // XXX not right, need to sync with ARP
}

/*
 * Handle an ABORT_REQ CPL message.
 */
static int do_abort_req(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_abort_req *req = cplhdr(skb);
    tid_t tid = GET_TID(req);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    VALIDATE_SOCK(d, sk);

    PROCESS_CPL_MSG_REF(process_abort_req, sk, skb);
    return 0;
}

/*
 * Process a CPL_L2T_WRITE_RPL.  Just log a message if we get bad status.
 * Note that the TID in the reply is really the L2T index it refers to.
 */
static int do_l2t_write_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_l2t_write_rpl *rpl = cplhdr(skb);

    if (rpl->status != CPL_ERR_NONE)
	printk(KERN_ERR
	       "Unexpected status %d for CPL_L2T_WRITE_RPL for entry %d\n",
	       rpl->status, GET_TID(rpl));

    return CPL_RET_BUF_DONE;
}

/*
 * Process a SET_TCB_RPL CPL message.  Just log a message if we get bad status.
 */
static int do_set_tcb_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_set_tcb_rpl *rpl = cplhdr(skb);

    if (rpl->status != CPL_ERR_NONE)
	printk(KERN_ERR
	       "Unexpected status %d for CPL_SET_TCB_RPL for tid %u\n",
	       rpl->status, GET_TID(rpl));

#ifdef CONFIG_CHELSIO_T1_CM
    else {
        tid_t tid =  GET_TID(rpl);
        struct tom_data *d = TOM_DATA(dev);
        struct sock *sk = lookup_tid(&d->tid_maps, tid);
                                                                                
        if (!sk) {
            printk(KERN_ERR "CPL_SET_TCB_RPL for tid %u, NO sk!\n", tid);
        } else if (sock_flag(sk, SOCK_MIGRATE)) {
            PROCESS_CPL_MSG(t1_cm_rx_cpl, sk, skb);
            return 0;
        }
    }
#endif
    return CPL_RET_BUF_DONE;
}

/*
 * Process a GET_TCB_RPL CPL message.  Just log a message if we get bad status.
 */
static int do_get_tcb_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_get_tcb_rpl *rpl = cplhdr(skb);
                                                                                
    if (rpl->status != CPL_ERR_NONE) {
        printk(KERN_ERR
               "Unexpected status %d for CPL_GET_TCB_RPL for tid %u\n",
               rpl->status, GET_TID(rpl));
    }
#ifdef CONFIG_CHELSIO_T1_CM
    else {
        tid_t tid =  GET_TID(rpl);
        struct tom_data *d = TOM_DATA(dev);
        struct sock *sk = lookup_tid(&d->tid_maps, tid);
                                                                                
        if (!sk) {
            printk(KERN_ERR "CPL_GET_TCB_RPL for tid %u, NO sk!\n", tid);
        } else if (sock_flag(sk, SOCK_MIGRATE)) {
            PROCESS_CPL_MSG(t1_cm_rx_cpl, sk, skb);
            return 0;
        } else {
            printk(KERN_ERR
               "Unexpected CPL_GET_TCB_RPL for tid %u\n", GET_TID(rpl));
        }
    }
#endif
    return CPL_RET_BUF_DONE;
}

#ifdef CONFIG_CHELSIO_T1_CM
int t1_read_pm_memory(struct toedev *dev, struct sock *sk, __u32 addr, __u32 length)
{
    struct sk_buff *skb;
    struct cpl_pcmd_read *req;
    struct tom_data *d = TOM_DATA(dev);
    int atid = alloc_atid(&d->tid_maps);
  
    if (atid < 0) goto err;
    
    skb = alloc_skb(sizeof(struct cpl_pcmd_read), GFP_ATOMIC);
    if (!skb) goto free_atid;

    skb->priority = CPL_PRIORITY_CONTROL;
    skb->sk = sk;

    sock_hold(sk);
    ATID2ENTRY(&d->tid_maps, atid).sk = sk;

    req = (struct cpl_pcmd_read *)__skb_put(skb, sizeof(*req));
    OPCODE_TID(req) = htonl(MK_OPCODE_TID(CPL_PCMD_READ, atid));
    req->addr = htonl(addr);
    req->len = htons(length);

    toe_send(dev, skb);
    return 0;

free_atid:
    free_atid(&d->tid_maps, atid);

err:
    return -1;
}

static int do_pcmd_read_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_pcmd_read_rpl *rpl = cplhdr(skb);
    tid_t atid =  GET_TID(rpl);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_atid(&d->tid_maps, atid);

    if (!sk) {
        printk(KERN_ERR "%s: NO sk found for PCMD_READ_RPL atid=0x%x.\n",
               dev->name, atid);
        return CPL_RET_UNKNOWN_TID | CPL_RET_BUF_DONE;
    }

    free_atid(&d->tid_maps, atid);

    if (sock_flag(sk, SOCK_MIGRATE)) {
        PROCESS_CPL_MSG(t1_cm_rx_cpl, sk, skb);
        return 0;
    } 

    printk(KERN_ERR "%s: PCMD_READ_RPL, sk not in migration.\n",
           dev->name);
    return CPL_RET_BUF_DONE;
}
                                                                                
#endif
static struct dst_entry *route_req(struct sock *sk, struct open_request *req)
{
    struct rtable *rt;

#ifdef ip_route_output_flow
    struct rt_key fl = {
	.dst = req->af.v4_req.rmt_addr,
	.src = req->af.v4_req.loc_addr,
	.oif = sk->bound_dev_if,
	.tos = RT_CONN_FLAGS(sk)
    };
#else
    struct flowi fl = { .oif = sk->sk_bound_dev_if,
			.nl_u = { .ip4_u =
				  { .daddr = req->af.v4_req.rmt_addr,
				    .saddr = req->af.v4_req.loc_addr,
				    .tos = RT_CONN_FLAGS(sk) } },
			.proto = IPPROTO_TCP,
			.uli_u = { .ports =
				   { .sport = inet_addr_info(sk)->sport,
				     .dport = req->rmt_port } } };
#endif

    if (ip_route_output_flow(&rt, &fl, sk, 0)) {
	IP_INC_STATS_BH(IPSTATS_MIB_OUTNOROUTES);
	return NULL;
    }
    return &rt->u.dst;
}

/*
 * Process a CPL_SMT_WRITE_RPL.  Just log a message if we get bad status.
 * Note that the TID in the reply is really the SMT index it refers to.
 */
static int do_smt_write_rpl(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_smt_write_rpl *rpl = cplhdr(skb);

    if (rpl->status != CPL_ERR_NONE)
	printk(KERN_ERR
	       "Unexpected status %d for CPL_SMT_WRITE_RPL for entry %d\n",
	       rpl->status, GET_TID(rpl));

    return CPL_RET_BUF_DONE;
}

static int chelsio_or_send_synack(struct sock *sk, struct open_request *req,
				  struct dst_entry *dst)
{
    printk(KERN_ERR "%s() called\n", __FUNCTION__);
    return 0;
}

static void chelsio_or_send_ack(struct sk_buff *skb, struct open_request *req)
{
    printk(KERN_ERR "%s() called\n", __FUNCTION__);
}

static void chelsio_or_free(struct open_request *req)
{
    printk(KERN_ERR "%s() called\n", __FUNCTION__);
}

static void chelsio_or_send_reset(struct sk_buff *skb)
{
    printk(KERN_ERR "%s() called\n", __FUNCTION__);
}
    
static struct or_calltable or_chelsio = {
    .family      = PF_INET,
    .rtx_syn_ack = chelsio_or_send_synack,
    .send_ack    = chelsio_or_send_ack,
    .destructor  = chelsio_or_free,
    .send_reset  = chelsio_or_send_reset
};

/*
 * Handle an ARP failure for a CPL_PASS_ACCEPT_RPL.  We convert the message to
 * a reject and send it along.
 */
static void pass_accept_rpl_arp_failure(struct toedev *dev,
					struct sk_buff *skb)
{
    struct cpl_pass_accept_rpl *rpl = cplhdr(skb);
    struct sock *sk = skb->sk;

    rpl->opt0h = htonl(V_TCAM_BYPASS);
    rpl->status = CPL_PASS_OPEN_REJECT;
    toe_send(dev, skb);
    TCP_INC_STATS_BH(TCP_MIB_ATTEMPTFAILS);

    // XXX not entirely correct, need to deal with backlog and open request
    sock_hold(sk);
    local_bh_disable();
    bh_lock_sock(sk);
    BUG_TRAP(sk->sk_state == TCP_SYN_RECV);
    if (sk->sk_state == TCP_SYN_RECV) {
	t1_release_offload_resources(sk);
	connection_done(sk);
    }
    bh_unlock_sock(sk);
    local_bh_enable();
    sock_put(sk);
}

/*
 * Create a new socket as a child of the listening socket 'lsk' and initialize
 * with the information in the supplied PASS_ACCEPT_REQ message.
 */
static struct sock *mk_pass_sock(struct sock *lsk, struct toedev *dev, int tid,
				 struct cpl_pass_accept_req *req)
{
    struct sock *newsk;
    struct l2t_entry *e;
    struct dst_entry *dst;
    struct tcp_opt *newtp;
    struct open_request *oreq = tcp_openreq_alloc();

    if (!oreq) goto out_err;

    memset(oreq, 0, sizeof(*oreq));
    oreq->rcv_isn = ntohl(req->rcv_isn);
    oreq->rmt_port = req->peer_port;
    oreq->af.v4_req.loc_addr = req->local_ip;
    oreq->af.v4_req.rmt_addr = req->peer_ip;
    oreq->class = &or_chelsio;
    if (sysctl_tcp_window_scaling) {
	oreq->wscale_ok = 1;
	oreq->snd_wscale = req->tcp_options.wsf;
    }

    dst = route_req(lsk, oreq);
    if (!dst) goto free_or;

    e = l2t_get(dev, dst->neighbour);
    if (!e) goto free_dst;

    newsk = tcp_create_openreq_child(lsk, oreq, tcphdr_skb);
    if (!newsk) goto free_l2t;

    if (sock_flag(newsk, SOCK_KEEPOPEN))
	tcp_delete_keepalive_timer(newsk);
    newsk->sk_user_data = oreq;
    newsk->sk_dst_cache = dst;
    tcp_v4_setup_caps(newsk, dst);

    newtp = tcp_sk(newsk);
    TID(newtp) = tid;
    L2T_ENTRY(newtp) = e;
    TOE_DEV(newtp) = dev;
    TX_PAGES_ALLOCED(newtp) = 0;
    CLOSE_FLAGS(newtp) = 0;
    newtp->rcv_wnd = max(select_rcv_wnd(newsk), MIN_RCV_WND);
    newtp->recv_wscale = select_rcv_wscale(tcp_full_space(newsk),
					   newtp->winscale_ok);

    inet_addr_info(newsk)->daddr = req->peer_ip;
    inet_addr_info(newsk)->rcv_saddr = req->local_ip;
    inet_addr_info(newsk)->saddr = req->local_ip;

    __tcp_v4_hash(newsk, 0);
    __tcp_inherit_port(lsk, newsk);
    install_offload_ops(newsk);
    bh_unlock_sock(newsk);          // counters tcp_create_openreq_child()
    return newsk;

 free_l2t:
    l2t_release(L2DATA(dev), e);
 free_dst:
    dst_release(dst);
 free_or:
    tcp_openreq_fastfree(oreq);
 out_err:
    return NULL;
}

/*
 * Send a deferred reject to an accept request.
 */
static void reject_pass_accept(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_pass_accept_rpl *rpl;
    struct cpl_pass_accept_req *req = cplhdr(skb);
    tid_t tid = G_PASS_OPEN_TID(ntohl(req->tos_tid));
    struct sk_buff *reply_skb = alloc_skb_nofail(sizeof(*rpl));

    reply_skb->priority = CPL_PRIORITY_SETUP;
    rpl = cplhdr(reply_skb);
    OPCODE_TID(rpl) = htonl(MK_OPCODE_TID(CPL_PASS_ACCEPT_RPL, tid));
    rpl->peer_ip = req->peer_ip;         // req->peer_ip is not overwritten
    rpl->opt0h = htonl(V_TCAM_BYPASS);
    rpl->opt0l = 0;
    rpl->status = CPL_PASS_OPEN_REJECT;
    toe_send(dev, reply_skb);
    kfree_skb(skb);
}

/*
 * Process a CPL_PASS_ACCEPT_REQ message.  Does the part that needs the socket
 * lock held.  Note that the sock here is a listening socket that is not owned
 * by the TOE.
 */
static void process_pass_accept_req(struct sock *sk, struct sk_buff *skb)
{
    int rt_flags;
    struct sock *newsk;
    struct l2t_entry *e;
    struct tcp_opt *newtp;
    struct sk_buff *reply_skb;
    struct cpl_pass_accept_rpl *rpl;
    struct cpl_pass_accept_req *req = cplhdr(skb);
    tid_t tid = G_PASS_OPEN_TID(ntohl(req->tos_tid));
    struct toedev *dev = BLOG_SKB_CB(skb)->dev;
    struct tom_data *d = TOM_DATA(dev);
    struct net_device *netdev;

    reply_skb = get_cpl_reply_skb(skb, sizeof(*rpl), GFP_ATOMIC);
    if (unlikely(!reply_skb)) {
	t1_defer_reply(skb, dev, reject_pass_accept);
	goto out;
    }
    reply_skb->priority = CPL_PRIORITY_SETUP;

    if (sk->sk_state != TCP_LISTEN)
	goto reject;
    if (tcp_synq_is_full(sk))
	goto reject;
    if (sk_acceptq_is_full(sk) && d->conf.soft_backlog_limit)
	goto reject;

    BUG_TRAP(!skb->dst);

    if (d->max_nports > 1) {
        struct toe_dev_byhwaddr hwdev;
        hwdev.mac = req->dst_mac;
        if (dev->ctl(dev, TOE_GET_DEV_BY_HWADDR, &hwdev) < 0) {
            goto reject;
	}
        if ( (netdev = (struct net_device *)hwdev.dev) == NULL) {
            goto reject;
        }
    } else {
        netdev = dev->lldev;
    }

    if (ip_route_input(skb, req->local_ip, req->peer_ip,
		       G_PASS_OPEN_TOS(ntohl(req->tos_tid)), netdev))
	goto reject;

    rt_flags = ((struct rtable *)skb->dst)->rt_flags &
	(RTCF_BROADCAST | RTCF_MULTICAST | RTCF_LOCAL);
    dst_release(skb->dst);            // done with the input route, release it
    skb->dst = NULL;
    if (rt_flags != RTCF_LOCAL)
	goto reject;

    newsk = mk_pass_sock(sk, dev, tid, req);
    if (!newsk)
	goto reject;
    tcp_synq_added(sk);

    /* Don't get a reference, newsk starts out with ref count 2 */
    __sk_insert_tid(d, newsk, tid);

    reply_skb->sk = newsk;
    SET_ARP_FAILURE_HANDLER(reply_skb, pass_accept_rpl_arp_failure);

    newtp = tcp_sk(newsk);
    e = L2T_ENTRY(newtp);

    rpl = cplhdr(reply_skb);
    OPCODE_TID(rpl) = htonl(MK_OPCODE_TID(CPL_PASS_ACCEPT_RPL, tid));
    rpl->peer_ip = req->peer_ip;         // req->peer_ip is not overwritten
    rpl->opt0h = htonl(calc_opt0h(newsk, newtp) |
			V_MSS_IDX_OVERRIDE |
			V_MSS_IDX(best_mtu_index(dev, netdev->mtu)) |
			V_L2T_IDX(l2e_idx(L2DATA(dev), e)));
    rpl->opt0l = htonl(calc_opt0l(newsk, newtp));
    rpl->status = CPL_PASS_OPEN_ACCEPT;
    l2t_send(dev, reply_skb, e);
    kfree_skb(skb);
    return;

 reject:
    rpl = cplhdr(reply_skb);
    OPCODE_TID(rpl) = htonl(MK_OPCODE_TID(CPL_PASS_ACCEPT_RPL, tid));
    rpl->peer_ip = req->peer_ip;         // req->peer_ip is not overwritten
    rpl->opt0h = htonl(V_TCAM_BYPASS);
    rpl->opt0l = 0;
    rpl->status = CPL_PASS_OPEN_REJECT;
    toe_send(dev, reply_skb);
    kfree_skb(skb);
 out:
    TCP_INC_STATS_BH(TCP_MIB_ATTEMPTFAILS);
}

/*
 * Handle a CPL_PASS_ACCEPT_REQ message.  XXX lots missing
 */
static int do_pass_accept_req(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_pass_accept_req *req = cplhdr(skb);
    tid_t stid = GET_TID(req);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_stid(&d->tid_maps, stid);

#if VALIDATE_TID
    tid_t tid = G_PASS_OPEN_TID(ntohl(req->tos_tid));

    if (!sk)
	return CPL_RET_UNKNOWN_TID | CPL_RET_BUF_DONE;
    if (!valid_new_tid(&d->tid_maps, tid)) {
	printk(KERN_ERR "%s: passive open uses existing TID %d\n", dev->name,
	       tid);
	return CPL_RET_BUF_DONE;
    }
#endif

    BLOG_SKB_CB(skb)->dev = dev;
    PROCESS_CPL_MSG(process_pass_accept_req, sk, skb);
    return 0;
}

/*
 * Add a passively open socket to its parent's accept queue.  Note that the
 * child may be in any state by now, including TCP_CLOSE.  We can guarantee
 * though that it has not been orphaned yet.
 */
static void add_pass_open_to_parent(struct sock *child, struct sock *lsk,
				    struct toedev *dev)
{
    struct open_request *oreq = child->sk_user_data;

    child->sk_user_data = NULL;

    if (lsk->sk_state == TCP_LISTEN)
	tcp_synq_removed(lsk, oreq);

    if (lsk->sk_state != TCP_LISTEN ||
	(sk_acceptq_is_full(lsk) && !TOM_TUNABLE(dev, soft_backlog_limit))) {
	if (lsk->sk_state == TCP_LISTEN)
	    NET_INC_STATS_BH(LINUX_MIB_LISTENOVERFLOWS);
	NET_INC_STATS_BH(LINUX_MIB_LISTENDROPS);
	tcp_openreq_fastfree(oreq);
	add_to_reap_list(child);
    } else {
	tcp_acceptq_queue(lsk, oreq, child);
	lsk->sk_data_ready(lsk, 0);
    }
}

/*
 * This is run from a listener's backlog to add a child socket to its accept
 * queue.  Note that at this point the child is not locked and we intentionally
 * do not bother locking it as the only fields we may be using are sk_pair,
 * sk_user_data, and the open request and there aren't any concurrent users
 * for them.
 */
static void bl_add_pass_open_to_parent(struct sock *lsk, struct sk_buff *skb)
{
    struct sock *child = skb->sk;

    skb->sk = NULL;
    add_pass_open_to_parent(child, lsk, BLOG_SKB_CB(skb)->dev);
    __kfree_skb(skb);
}

/*
 * Process a CPL_PASS_ESTABLISH message.  XXX a lot of the locking doesn't work
 * if we are in TCP_SYN_RECV due to crossed SYNs
 */
static int do_pass_establish(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_pass_establish *req = cplhdr(skb);
    tid_t stid = GET_TID(req);
    tid_t tid = G_PASS_OPEN_TID(ntohl(req->tos_tid));
    struct tom_data *d = TOM_DATA(dev);
    struct sock *lsk = lookup_stid(&d->tid_maps, stid);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    /*
     * There isn't much we can validate here as in AUTO mode the TID is new and
     * the listening socket may also not exist anymore.
     */

    if (unlikely(!sk)) {                                // auto mode
	struct cpl_pass_accept_req par;

	// Synthesize a PASS_ACCEPT_REQ for mk_pass_sock()
	memset(&par, 0, sizeof(par));
	par.rcv_isn = htonl(ntohl(req->rcv_isn) - 1);  // compensate for SYN
	par.local_ip = req->local_ip;
	par.peer_ip = req->peer_ip;
	par.peer_port = req->peer_port;

	sk = mk_pass_sock(lsk, dev, tid, &par);
	if (!sk) return CPL_RET_BUF_DONE;   // XXX need to abort

	__sk_insert_tid(d, sk, tid);
    }

    bh_lock_sock(sk);
    if (unlikely(sock_owned_by_user(sk))) {
	// This can only happen in simultaneous opens.  XXX TBD
	__kfree_skb(skb);
    } else {
	// Finish off socket initialization now that we have the SND_ISN.
	struct tcp_opt *tp = tcp_sk(sk);
	u32 snd_isn = ntohl(req->snd_isn);  // real SND_ISN + 1

	tp->pushed_seq = tp->write_seq = tp->snd_nxt = tp->snd_una = snd_isn;
	SND_MAX(tp) = snd_isn;
	TX_PAGES_ALLOCED(tp) = 0;
	tp->mss_cache = d->conf.mss;
	inet_sk(sk)->id = tp->write_seq ^ jiffies;

	dst_confirm(sk->sk_dst_cache);

	mb();
	tcp_set_state(sk, TCP_ESTABLISHED);
	if (sk->sk_socket) {                       // simultaneous opens only
	    sk->sk_state_change(sk);
	    sk_wake_async(sk, 0, POLL_OUT);
	}

	if (unlikely(!lsk)) {         // parent has gone away in the meantime
	    if (sk->sk_user_data) {
		struct open_request *oreq = sk->sk_user_data;

		tcp_openreq_fastfree(oreq);
		sk->sk_user_data = NULL;
	    }
	    add_to_reap_list(sk);
	    __kfree_skb(skb);
	    goto unlock;
	}
	// XXX need to handle crossed SYNs
	if (TOM_TUNABLE(dev, tx_wscale))
	    t1_set_tx_wscale(sk, TOM_TUNABLE(dev, tx_wscale));

	/*
	 * The state for the new connection is now up to date.  All that
	 * remains is to add the new connection to its parent's accept queue.
	 */
	bh_lock_sock(lsk);
	if (likely(!sock_owned_by_user(lsk))) {
	    __kfree_skb(skb);
	    add_pass_open_to_parent(sk, lsk, dev);
	} else {
	    skb->sk = sk;
	    BLOG_SKB_CB(skb)->dev = dev;
	    SET_BLOG_CPL_HANDLER(skb, bl_add_pass_open_to_parent);
	    sk_add_backlog(lsk, skb);
	}
	bh_unlock_sock(lsk);
    }
 unlock:
    bh_unlock_sock(sk);
    return 0;
}

/*
 * Fill in the right TID for CPL messages waiting in the out-of-order queue
 * and send them to the TOE.
 */
static void fixup_and_send_ofo(struct tcp_opt *tp)
{
    tid_t tid = TID(tp);
    struct toedev *dev = TOE_DEV(tp);

    while (skb_queue_len(&tp->out_of_order_queue)) {
	struct sk_buff *skb = tp->out_of_order_queue.next;
	union opcode_tid *p = cplhdr(skb);

	__skb_unlink(skb, skb->list);
	p->opcode_tid = htonl(MK_OPCODE_TID(p->opcode, tid));
	toe_send(dev, skb);
    }
}

/*
 * Adjust the start sequence numbers of buffers in the write queue.
 */
static inline void fixup_pending_writeq_buffers(struct sock *sk,
						struct tcp_opt *tp)
{
    struct sk_buff *skb;

    skb_queue_walk(&sk->sk_write_queue, skb) {
	TCP_SKB_CB(skb)->seq = tp->write_seq;
	tp->write_seq += skb->len;
    }
}

/*
 * Updates socket state from an active establish CPL message.  Runs with the
 * socket lock held.
 */
static void sock_act_establish(struct sock *sk, struct sk_buff *skb)
{
    struct cpl_act_establish *req = cplhdr(skb);
    u32 snd_isn = ntohl(req->snd_isn);        /* real SND_ISN + 1 */
    u32 rcv_isn = ntohl(req->rcv_isn);        /* real RCV_ISN + 1 */
    struct tcp_opt *tp = tcp_sk(sk);

    BUG_TRAP(TX_PAGES_ALLOCED(tp) == 0);
    BUG_TRAP(sk->sk_state == TCP_SYN_SENT);
    if (sk->sk_state != TCP_SYN_SENT)
	printk(KERN_ERR "TID %u expected SYN_SENT, found %d\n", TID(tp),
	       sk->sk_state);

    tp->rcv_tstamp = tcp_time_stamp;
    tp->copied_seq = tp->rcv_wup = tp->rcv_nxt = rcv_isn;
    tp->pushed_seq = tp->write_seq = tp->snd_nxt = tp->snd_una = snd_isn;
    SND_MAX(tp) = snd_isn;
    inet_sk(sk)->id = tp->write_seq ^ jiffies;

    dst_confirm(sk->sk_dst_cache);

    /*
     * tcp_poll() does not lock socket, make sure initial values are committed
     * before changing to ESTABLISHED.
     */
    mb();
    tcp_set_state(sk, TCP_ESTABLISHED);

    /*
     * Now that we finally have a TID send any CPL messages that we had to
     * defer for lack of a TID.
     */
    if (skb_queue_len(&tp->out_of_order_queue))
	fixup_and_send_ofo(tp);

    if (likely(!sock_flag(sk, SOCK_DEAD))) {
	sk->sk_state_change(sk);
	sk_wake_async(sk, 0, POLL_OUT);
    }

    __kfree_skb(skb);

    if (TOM_TUNABLE(TOE_DEV(tp), tx_wscale))
	t1_set_tx_wscale(sk, TOM_TUNABLE(TOE_DEV(tp), tx_wscale));

    /*
     * Currently the send queue must be empty at this point because the socket
     * layer does not send anything before a connection is established.  To be
     * future proof though we handle the possibility that there are pending
     * buffers to send (either TX_DATA or CLOSE_CON_REQ).  First we need to
     * adjust the sequence number of the buffers according to the just
     * learned write_seq, and then we send them on their way.
     */
    fixup_pending_writeq_buffers(sk, tp);
    if (t1_push_frames(sk))
	sk->sk_write_space(sk);
}

/*
 * Process a CPL_ACT_ESTABLISH message.
 */
static int do_act_establish(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_act_establish *req = cplhdr(skb);
    tid_t atid = GET_TID(req);
    tid_t tid = G_PASS_OPEN_TID(ntohl(req->tos_tid));
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_atid(&d->tid_maps, atid);

#if VALIDATE_TID
    if (!sk)
	return CPL_RET_UNKNOWN_TID | CPL_RET_BUF_DONE;
    if (!valid_new_tid(&d->tid_maps, tid)) {
	printk(KERN_ERR "%s: active establish uses existing TID %u\n",
	       dev->name, tid);
	return CPL_RET_BUF_DONE;
    }
#endif

    TID(tcp_sk(sk)) = tid;
    sk_insert_tid(d, sk, tid);
    free_atid(&d->tid_maps, atid);

    PROCESS_CPL_MSG(sock_act_establish, sk, skb);
    return 0;
}

/*
 * Process an acknowledgment of TX data.  Advance snd_una and send the next
 * batch of TX data from the write queue.
 */
static void tx_ack(struct sock *sk, struct sk_buff *skb)
{
    struct tcp_opt *tp = tcp_sk(sk);
    struct tom_data *d = TOM_DATA(TOE_DEV(tp));
    struct cpl_tx_data_ack *hdr = cplhdr(skb);
    u32 ack_seq = ntohl(hdr->ack_seq);

    dst_confirm(sk->sk_dst_cache);
    tp->rcv_tstamp = tcp_time_stamp;

    if (unlikely(!after(ack_seq, tp->snd_una))) {
#if VALIDATE_SEQ
#if DEBUG
	if (ack_seq == tp->snd_una)
		printk(KERN_NOTICE
		       "%s:  TX_DATA_ACK for TID %u did not advance snd_una\n",
		       d->dev->name, TID(tp));
	else
		printk(KERN_ERR "%s: unexpected sequence # %u in TX_DATA_ACK "
		       "for TID %u, snd_una %u\n", d->dev->name, ack_seq,
		       TID(tp), tp->snd_una);
#endif
#endif
	goto out_free;
    }

    tp->snd_una = ack_seq;
    if (tp->snd_una == tp->write_seq && !WRITE_PENDING(sk, tp)) {
	free_tx_pages(tp, TX_PAGES_ALLOCED(tp));
	BUG_TRAP(tp->snd_nxt == SND_MAX(tp));
    } else {
	if (TX_PAGES_ALLOCED(tp) > 2 && pm_pressure(d) &&
	    tx_pages_used(tp, d) < TX_PAGES_ALLOCED(tp))
	    free_tx_pages(tp, 1);
	update_sndmax(tp, d);
	BUG_TRAP(!after(tp->snd_nxt, SND_MAX(tp)));
    }
#if VALIDATE_SEQ
    if (!between(tp->snd_nxt, tp->snd_una, SND_MAX(tp)))
	printk(KERN_ERR
	       "TID %u violation: snd_una %u, snd_max %u, snd_nxt %u\n",
	       TID(tp), tp->snd_una, SND_MAX(tp), tp->snd_nxt);
#endif

    if ((CLOSE_FLAGS(tp) & TX_HALT) && tp->snd_nxt == tp->snd_una) {
	CLOSE_FLAGS(tp) &= ~TX_HALT;
    }

    if (skb_queue_len(&sk->sk_write_queue) && t1_push_frames(sk))
	sk->sk_write_space(sk);

 out_free:
    __kfree_skb(skb);
}

/*
 * Handler for TX_DATA_ACK CPL messages.
 */
static int do_tx_data_ack(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_tx_data_ack *hdr = cplhdr(skb);
    tid_t tid = GET_TID(hdr);
    struct tom_data *d = TOM_DATA(dev);
    struct sock *sk = lookup_tid(&d->tid_maps, tid);

    VALIDATE_SOCK(d, sk);

    PROCESS_CPL_MSG(tx_ack, sk, skb);
    return 0;
}

/*
 * Process a MSS_CHANGE CPL message, internal driver -> TOM.
 */
static int do_mss_change(struct toedev *dev, struct sk_buff *skb)
{
    struct cpl_mss_change *req = cplhdr(skb);
    struct tom_data *d = TOM_DATA(dev);
    u32 	port = G_TID(ntohl(get_unaligned(&OPCODE_TID(req))));
    u32 	mss = ntohl(req->mss);

    if ( (port < d->max_nports) && (mss) ) {
        d->lmss[port] = mss;
    }

    return CPL_RET_BUF_DONE;
}

/*
 * Disconnect offloaded established but not yet accepted connections sitting
 * on a server's accept_queue.  We just send an ABORT_REQ at this point and
 * finish off the disconnect later as we may need to wait for the ABORT_RPL.
 */
void t1_disconnect_acceptq(struct sock *listen_sk)
{
    struct tcp_opt *tp = tcp_sk(listen_sk);
    struct open_request **pprev = &tp->accept_queue;

    while (*pprev) {
	struct open_request *req = *pprev;

	if (req->class == &or_chelsio) {   // one of ours
	    struct sock *child = req->sk;

	    *pprev = req->dl_next;
	    sk_acceptq_removed(listen_sk);
	    tcp_openreq_fastfree(req);
	    if (is_bound(child))           // release the port immediately
		tcp_put_port(child);
	    reset_listen_child(child);
	} else
	    pprev = &req->dl_next;
    }
}

int __init t1_init_cpl_io(void)
{
    tcphdr_skb = alloc_skb(sizeof(struct tcphdr), GFP_KERNEL);
    if (!tcphdr_skb) {
	printk(KERN_ERR "Chelsio TCP offload: can't allocate sk_buff\n");
	return -1;
    }
    skb_put(tcphdr_skb, sizeof(struct tcphdr));
    tcphdr_skb->h.raw = tcphdr_skb->data;
    memset(tcphdr_skb->data, 0, tcphdr_skb->len);

    t1_register_cpl_handler(CPL_PASS_ESTABLISH, do_pass_establish);
    t1_register_cpl_handler(CPL_ACT_ESTABLISH, do_act_establish);
    t1_register_cpl_handler(CPL_ACT_OPEN_RPL, do_act_open_rpl);
    t1_register_cpl_handler(CPL_PASS_ACCEPT_REQ, do_pass_accept_req);
    t1_register_cpl_handler(CPL_RX_DATA, do_rx_data);
    t1_register_cpl_handler(CPL_TX_DATA_ACK, do_tx_data_ack);
    t1_register_cpl_handler(CPL_PEER_CLOSE, do_peer_close);
    t1_register_cpl_handler(CPL_ABORT_REQ, do_abort_req);
    t1_register_cpl_handler(CPL_ABORT_RPL, do_abort_rpl);
    t1_register_cpl_handler(CPL_CLOSE_CON_RPL, do_close_con_rpl);
    t1_register_cpl_handler(CPL_L2T_WRITE_RPL, do_l2t_write_rpl);
    t1_register_cpl_handler(CPL_SMT_WRITE_RPL, do_smt_write_rpl);
    t1_register_cpl_handler(CPL_SET_TCB_RPL, do_set_tcb_rpl);
    t1_register_cpl_handler(CPL_GET_TCB_RPL, do_get_tcb_rpl);
    t1_register_cpl_handler(CPL_MSS_CHANGE, do_mss_change);
#ifdef CONFIG_CHELSIO_T1_CM
    t1_register_cpl_handler(CPL_PCMD_READ_RPL, do_pcmd_read_rpl);
#endif
    return 0;
}

// Local Variables:
// c-basic-offset: 4
// End:
